/************************************************************************
 *									*
 *            FAT.C							*
 *            =======							*
 *                                        W.G. Reissnegger		*
 *                                        C&C Research Labs		*
 *                                        NEC USA Inc.			*
 *                                        4 Independence Way		*
 *                                        Princeton NJ 08540		*
 *									*
 *   DOS FAT12 and FAT16 operations on the FAT itself                   *
 *                                                                      *
 *   Copyright (C) 1997, 1998, 1999, 2000 NEC USA, Inc.			*
 *                 All Rights Reserved					*
 *									*
 *   This program is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License as     *
 *   published by the Free Software Foundation; either version 2 of     *
 *   the license, or (at your option) any later version.                *
 *									*
 *   This program is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of     *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      *
 *   GNU General Public License for more details.                       *
 *                                                                      *
 *   You should have received a copy of the GNU General Public          *
 *   License along with this program;  if not, write to the             *
 *       Free Software Foundation, Inc.,                                *
 *       59 Temple Place - Suite 330,                                   *
 *       Boston, MA 02111-1307                                          *
 *                                                                      *
 ************************************************************************/

#include <Messages.h>
#include <rome_if.h>
#include <string.h>
#include "msdos.h"

/************************************************************************/

#define FAT_MEM_DEBUG			0xfa

#define D(c)

/************************************************************************/

int dos_fat_read(FILE *fp, MSDOS_FS_INFO *inf)
{
    ROME_MESSAGE msg;
    char *p;
    int fat_len = inf->fat_length;
    uint fat_lba = inf->fat_lba;

    inf->fat_entries = (inf->fat_length * inf->sector_size) / 2;

    if (DOS_FAT12 == inf->fat_type)
    {
	inf->fat_entries >>= 1;
	inf->fat_entries *= 3;
    }
    else if (DOS_FAT16 != inf->fat_type)
    {
	rome_fatal("Invalid FAT type\n");
    }

    p = inf->fat = rome_alloc(inf->fat_length * inf->sector_size, FAT_MEM_DEBUG, 0);

    while (fat_len)
    {
	msg.b_immed = fat_lba;
        msg.b_wptr = (uchar *)(inf->sector_size);
	rome_getmblk(fp, &msg);
	rome_await_message(&msg, 0);

	if  (msg.m_errno)
	{
	    return msg.m_errno;
	}
	memcpy(p, msg.b_rptr, inf->sector_size);

	rome_retmblk(&msg);

	p += inf->sector_size;
	fat_len--;
	fat_lba++;
    }
    return 0;
}

int dos_fat_write(FILE *fp, MSDOS_FS_INFO *inf)
{
    ROME_MESSAGE msg;
    char *p;
    int fat_len = inf->fat_length;
    uint fat_lba = inf->fat_lba;

    D(rome_kprintf("fat_write(): DEBUG Writing FAT (lba %d)\n", fat_lba));

    p = inf->fat;

    while (fat_len)
    {
	rome_newmblk(fp, &msg, inf->sector_size);
	rome_await_message(&msg, 0);

	msg.b_immed = fat_lba;
	memcpy(msg.b_wptr, p, inf->sector_size);

	msg.b_wptr += inf->sector_size;

	rome_outmblk(fp, &msg);
	rome_await_message(&msg, 0);

	/* write the FAT copy */

	msg.b_immed = fat_lba + inf->fat_length;
	rome_outmblk(fp, &msg);
	rome_await_message(&msg, 0);

	rome_retmblk(&msg);

	p += inf->sector_size;
	fat_len--;
	fat_lba++;
    }

    return 0;
}

ushort dos_fat_lookup_entry(ushort idx, MSDOS_FS_INFO *inf)
{
    ushort adj_idx = 0;
    ushort tmp;
    ushort entry = 0;		/* make compiler happy */

    if (idx >= inf->fat_entries)
    {
	rome_kprintf("fat_lookup_entry(): Illegal FAT index %d\n", idx);
	return 0;
    }

    if (DOS_FAT16 == inf->fat_type)
    {
	idx <<= 1;
	entry = *((ushort *) & inf->fat[idx]);

	if (FAT16_IS_EOF(entry))
	{
	    entry = 1;
	}
    }
    else if (DOS_FAT12 == inf->fat_type)
    {
	adj_idx = (idx >> 1) * 3 + (idx & 1);
	tmp = inf->fat[adj_idx] | inf->fat[adj_idx + 1] << 8;
	entry = (idx & 1) ? tmp >> 4 : tmp & 0x0fff;

	if (FAT12_IS_EOF(entry))
	{
	    entry = 1;
	}
    }
    else
    {
	rome_fatal("Illegal FAT type\n");
    }

    return entry;
}

int dos_fat_write_entry(MSDOS_FS_INFO *inf, ushort idx, ushort entry)
{
    ushort adj_idx = 0;

    if (idx >= inf->fat_entries)
    {
	rome_kprintf("fat_write_entry(): Illegal FAT index %d\n", idx);
	return -1;
    }

    if (DOS_FAT16 == inf->fat_type)
    {
	idx <<= 1;
	*((ushort *) & inf->fat[idx]) = entry;
    }
    else if (DOS_FAT12 == inf->fat_type)
    {
	/* check validity of entry */
	if (entry & 0xf000)
	{
	    if (entry != FAT_EOF)
	    {
		rome_kprintf("fat_write_entry(): Invalid entry 0x%04x, "
			     "truncating it to 12 bit\n", entry);
	    }
	    entry &= 0xfff;
	}
	adj_idx = (idx >> 1) * 3 + (idx & 1);
	if (idx & 1)
	{
	    entry <<= 4;
	    inf->fat[adj_idx] = (inf->fat[adj_idx] & 0x0f) | (entry & 0xf0);
	    inf->fat[adj_idx + 1] = (entry >> 8);
	}
	else
	{
	    inf->fat[adj_idx] = entry & 0xff;
	    inf->fat[adj_idx + 1] = (inf->fat[adj_idx + 1] & 0xf0) |
		(entry >> 8);
	}
    }
    else
    {
	rome_fatal("Illegal FAT type\n");
    }
    return 0;
}

ushort dos_fat_lookup_free(MSDOS_FS_INFO *inf)
{
    int idx = 2;

    while (inf->fat_entries > idx)
    {
	if (0 == dos_fat_lookup_entry(idx, inf))
	{
	    D(rome_kprintf("FAT-free at %x\n", idx);)
	    return idx;
	}
	idx++;
    }

    rome_fatal("full FAT");
    return 0;
}

ushort dos_fat_seek_entry(MSDOS_FS_INFO *inf, ushort start, register ushort idx)
{
    register ushort entry = start;

    while (idx && (1 != entry))
    {
	entry = dos_fat_lookup_entry(entry, inf);
	idx--;
    }

    /* if not the last entry report 0 (idx not in file scope) */

    return idx ? 0 : entry;
}

int dos_fat_free_chain(MSDOS_FS_INFO *inf, ushort start)
{
    int count = 0;
    ushort current = start;

    if (start >= inf->fat_entries)
    {
	rome_kprintf("fat_free_chain(): Illegal FAT index %d\n", start);
	return 0;
    }

    while (1 < current)		/* 1 = EOF, 0 = free */
    {
	ushort next = dos_fat_lookup_entry(current, inf);

	/* mark current free */
	if (0 != dos_fat_write_entry(inf, current, 0))
	{
	    rome_kprintf("fat_free_chain(): failed\n");
	    return -1;
	}
	count++;
	current = next;
    }

    D(rome_kprintf("fat_free_chain(): DEBUG Freed %d FAT entries\n", count));

    return 0;
}
