#ifndef MMSP2_940_2_H_INCLUDED
#define MMSP2_940_2_H_INCLUDED
/* **************************************************************************
 *    Copyright (C) 2005 by Rob Brown                                       *
 *    rob@cobbleware.com                                                    *
 *                                                                          *
 *    This program is free software; you can redistribute it and/or modify  *
 *    it under the terms of the GNU General Public License as published by  *
 *    the Free Software Foundation; either version 2 of the License, or     *
 *    (at your option) any later version.                                   *
 *                                                                          *
 *    This program is distributed in the hope that it will be useful,       *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *    GNU General Public License for more details.                          *
 *                                                                          *
 *    You should have received a copy of the GNU General Public License     *
 *    along with this program; if not, write to the                         *
 *    Free Software Foundation, Inc.,                                       *
 *    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 * **************************************************************************
 */

/* This file defines the MMSP2 special function registers.
 * 
 * The registers are each defined in 3 forms:
 * 
 * 1. As a constant number, the address of the register. These are prefixed with "nMSP_"
 *    e.g. nMSP_SPECIAL_FUNCTION_REGISTER <=> (0xC0001234)
 * 
 * 2. As a pointer to the register. These are prefixed with "pMSP_". The type of the register is defined
 *    by its size, ie. "unsigned char" for 1 byte, "unsigned short" for 2 bytes, and "unsigned int" for 4 bytes.
 *    e.g. pMSP_SPECIAL_FUNCTION_REGISTER <=> ((volatile unsigned short *)nMSP_SPECIAL_FUNCTION_REGISTER)
 * 
 * 3. As a memory access which can be used as an l-value or an r-value. These are prefixed with "MSP_"
 *    e.g. MSP_SPECIAL_FUNCTION_REGISTER <=> (*pMSP_SPECIAL_FUNCTION_REGISTER)
 */

/* ***** Memory Controller ***** */

#define nMSP_MEMCFGW              (0xBE003A00)
#define nMSP_MEMTIMEW0            (0xBE003A02)
#define nMSP_MEMTIMEW1            (0xBE003A04)
#define nMSP_MRMTIMEW2            (0xBE003A06)
#define nMSP_MEMTIMEW3            (0xBE003A08)
#define nMSP_MEMTIMEW4            (0xBE003A0A)
#define nMSP_MEMWAITCTRLW         (0xBE003A0E)
#define nMSP_MEMPAGEW             (0xBE003A10)
#define nMSP_MEMIDETIMEW          (0xBE003A12)
#define nMSP_MEMPCMCIAMW          (0xBE003A14)
#define nMSP_MEMPCMCIAAW          (0xBE003A16)
#define nMSP_MEMPCMCIAIW          (0xBE003A18)
#define nMSP_MEMPCMCIAWAITW       (0xBE003A1A)
#define nMSP_MEMIDEWAITW          (0xBE003A1C)
#define nMSP_MEMDTIMEOUTW         (0xBE003A20)
#define nMSP_MEMDMACTRLW          (0xBE003A22)
#define nMSP_MEMDMAPOLW           (0xBE003A24)
#define nMSP_MEMDMATIMEW0         (0xBE003A26)
#define nMSP_MEMDMATIMEW1         (0xBE003A28)
#define nMSP_MEMDMATIMEW2         (0xBE003A2A)
#define nMSP_MEMDMATIMEW3         (0xBE003A2C)
#define nMSP_MEMDMATIMEW4         (0xBE003A2E)
#define nMSP_MEMDMATIMEW5         (0xBE003A30)
#define nMSP_MEMDMATIMEW6         (0xBE003A32)
/* Reserved... */
#define nMSP_MEMDMATIMEW7         (0xBE003A34)
#define nMSP_MEMDMATIMEW8         (0xBE003A36)
#define nMSP_MEMDMASTRBW          (0xBE003A38)

/* NAND Flash control */
#define nMSP_MEMNANDCTRLW         (0xBE003A3A)
#define nMSP_MEMNANDTIMEW         (0xBE003A3C)
#define nMSP_MEMNANDECC0W         (0xBE003A3E)
#define nMSP_MEMNANDECC1W         (0xBE003A40)
#define nMSP_MEMNANDECC2W         (0xBE003A42)
#define nMSP_MEMNANDCNTW          (0xBE003A44)

/* NAND Flash Access
 * 
 * Beware! These addresses will change if the Shadow setting is changed. The datasheet says that
 * the CPU can change the Shadow setting, but doesn't say how. At power-up, Shadow is set by an
 * external pin.
 */
#define nMSP_NFDATA               (0x9A000000)
#define nMSP_NFCMD                (0x9A000010)
#define nMSP_NFADDR               (0x9A000018)

/* Bank A Memory Control */
#define nMSP_MEMCFGX              (0xBE003800)
#define nMSP_MEMTIMEX0            (0xBE003802)
#define nMSP_MEMTIMEX1            (0xBE003804)
#define nMSP_MEMACTPWDX           (0xBE003806)
#define nMSP_MEMREFX              (0xBE003808)

/* ***** Dual CPU Interface ***** */

#define nMSP_DUAL920DATA0         (0xBE003B00)
#define nMSP_DUAL920DATA1         (0xBE003B02)
#define nMSP_DUAL920DATA2         (0xBE003B04)
#define nMSP_DUAL920DATA3         (0xBE003B06)
#define nMSP_DUAL920DATA4         (0xBE003B08)
#define nMSP_DUAL920DATA5         (0xBE003B0A)
#define nMSP_DUAL920DATA6         (0xBE003B0C)
#define nMSP_DUAL920DATA7         (0xBE003B0E)
#define nMSP_DUAL920DATA8         (0xBE003B10)
#define nMSP_DUAL920DATA9         (0xBE003B12)
#define nMSP_DUAL920DATA10        (0xBE003B14)
#define nMSP_DUAL920DATA11        (0xBE003B16)
#define nMSP_DUAL920DATA12        (0xBE003B18)
#define nMSP_DUAL920DATA13        (0xBE003B1A)
#define nMSP_DUAL920DATA14        (0xBE003B1C)
#define nMSP_DUAL920DATA15        (0xBE003B1E)

#define nMSP_DUAL940DATA0         (0xBE003B20)
#define nMSP_DUAL940DATA1         (0xBE003B22)
#define nMSP_DUAL940DATA2         (0xBE003B24)
#define nMSP_DUAL940DATA3         (0xBE003B26)
#define nMSP_DUAL940DATA4         (0xBE003B28)
#define nMSP_DUAL940DATA5         (0xBE003B2A)
#define nMSP_DUAL940DATA6         (0xBE003B2C)
#define nMSP_DUAL940DATA7         (0xBE003B2E)
#define nMSP_DUAL940DATA8         (0xBE003B30)
#define nMSP_DUAL940DATA9         (0xBE003B32)
#define nMSP_DUAL940DATA10        (0xBE003B34)
#define nMSP_DUAL940DATA11        (0xBE003B36)
#define nMSP_DUAL940DATA12        (0xBE003B38)
#define nMSP_DUAL940DATA13        (0xBE003B3A)
#define nMSP_DUAL940DATA14        (0xBE003B3C)
#define nMSP_DUAL940DATA15        (0xBE003B3E)

#define nMSP_DUALINT920           (0xBE003B40)
#define nMSP_DUALINT940           (0xBE003B42)
#define nMSP_DUALPEND920          (0xBE003B44)
#define nMSP_DUALPEND940          (0xBE003B46)
#define nMSP_DUALCTRL940          (0xBE003B48)

/* ***** Clocks and Power Management ***** */

#define nMSP_PWMODEREG            (0xBE000900)
#define nMSP_CLKCHGSTREG          (0xBE000902)
#define nMSP_SYSCLKENREG          (0xBE000904)
#define nMSP_COMCLKENREG          (0xBE000908)
#define nMSP_VGCLKENREG           (0xBE00090A)
#define nMSP_ASCLKENREG           (0xBE00090C)
#define nMSP_FPLLSETVREG          (0xBE000910)
#define nMSP_FPLLVSETREG          (0xBE000912)
#define nMSP_UPLLSETVREG          (0xBE000914)
#define nMSP_UPLLVSETREG          (0xBE000916)
#define nMSP_APLLSETVREG          (0xBE000918)
#define nMSP_APLLVSETREG          (0xBE00091A)
#define nMSP_SYSCSETREG           (0xBE00091C)
#define nMSP_ESYSCSETREG          (0xBE00091E)
#define nMSP_UIRMCSETREG          (0xBE000920)
#define nMSP_AUDICSETREG          (0xBE000922)
#define nMSP_DISPCSETREG          (0xBE000924)
#define nMSP_IMAGCSETREG          (0xBE000926)
#define nMSP_URT0CSETREG          (0xBE000928)
#define nMSP_URT1CSETREG          (0xBE00092A)

/* Linux code in open2x kernel source disagrees with this! It puts "A940TMODE" at C000092C
 * and doesn't seem to have SPDICSETREG at all.
 */
#define nMSP_A940TIDLE            (0xBE00092C)
#define nMSP_SPDICSETREG          (0xBE00092E)

/* ***** Real Time Clock *****
 * 
 * The preliminary data sheet gets a little confused about these, and lists several
 * of them twice. That's OK apart from the RTC Interrupt Status Register, which is
 * listed as "RTSTRLREG" and "RTCSTREG". Luckily, both of these are given the same
 * address which is C0000C24. The name RTCSTREG is used here.
 */
#define nMSP_RTCTSETREG           (0xBE000C00)
#define nMSP_RTCTCNTREG           (0xBE000C04)
#define nMSP_RTCSTCNTREG          (0xBE000C08)
#define nMSP_TICKSETREG           (0xBE000C0A)
#define nMSP_ALARMTREG            (0xBE000C0C)
#define nMSP_PWRMGRREG            (0xBE000C10)
#define nMSP_CLKMGRREG            (0xBE000C12)
#define nMSP_RSTCTRLREG           (0xBE000C14)
#define nMSP_RSTSTREG             (0xBE000C16)
#define nMSP_BOOTCTRLREG          (0xBE000C18)
#define nMSP_LOCKTIMEREG          (0xBE000C1A)
#define nMSP_RSTTIMEREG           (0xBE000C1C)
#define nMSP_EXTCTRLREG           (0xBE000C1E)
#define nMSP_STOPTSETREG          (0xBE000C20)
#define nMSP_RTCCTRLREG           (0xBE000C22)
#define nMSP_RTCSTREG             (0xBE000C24)

/* ***** DMA Controller ***** */

#define nMSP_DMAINT               (0xBE000000)

#define nMSP_DCH0SRM              (0xBE000100)
#define nMSP_DCH0TRM              (0xBE000102)
#define nMSP_DCH1SRM              (0xBE000104)
#define nMSP_DCH1TRM              (0xBE000106)
#define nMSP_DCH2SRM              (0xBE000108)
#define nMSP_DCH2TRM              (0xBE00010A)
#define nMSP_DCH3SRM              (0xBE00010C)
#define nMSP_DCH3TRM              (0xBE00010E)
#define nMSP_DCH4SRM              (0xBE000110)
#define nMSP_DCH4TRM              (0xBE000112)
#define nMSP_DCH5SRM              (0xBE000114)
#define nMSP_DCH5TRM              (0xBE000116)
#define nMSP_DCH6SRM              (0xBE000118)
#define nMSP_DCH6TRM              (0xBE00011A)
#define nMSP_DCH7SRM              (0xBE00011C)
#define nMSP_DCH7TRM              (0xBE00011E)
#define nMSP_DCH8SRM              (0xBE000120)
#define nMSP_DCH8TRM              (0xBE000122)
#define nMSP_DCH9SRM              (0xBE000124)
#define nMSP_DCH9TRM              (0xBE000126)
#define nMSP_DCH10SRM             (0xBE000128)
#define nMSP_DCH10TRM             (0xBE00012A)
#define nMSP_DCH11SRM             (0xBE00012C)
#define nMSP_DCH11TRM             (0xBE00012E)
#define nMSP_DCH12SRM             (0xBE000130)
#define nMSP_DCH12TRM             (0xBE000132)
#define nMSP_DCH13SRM             (0xBE000134)
#define nMSP_DCH13TRM             (0xBE000136)
#define nMSP_DCH14SRM             (0xBE000138)
#define nMSP_DCH14TRM             (0xBE00013A)
#define nMSP_DCH15SRM             (0xBE00013C)
#define nMSP_DCH15TRM             (0xBE00013E)

#define nMSP_DMA0COM0             (0xBE000200)
#define nMSP_DMA0COM1             (0xBE000202)
#define nMSP_DMA0COM2             (0xBE000204)
#define nMSP_DMA0CONS             (0xBE000206)
#define nMSP_SRC0LADDR            (0xBE000208)
#define nMSP_SRC0HADDR            (0xBE00020A)
#define nMSP_TRG0LADDR            (0xBE00020C)
#define nMSP_TRG0HADDR            (0xBE00020E)

#define nMSP_DMA1COM0             (0xBE000210)
#define nMSP_DMA1COM1             (0xBE000212)
#define nMSP_DMA1COM2             (0xBE000214)
#define nMSP_DMA1CONS             (0xBE000216)
#define nMSP_SRC1LADDR            (0xBE000218)
#define nMSP_SRC1HADDR            (0xBE00021A)
#define nMSP_TRG1LADDR            (0xBE00021C)
#define nMSP_TRG1HADDR            (0xBE00021E)

#define nMSP_DMA2COM0             (0xBE000220)
#define nMSP_DMA2COM1             (0xBE000222)
#define nMSP_DMA2COM2             (0xBE000224)
#define nMSP_DMA2CONS             (0xBE000226)
#define nMSP_SRC2LADDR            (0xBE000228)
#define nMSP_SRC2HADDR            (0xBE00022A)
#define nMSP_TRG2LADDR            (0xBE00022C)
#define nMSP_TRG2HADDR            (0xBE00022E)

#define nMSP_DMA3COM0             (0xBE000230)
#define nMSP_DMA3COM1             (0xBE000232)
#define nMSP_DMA3COM2             (0xBE000234)
#define nMSP_DMA3CONS             (0xBE000236)
#define nMSP_SRC3LADDR            (0xBE000238)
#define nMSP_SRC3HADDR            (0xBE00023A)
#define nMSP_TRG3LADDR            (0xBE00023C)
#define nMSP_TRG3HADDR            (0xBE00023E)

#define nMSP_DMA4COM0             (0xBE000240)
#define nMSP_DMA4COM1             (0xBE000242)
#define nMSP_DMA4COM2             (0xBE000244)
#define nMSP_DMA4CONS             (0xBE000246)
#define nMSP_SRC4LADDR            (0xBE000248)
#define nMSP_SRC4HADDR            (0xBE00024A)
#define nMSP_TRG4LADDR            (0xBE00024C)
#define nMSP_TRG4HADDR            (0xBE00024E)

#define nMSP_DMA5COM0             (0xBE000250)
#define nMSP_DMA5COM1             (0xBE000252)
#define nMSP_DMA5COM2             (0xBE000254)
#define nMSP_DMA5CONS             (0xBE000256)
#define nMSP_SRC5LADDR            (0xBE000258)
#define nMSP_SRC5HADDR            (0xBE00025A)
#define nMSP_TRG5LADDR            (0xBE00025C)
#define nMSP_TRG5HADDR            (0xBE00025E)

#define nMSP_DMA6COM0             (0xBE000260)
#define nMSP_DMA6COM1             (0xBE000262)
#define nMSP_DMA6COM2             (0xBE000264)
#define nMSP_DMA6CONS             (0xBE000266)
#define nMSP_SRC6LADDR            (0xBE000268)
#define nMSP_SRC6HADDR            (0xBE00026A)
#define nMSP_TRG6LADDR            (0xBE00026C)
#define nMSP_TRG6HADDR            (0xBE00026E)

#define nMSP_DMA7COM0             (0xBE000270)
#define nMSP_DMA7COM1             (0xBE000272)
#define nMSP_DMA7COM2             (0xBE000274)
#define nMSP_DMA7CONS             (0xBE000276)
#define nMSP_SRC7LADDR            (0xBE000278)
#define nMSP_SRC7HADDR            (0xBE00027A)
#define nMSP_TRG7LADDR            (0xBE00027C)
#define nMSP_TRG7HADDR            (0xBE00027E)

#define nMSP_DMA8COM0             (0xBE000280)
#define nMSP_DMA8COM1             (0xBE000282)
#define nMSP_DMA8COM2             (0xBE000284)
#define nMSP_DMA8CONS             (0xBE000286)
#define nMSP_SRC8LADDR            (0xBE000288)
#define nMSP_SRC8HADDR            (0xBE00028A)
#define nMSP_TRG8LADDR            (0xBE00028C)
#define nMSP_TRG8HADDR            (0xBE00028E)

#define nMSP_DMA9COM0             (0xBE000290)
#define nMSP_DMA9COM1             (0xBE000292)
#define nMSP_DMA9COM2             (0xBE000294)
#define nMSP_DMA9CONS             (0xBE000296)
#define nMSP_SRC9LADDR            (0xBE000298)
#define nMSP_SRC9HADDR            (0xBE00029A)
#define nMSP_TRG9LADDR            (0xBE00029C)
#define nMSP_TRG9HADDR            (0xBE00029E)

#define nMSP_DMA10COM0            (0xBE0002A0)
#define nMSP_DMA10COM1            (0xBE0002A2)
#define nMSP_DMA10COM2            (0xBE0002A4)
#define nMSP_DMA10CONS            (0xBE0002A6)
#define nMSP_SRC10LADDR           (0xBE0002A8)
#define nMSP_SRC10HADDR           (0xBE0002AA)
#define nMSP_TRG10LADDR           (0xBE0002AC)
#define nMSP_TRG10HADDR           (0xBE0002AE)

#define nMSP_DMA11COM0            (0xBE0002B0)
#define nMSP_DMA11COM1            (0xBE0002B2)
#define nMSP_DMA11COM2            (0xBE0002B4)
#define nMSP_DMA11CONS            (0xBE0002B6)
#define nMSP_SRC11LADDR           (0xBE0002B8)
#define nMSP_SRC11HADDR           (0xBE0002BA)
#define nMSP_TRG11LADDR           (0xBE0002BC)
#define nMSP_TRG11HADDR           (0xBE0002BE)

#define nMSP_DMA12COM0            (0xBE0002C0)
#define nMSP_DMA12COM1            (0xBE0002C2)
#define nMSP_DMA12COM2            (0xBE0002C4)
#define nMSP_DMA12CONS            (0xBE0002C6)
#define nMSP_SRC12LADDR           (0xBE0002C8)
#define nMSP_SRC12HADDR           (0xBE0002CA)
#define nMSP_TRG12LADDR           (0xBE0002CC)
#define nMSP_TRG12HADDR           (0xBE0002CE)

#define nMSP_DMA13COM0            (0xBE0002D0)
#define nMSP_DMA13COM1            (0xBE0002D2)
#define nMSP_DMA13COM2            (0xBE0002D4)
#define nMSP_DMA13CONS            (0xBE0002D6)
#define nMSP_SRC13LADDR           (0xBE0002D8)
#define nMSP_SRC13HADDR           (0xBE0002DA)
#define nMSP_TRG13LADDR           (0xBE0002DC)
#define nMSP_TRG13HADDR           (0xBE0002DE)

#define nMSP_DMA14COM0            (0xBE0002E0)
#define nMSP_DMA14COM1            (0xBE0002E2)
#define nMSP_DMA14COM2            (0xBE0002E4)
#define nMSP_DMA14CONS            (0xBE0002E6)
#define nMSP_SRC14LADDR           (0xBE0002E8)
#define nMSP_SRC14HADDR           (0xBE0002EA)
#define nMSP_TRG14LADDR           (0xBE0002EC)
#define nMSP_TRG14HADDR           (0xBE0002EE)

#define nMSP_DMA15COM0            (0xBE0002F0)
#define nMSP_DMA15COM1            (0xBE0002F2)
#define nMSP_DMA15COM2            (0xBE0002F4)
#define nMSP_DMA15CONS            (0xBE0002F6)
#define nMSP_SRC15LADDR           (0xBE0002F8)
#define nMSP_SRC15HADDR           (0xBE0002FA)
#define nMSP_TRG15LADDR           (0xBE0002FC)
#define nMSP_TRG15HADDR           (0xBE0002FE)

/* ***** Interrupt Controller ***** */

#define nMSP_SRCPEND              (0xBE000800)
#define nMSP_INTMODE              (0xBE000804)
#define nMSP_INTMASK              (0xBE000808)
#define nMSP_PRIORITY             (0xBE00080C)
#define nMSP_INTPEND              (0xBE000810)
#define nMSP_INTOFFSET            (0xBE000814)

/* ***** Timer / Watchdog Timer ***** */

#define nMSP_TCOUNT               (0xBE000A00)
#define nMSP_TMATCH0              (0xBE000A04)
#define nMSP_TMATCH1              (0xBE000A08)
#define nMSP_TMATCH2              (0xBE000A0C)
#define nMSP_TMATCH3              (0xBE000A10)
#define nMSP_TCONTROL             (0xBE000A14)
#define nMSP_TSTATUS              (0xBE000A16)
#define nMSP_TINTEN               (0xBE000A18)

/* ***** UART ***** */

#define nMSP_LCON0                (0xBE001200)
#define nMSP_UCON0                (0xBE001202)
#define nMSP_FCON0                (0xBE001204)
#define nMSP_MCON0                (0xBE001206)
#define nMSP_TRSTATUS0            (0xBE001208)
#define nMSP_ESTATUS0             (0xBE00120A)
#define nMSP_FSTATUS0             (0xBE00120C)
#define nMSP_MSTATUS0             (0xBE00120E)
#define nMSP_THB0                 (0xBE001210)
#define nMSP_RHB0                 (0xBE001212)
#define nMSP_BRD0                 (0xBE001214)
#define nMSP_TIMEOUTREG0          (0xBE001216)

#define nMSP_LCON1                (0xBE001220)
#define nMSP_UCON1                (0xBE001222)
#define nMSP_FCON1                (0xBE001224)
#define nMSP_MCON1                (0xBE001226)
#define nMSP_TRSTATUS1            (0xBE001228)
#define nMSP_ESTATUS1             (0xBE00122A)
#define nMSP_FSTATUS1             (0xBE00122C)
#define nMSP_MSTATUS1             (0xBE00122E)
#define nMSP_THB1                 (0xBE001230)
#define nMSP_RHB1                 (0xBE001232)
#define nMSP_BRD1                 (0xBE001234)
#define nMSP_TIMEOUTREG1          (0xBE001236)

#define nMSP_LCON2                (0xBE001240)
#define nMSP_UCON2                (0xBE001242)
#define nMSP_FCON2                (0xBE001244)
#define nMSP_MCON2                (0xBE001246)
#define nMSP_TRSTATUS2            (0xBE001248)
#define nMSP_ESTATUS2             (0xBE00124A)
#define nMSP_FSTATUS2             (0xBE00124C)
#define nMSP_MSTATUS2             (0xBE00124E)
#define nMSP_THB2                 (0xBE001250)
#define nMSP_RHB2                 (0xBE001252)
#define nMSP_BRD2                 (0xBE001254)
#define nMSP_TIMEOUTREG2          (0xBE001256)

#define nMSP_LCON3                (0xBE001260)
#define nMSP_UCON3                (0xBE001262)
#define nMSP_FCON3                (0xBE001264)
#define nMSP_MCON3                (0xBE001266)
#define nMSP_TRSTATUS3            (0xBE001268)
#define nMSP_ESTATUS3             (0xBE00126A)
#define nMSP_FSTATUS3             (0xBE00126C)
#define nMSP_MSTATUS3             (0xBE00126E)
#define nMSP_THB3                 (0xBE001270)
#define nMSP_RHB3                 (0xBE001272)
#define nMSP_BRD3                 (0xBE001274)
#define nMSP_TIMEOUTREG3          (0xBE001276)

#define nMSP_INTSTATREG           (0xBE001280)
#define nMSP_PORTCON              (0xBE001282)

/* ***** GPIO ***** */

#define nMSP_GPIOAALTFNLOW        (0xBE001020)
#define nMSP_GPIOBALTFNLOW        (0xBE001022)
#define nMSP_GPIOCALTFNLOW        (0xBE001024)
#define nMSP_GPIODALTFNLOW        (0xBE001026)
#define nMSP_GPIOEALTFNLOW        (0xBE001028)
#define nMSP_GPIOFALTFNLOW        (0xBE00102A)
#define nMSP_GPIOGALTFNLOW        (0xBE00102C)
#define nMSP_GPIOHALTFNLOW        (0xBE00102E)
#define nMSP_GPIOIALTFNLOW        (0xBE001030)
#define nMSP_GPIOJALTFNLOW        (0xBE001032)
#define nMSP_GPIOKALTFNLOW        (0xBE001034)
#define nMSP_GPIOLALTFNLOW        (0xBE001036)
#define nMSP_GPIOMALTFNLOW        (0xBE001038)
#define nMSP_GPIONALTFNLOW        (0xBE00103A)
#define nMSP_GPIOOALTFNLOW        (0xBE00103C)

#define nMSP_GPIOAALTFNHI         (0xBE001040)
#define nMSP_GPIOBALTFNHI         (0xBE001042)
#define nMSP_GPIOCALTFNHI         (0xBE001044)
#define nMSP_GPIODALTFNHI         (0xBE001046)
#define nMSP_GPIOEALTFNHI         (0xBE001048)
#define nMSP_GPIOFALTFNHI         (0xBE00104A)
#define nMSP_GPIOGALTFNHI         (0xBE00104C)
#define nMSP_GPIOHALTFNHI         (0xBE00104E)
#define nMSP_GPIOIALTFNHI         (0xBE001050)
#define nMSP_GPIOJALTFNHI         (0xBE001052)
#define nMSP_GPIOKALTFNHI         (0xBE001054)
#define nMSP_GPIOLALTFNHI         (0xBE001056)
#define nMSP_GPIOMALTFNHI         (0xBE001058)
#define nMSP_GPIONALTFNHI         (0xBE00105A)
#define nMSP_GPIOOALTFNHI         (0xBE00105C)

#define nMSP_GPIOAOUT             (0xBE001060)
#define nMSP_GPIOBOUT             (0xBE001062)
#define nMSP_GPIOCOUT             (0xBE001064)
#define nMSP_GPIODOUT             (0xBE001066)
#define nMSP_GPIOEOUT             (0xBE001068)
#define nMSP_GPIOFOUT             (0xBE00106A)
#define nMSP_GPIOGOUT             (0xBE00106C)
#define nMSP_GPIOHOUT             (0xBE00106E)
#define nMSP_GPIOIOUT             (0xBE001070)
#define nMSP_GPIOJOUT             (0xBE001072)
#define nMSP_GPIOKOUT             (0xBE001074)
#define nMSP_GPIOLOUT             (0xBE001076)
#define nMSP_GPIOMOUT             (0xBE001078)
#define nMSP_GPIONOUT             (0xBE00107A)
#define nMSP_GPIOOOUT             (0xBE00107C)

#define nMSP_GPIOAEVTTYPLOW       (0xBE001080)
#define nMSP_GPIOBEVTTYPLOW       (0xBE001082)
#define nMSP_GPIOCEVTTYPLOW       (0xBE001084)
#define nMSP_GPIODEVTTYPLOW       (0xBE001086)
#define nMSP_GPIOEEVTTYPLOW       (0xBE001088)
#define nMSP_GPIOFEVTTYPLOW       (0xBE00108A)
#define nMSP_GPIOGEVTTYPLOW       (0xBE00108C)
#define nMSP_GPIOHEVTTYPLOW       (0xBE00108E)
#define nMSP_GPIOIEVTTYPLOW       (0xBE001090)
#define nMSP_GPIOJEVTTYPLOW       (0xBE001092)
#define nMSP_GPIOKEVTTYPLOW       (0xBE001094)
#define nMSP_GPIOLEVTTYPLOW       (0xBE001096)
#define nMSP_GPIOMEVTTYPLOW       (0xBE001098)
#define nMSP_GPIONEVTTYPLOW       (0xBE00109A)
#define nMSP_GPIOOEVTTYPLOW       (0xBE00109C)

#define nMSP_GPIOAEVTTYPHI        (0xBE0010A0)
#define nMSP_GPIOBEVTTYPHI        (0xBE0010A2)
#define nMSP_GPIOCEVTTYPHI        (0xBE0010A4)
#define nMSP_GPIODEVTTYPHI        (0xBE0010A6)
#define nMSP_GPIOEEVTTYPHI        (0xBE0010A8)
#define nMSP_GPIOFEVTTYPHI        (0xBE0010AA)
#define nMSP_GPIOGEVTTYPHI        (0xBE0010AC)
#define nMSP_GPIOHEVTTYPHI        (0xBE0010AE)
#define nMSP_GPIOIEVTTYPHI        (0xBE0010B0)
#define nMSP_GPIOJEVTTYPHI        (0xBE0010B2)
#define nMSP_GPIOKEVTTYPHI        (0xBE0010B4)
#define nMSP_GPIOLEVTTYPHI        (0xBE0010B6)
#define nMSP_GPIOMEVTTYPHI        (0xBE0010B8)
#define nMSP_GPIONEVTTYPHI        (0xBE0010BA)
#define nMSP_GPIOOEVTTYPHI        (0xBE0010BC)

#define nMSP_GPIOAPUENB           (0xBE0010C0)
#define nMSP_GPIOBPUENB           (0xBE0010C2)
#define nMSP_GPIOCPUENB           (0xBE0010C4)
#define nMSP_GPIODPUENB           (0xBE0010C6)
#define nMSP_GPIOEPUENB           (0xBE0010C8)
#define nMSP_GPIOFPUENB           (0xBE0010CA)
#define nMSP_GPIOGPUENB           (0xBE0010CC)
#define nMSP_GPIOHPUENB           (0xBE0010CE)
#define nMSP_GPIOIPUENB           (0xBE0010D0)
#define nMSP_GPIOJPUENB           (0xBE0010D2)
#define nMSP_GPIOKPUENB           (0xBE0010D4)
#define nMSP_GPIOLPUENB           (0xBE0010D6)
#define nMSP_GPIOMPUENB           (0xBE0010D8)
#define nMSP_GPIONPUENB           (0xBE0010DA)
#define nMSP_GPIOOPUENB           (0xBE0010DC)

#define nMSP_GPIOAINTENB          (0xBE0010E0)
#define nMSP_GPIOBINTENB          (0xBE0010E2)
#define nMSP_GPIOCINTENB          (0xBE0010E4)
#define nMSP_GPIODINTENB          (0xBE0010E6)
#define nMSP_GPIOEINTENB          (0xBE0010E8)
#define nMSP_GPIOFINTENB          (0xBE0010EA)
#define nMSP_GPIOGINTENB          (0xBE0010EC)
#define nMSP_GPIOHINTENB          (0xBE0010EE)
#define nMSP_GPIOIINTENB          (0xBE0010F0)
#define nMSP_GPIOJINTENB          (0xBE0010F2)
#define nMSP_GPIOKINTENB          (0xBE0010F4)
#define nMSP_GPIOLINTENB          (0xBE0010F6)
#define nMSP_GPIOMINTENB          (0xBE0010F8)
#define nMSP_GPIONINTENB          (0xBE0010FA)
#define nMSP_GPIOOINTENB          (0xBE0010FC)

#define nMSP_GPIOAEVT             (0xBE001100)
#define nMSP_GPIOBEVT             (0xBE001102)
#define nMSP_GPIOCEVT             (0xBE001104)
#define nMSP_GPIODEVT             (0xBE001106)
#define nMSP_GPIOEEVT             (0xBE001108)
#define nMSP_GPIOFEVT             (0xBE00110A)
#define nMSP_GPIOGEVT             (0xBE00110C)
#define nMSP_GPIOHEVT             (0xBE00110E)
#define nMSP_GPIOIEVT             (0xBE001110)
#define nMSP_GPIOJEVT             (0xBE001112)
#define nMSP_GPIOKEVT             (0xBE001114)
#define nMSP_GPIOLEVT             (0xBE001116)
#define nMSP_GPIOMEVT             (0xBE001118)
#define nMSP_GPIONEVT             (0xBE00111A)
#define nMSP_GPIOOEVT             (0xBE00111C)

#define nMSP_GPIOPADSEL           (0xBE001120)

#define nMSP_GPIOGREVTSTAT        (0xBE001140)

#define nMSP_GPIOAPINLVL          (0xBE001180)
#define nMSP_GPIOBPINLVL          (0xBE001182)
#define nMSP_GPIOCPINLVL          (0xBE001184)
#define nMSP_GPIODPINLVL          (0xBE001186)
#define nMSP_GPIOEPINLVL          (0xBE001188)
#define nMSP_GPIOFPINLVL          (0xBE00118A)
#define nMSP_GPIOGPINLVL          (0xBE00118C)
#define nMSP_GPIOHPINLVL          (0xBE00118E)
#define nMSP_GPIOIPINLVL          (0xBE001190)
#define nMSP_GPIOJPINLVL          (0xBE001192)
#define nMSP_GPIOKPINLVL          (0xBE001194)
#define nMSP_GPIOLPINLVL          (0xBE001196)
#define nMSP_GPIOMPINLVL          (0xBE001198)
#define nMSP_GPIONPINLVL          (0xBE00119A)
#define nMSP_GPIOOPINLVL          (0xBE00119C)

#define nMSP_GPIOEOUT_ALIVE       (0xBE0011A0)
#define nMSP_GPIOEOUTENB_ALIVE    (0xBE0011A2)
#define nMSP_GPIOEPUENB_ALIVE     (0xBE0011A4)
#define nMSP_GPIOFOUT_ALIVE       (0xBE0011A6)
#define nMSP_GPIOFOUTENB_ALIVE    (0xBE0011A8)
#define nMSP_GPIOFPUENB_ALIVE     (0xBE0011AA)
#define nMSP_GPIOGOUT_ALIVE       (0xBE0011AC)
#define nMSP_GPIOGOUTENB_ALIVE    (0xBE0011AE)
#define nMSP_GPIOGPUENB_ALIVE     (0xBE0011B0)
#define nMSP_GPIOIOUT_ALIVE       (0xBE0011B2)
#define nMSP_GPIOIOUTENB_ALIVE    (0xBE0011B4)
#define nMSP_GPIOIPUENB_ALIVE     (0xBE0011B6)
#define nMSP_GPIOJOUT_ALIVE       (0xBE0011B8)
#define nMSP_GPIOJOUTENB_ALIVE    (0xBE0011BA)
#define nMSP_GPIOJPUENB_ALIVE     (0xBE0011BC)

/* ********************************
 * ***** Video Post Processor *****
 * ********************************
 */

/* ***** Frame Dimension Converter *****
 * 
 * All Frame Dimension Converter registers are referenced to a value called "FDC_BASE", which
 * is never actually given a value. Perhaps FDC_BASE is specified in another SFR, but it has
 * not yet been found. Open2x project kernel source gives this as C0001838.
 */

/* ***** Scale Processor ***** */

#define nMSP_SC_CMD               (0xBE001800)
#define nMSP_SC_STATUS            (0xBE001802)
#define nMSP_SC_MIRROR            (0xBE001804)
#define nMSP_SC_PRE_VRATIO        (0xBE001806)
#define nMSP_SC_PRE_HRATIO        (0xBE001808)
#define nMSP_SC_POST_VRATIOL      (0xBE00180A)
#define nMSP_SC_POST_VRATIOH      (0xBE00180C)
#define nMSP_SC_POST_HRATIOL      (0xBE00180E)
#define nMSP_SC_POST_HRATIOH      (0xBE001810)
#define nMSP_SC_SRC_ODD_ADDRL     (0xBE001812)
#define nMSP_SC_SRC_ODD_ADDRH     (0xBE001814)
#define nMSP_SC_SRC_EVEN_ADDRL    (0xBE001816)
#define nMSP_SC_SRC_EVEN_ADDRH    (0xBE001818)
#define nMSP_SC_DST_ADDRL         (0xBE00181A)
#define nMSP_SC_DST_ADDRH         (0xBE00181C)
#define nMSP_SC_SRC_PXL_HEIGHT    (0xBE00181E)
#define nMSP_SC_SRC_PXL_WIDTH     (0xBE001820)
#define nMSP_SC_SRC_PXL_REQCNT    (0xBE001822)
#define nMSP_SC_DST_PXL_HEIGHT    (0xBE001824)
#define nMSP_SC_DST_PXL_WIDTH     (0xBE001826)

/* SC_DST_WPXL_WIDTH and SC_LUMA_OFFSET share the same address. This is probably a mistake.
 * Open2x project kernel source says SC_DST_WPXL_WIDTH is at C0001832.
 */
#define nMSP_SC_DST_WPXL_WIDTH    (0xBE001828)
#define nMSP_SC_LUMA_OFFSET       (0xBE001828)
#define nMSP_SC_CB_OFFSET         (0xBE00182A)
#define nMSP_SC_CR_OFFSET         (0xBE00182C)
#define nMSP_SC_DELAY             (0xBE00182E)
#define nMSP_SC_MEM_CNTR          (0xBE001830)
#define nMSP_SC_IRQ               (0xBE001834)
#define nMSP_SC_RESERVED          (0xBE001836)

/* ***** Multi Layer Controller ***** */

#define nMSP_MLC_OVLAY_CNTR       (0xBE002880)
#define nMSP_MLC_YUV_EFFECT       (0xBE002882)
#define nMSP_MLC_YUV_CNTL         (0xBE002884)
#define nMSP_MLC_YUVA_TP_HSC      (0xBE002886)
#define nMSP_MLC_YUVA_BT_HSC      (0xBE002888)
#define nMSP_MLC_YUVA_TP_HSCL     (0xBE00288A)
#define nMSP_MLC_YUVA_TP_HSCH     (0xBE00288C)
#define nMSP_MLC_YUVA_BT_HSCL     (0xBE00288E)
#define nMSP_MLC_YUVA_BT_HSCH     (0xBE002890)
#define nMSP_MLC_YUVA_TP_PXW      (0xBE002892)
#define nMSP_MLC_YUVA_BT_PXW      (0xBE002894)
#define nMSP_MLC_YUVA_STX         (0xBE002896)
#define nMSP_MLC_YUVA_ENDX        (0xBE002898)
#define nMSP_MLC_YUVA_TP_STY      (0xBE00289A)
#define nMSP_MLC_YUVA_TP_ENDY     (0xBE00289C)
#define nMSP_MLC_YUVA_BT_ENDY     (0xBE00289E)
#define nMSP_MLC_YUVA_TP_OADRL    (0xBE0028A0)
#define nMSP_MLC_YUVA_TP_OADRH    (0xBE0028A2)
#define nMSP_MLC_YUVA_TP_EADRL    (0xBE0028A4)
#define nMSP_MLC_YUVA_TP_EADRH    (0xBE0028A6)
#define nMSP_MLC_YUVA_BT_OADRL    (0xBE0028A8)
#define nMSP_MLC_YUVA_BT_OADRH    (0xBE0028AA)
#define nMSP_MLC_YUVA_BT_EADRL    (0xBE0028AC)
#define nMSP_MLC_YUVA_BT_EADRH    (0xBE0028AE)
#define nMSP_MLC_YUVB_TP_HSC      (0xBE0028B0)
#define nMSP_MLC_YUVB_BT_HSC      (0xBE0028B2)
#define nMSP_MLC_YUVB_TP_VSCL     (0xBE0028B4)
#define nMSP_MLC_YUVB_TP_VSCH     (0xBE0028B6)
#define nMSP_MLC_YUVB_BT_VSCL     (0xBE0028B8)
#define nMSP_MLC_YUVB_BT_VSCH     (0xBE0028BA)
#define nMSP_MLC_YUVB_TP_PXW      (0xBE0028BC)
#define nMSP_MLC_YUVB_BT_PXW      (0xBE0028BE)
#define nMSP_MLC_YUVB_STX         (0xBE0028C0)
#define nMSP_MLC_YUVB_ENDX        (0xBE0028C2)
#define nMSP_MLC_YUVB_TP_STY      (0xBE0028C4)
#define nMSP_MLC_YUVB_TP_ENDY     (0xBE0028C6)
#define nMSP_MLC_YUVB_BT_ENDY     (0xBE0028C8)
#define nMSP_MLC_YUVB_TP_OADRL    (0xBE0028CA)
#define nMSP_MLC_YUVB_TP_OADRH    (0xBE0028CC)
#define nMSP_MLC_YUVB_TP_EADRL    (0xBE0028CE)
#define nMSP_MLC_YUVB_TP_EADRH    (0xBE0028D0)
#define nMSP_MLC_YUVB_BT_OADRL    (0xBE0028D2)
#define nMSP_MLC_YUVB_BT_OADRH    (0xBE0028D4)
#define nMSP_MLC_YUVB_BT_EADRL    (0xBE0028D6)
#define nMSP_MLC_YUVB_BT_EADRH    (0xBE0028D8)

/* Still Image Control */
#define nMSP_MLC_STL_CNTL         (0xBE0028DA)

/* Mix/Mux Control */
#define nMSP_MLC_STL_MIXMUX       (0xBE0028DC)

/* MLC_RGB_ALPHAL and MLC_STL_ALPHAL share the same address. Seems OK, because the open2x
 * project kernel source agrees with it
 */
#define nMSP_MLC_RGB_ALPHAL       (0xBE0028DE)
#define nMSP_MLC_STL_ALPHAL       (0xBE0028DE)

#define nMSP_MLC_STL_ALPHAH       (0xBE0028E0)

#define nMSP_MLC_STL1_STX         (0xBE0028E2)
#define nMSP_MLC_STL1_ENDX        (0xBE0028E4)
#define nMSP_MLC_STL1_STY         (0xBE0028E6)
#define nMSP_MLC_STL1_ENDY        (0xBE0028E8)

#define nMSP_MLC_STL2_STX         (0xBE0028EA)
#define nMSP_MLC_STL2_ENDX        (0xBE0028EC)
#define nMSP_MLC_STL2_STY         (0xBE0028EE)
#define nMSP_MLC_STL2_ENDY        (0xBE0028F0)

#define nMSP_MLC_STL3_STX         (0xBE0028F2)
#define nMSP_MLC_STL3_ENDX        (0xBE0028F4)
#define nMSP_MLC_STL3_STY         (0xBE0028F6)
#define nMSP_MLC_STL3_ENDY        (0xBE0028F8)

#define nMSP_MLC_STL4_STX         (0xBE0028FA)
#define nMSP_MLC_STL4_ENDX        (0xBE0028FC)
#define nMSP_MLC_STL4_STY         (0xBE0028FE)
#define nMSP_MLC_STL4_ENDY        (0xBE002900)

#define nMSP_MLC_STL_CKEY_GB      (0xBE002902)
#define nMSP_MLC_STL_CKEY_R       (0xBE002904)

#define nMSP_MLC_STL_HSC          (0xBE002906)
#define nMSP_MLC_STL_VSCL         (0xBE002908)
#define nMSP_MLC_STL_VSCH         (0xBE00290A)
#define nMSP_MLC_STL_HW           (0xBE00290C)

#define nMSP_MLC_STL_OADRL        (0xBE00290E)
#define nMSP_MLC_STL_OADRH        (0xBE002910)
#define nMSP_MLC_STL_EADRL        (0xBE002912)
#define nMSP_MLC_STL_EADRH        (0xBE002914)

#define nMSP_MLC_OSD_OADRL        (0xBE002916)
#define nMSP_MLC_OSD_OADRH        (0xBE002918)
#define nMSP_MLC_OSD_EADRL        (0xBE00291A)
#define nMSP_MLC_OSD_EADRH        (0xBE00291C)

#define nMSP_MLC_HWC_CNTR         (0xBE00291E)
#define nMSP_MLC_HWC_STX          (0xBE002920)
#define nMSP_MLC_HWC_STY          (0xBE002922)

#define nMSP_MLC_HWC_FGR          (0xBE002924)
#define nMSP_MLC_HWC_FB           (0xBE002926)
#define nMSP_MLC_HWC_BGR          (0xBE002928)
#define nMSP_MLC_HWC_BB           (0xBE00292A)

#define nMSP_MLC_HWC_OADRL        (0xBE00292C)
#define nMSP_MLC_HWC_OADRH        (0xBE00292E)
#define nMSP_MLC_HWC_EADRL        (0xBE002930)
#define nMSP_MLC_HWC_EADRH        (0xBE002932)

#define nMSP_MLC_LUMA_ENH         (0xBE002934)

#define nMSP_MLC_HUECB1AB         (0xBE002936)
#define nMSP_MLC_HUECR1AB         (0xBE002938)
#define nMSP_MLC_HUECB2AB         (0xBE00293A)
#define nMSP_MLC_HUECR2AB         (0xBE00293C)
#define nMSP_MLC_HUECB3AB         (0xBE00293E)
#define nMSP_MLC_HUECR3AB         (0xBE002940)
#define nMSP_MLC_HUECB4AB         (0xBE002942)
#define nMSP_MLC_HUECR4AB         (0xBE002944)

#define nMSP_MLC_DITHER           (0xBE002946)

#define nMSP_MLC_OSD_PALLT_A      (0xBE002954)
#define nMSP_MLC_OSD_PALLT_D      (0xBE002956)

#define nMSP_MLC_STL_PALLT_A      (0xBE002958)
#define nMSP_MLC_STL_PALLT_D      (0xBE00295A)

#define nMSP_MLC_GAMMA_A          (0xBE00295C)
#define nMSP_MLC_GAMMA_D          (0xBE00295E)

#define nMSP_MLC_SPU_CTRL         (0xBE002974)
#define nMSP_MLC_SPU_DELAY        (0xBE002976)

#define nMSP_MLC_SPU_BASEI_ADDRL  (0xBE002978)
#define nMSP_MLC_SPU_BASEI_ADDRH  (0xBE00297A)
#define nMSP_MLC_SPU_HLI_ADDRL    (0xBE00297C)
#define nMSP_MLC_SPU_HLI_ADDRH    (0xBE00297E)

/* MLC_SPU_PAL is the start of a 64-byte area for storing 16 RGB triples */
#define nMSP_MLC_SPU_PAL          (0xBE002980)

/* ***** Display Controller ***** */

#define nMSP_DPC_CNTL             (0xBE002800)
#define nMSP_DPC_FPICNTL          (0xBE002802)
#define nMSP_DPC_FPIPOL1          (0xBE002804)
#define nMSP_DPC_FPIPOL2          (0xBE002806)
#define nMSP_DPC_FPIPOL3          (0xBE002808)
#define nMSP_DPC_FPIATV1          (0xBE00280A)
#define nMSP_DPC_FPIATV2          (0xBE00280C)
#define nMSP_DPC_FPIATV3          (0xBE00280E)
#define nMSP_DPC_X_MAX            (0xBE002816)
#define nMSP_DPC_Y_MAX            (0xBE002818)
#define nMSP_DPC_HS_WIDTH         (0xBE00281A)
#define nMSP_DPC_HS_STR           (0xBE00281C)
#define nMSP_DPC_HS_END           (0xBE00281E)
#define nMSP_DPC_V_SYNC           (0xBE002820)
#define nMSP_DPC_V_END            (0xBE002822)
#define nMSP_DPC_TV_BNK           (0xBE002824)
#define nMSP_DPC_DE               (0xBE002826)
#define nMSP_DPC_PS               (0xBE002828)
#define nMSP_DPC_FG               (0xBE00282A)
#define nMSP_DPC_LP               (0xBE00282C)
#define nMSP_DPC_CLKVH            (0xBE00282E)
#define nMSP_DPC_CLKVL            (0xBE002830)
#define nMSP_DPC_POL              (0xBE002832)
#define nMSP_DPC_CISSYNC          (0xBE002834)
#define nMSP_DPC_MID_SYNC         (0xBE002836)
#define nMSP_DPC_C_SYNC           (0xBE002838)
#define nMSP_DPC_Y_BLANK          (0xBE00283A)
#define nMSP_DPC_C_BLANK          (0xBE00283C)
#define nMSP_DPC_YP_CSYNC         (0xBE00283E)
#define nMSP_DPC_YN_CSYNC         (0xBE002840)
#define nMSP_DPC_CP_CSYNC         (0xBE002842)
#define nMSP_DPC_CN_CSYNC         (0xBE002844)
#define nMSP_DPC_INTR             (0xBE002846)
#define nMSP_DPC_CLKCNTL          (0xBE002848)

/* ***** 2D Graphic Accelerator ***** */

#define nMSP_MESGDSTCTRL          (0xDE020000)
#define nMSP_MESGDSTADDR          (0xDE020004)
#define nMSP_MESGDSTSTRIDE        (0xDE020008)

#define nMSP_MESGSRCCTRL          (0xDE02000C)
#define nMSP_MESGSRCADDR          (0xDE020010)
#define nMSP_MESGSRCSTRIDE        (0xDE020014)

#define nMSP_MESGSRCFORCOLOR      (0xDE020018)
#define nMSP_MESGSRCBACKCOLOR     (0xDE02001C)

#define nMSP_MESGPATCTRL          (0xDE020020)

#define nMSP_MESGFORCOLOR         (0xDE020024)
#define nMSP_MESGBACKCOLOR        (0xDE020028)
#define nMSP_MESGSIZE             (0xDE02002C)

#define nMSP_MESGCTRL             (0xDE020030)
#define nMSP_MESGSTATUS           (0xDE020034)
#define nMSP_MESGFIFOSTATUS       (0xDE020038)
#define nMSP_MESGFIFO             (0xDE02003C)

/* MESGPAT is the first word in a 32-word table of pattern data */
#define nMSP_MESGPAT              (0xDE020080)

/* ***** Image Signal Processor ***** */

#define nMSP_ISP_CON_1            (0xBE003000)
#define nMSP_ISP_CON_2            (0xBE003002)
#define nMSP_ISP_INT_CON          (0xBE003004)
#define nMSP_ISP_VPERIOD          (0xBE003006)
#define nMSP_ISP_VPW              (0xBE003008)
#define nMSP_ISP_HPERIOD          (0xBE00300A)
#define nMSP_ISP_HPW              (0xBE00300C)
#define nMSP_ISP_HACT_ST          (0xBE00300E)
#define nMSP_ISP_H_WIDTH          (0xBE003010)
#define nMSP_ISP_VACT_ST          (0xBE003012)
#define nMSP_ISP_V_HEIGHT         (0xBE003014)
#define nMSP_ISP_SRC_HWIDTH       (0xBE003016)
#define nMSP_ISP_RGBSV            (0xBE003018)
#define nMSP_ISP_RGBEV            (0xBE00301A)
#define nMSP_ISP_RGBSH            (0xBE00301C)
#define nMSP_ISP_RGBEH            (0xBE00301E)
#define nMSP_ISP_AESV             (0xBE003020)
#define nMSP_ISP_AEEV             (0xBE003022)
#define nMSP_ISP_AESH             (0xBE003024)
#define nMSP_ISP_AEEH             (0xBE003026)
#define nMSP_ISP_AELIM            (0xBE003028)
#define nMSP_ISP_AEVSTEP          (0xBE00302A)
#define nMSP_ISP_AEHSTEP          (0xBE00302C)
#define nMSP_ISP_AFSV             (0xBE00302E)
#define nMSP_ISP_AFEV             (0xBE003030)
#define nMSP_ISP_AFSH             (0xBE003032)
#define nMSP_ISP_AFEH             (0xBE003034)
#define nMSP_ISP_CNTRSTGAIN       (0xBE003036)
#define nMSP_ISP_CNTRSTLIM        (0xBE003038)
#define nMSP_ISP_RGBREC           (0xBE00303A)
#define nMSP_ISP_HUECB0AB         (0xBE003040)
#define nMSP_ISP_HUECR0AB         (0xBE003042)
#define nMSP_ISP_HUECB1AB         (0xBE003044)
#define nMSP_ISP_HUECR1AB         (0xBE003046)
#define nMSP_ISP_HUECB2AB         (0xBE003048)
#define nMSP_ISP_HUECR2AB         (0xBE00304A)
#define nMSP_ISP_HUECB3AB         (0xBE00304C)
#define nMSP_ISP_HUECR3AB         (0xBE00304E)
#define nMSP_ISP_YGAIN            (0xBE003054)
#define nMSP_ISP_CDLY             (0xBE003056)
#define nMSP_ISP_BRIGHT           (0xBE003058)
#define nMSP_ISP_YCLIP            (0xBE00305A)
#define nMSP_ISP_YCLIPSLANT       (0xBE00305C)
#define nMSP_ISP_CSUPSLANT        (0xBE00305E)
#define nMSP_ISP_SLICELMT         (0xBE003060)
#define nMSP_ISP_WBLMT            (0xBE003062)

/* The following SFRs are defined in the datasheet as having the upper 16 bits at a lower address
 * than the lower 16 bits, which is opposite to all the other double-hword registers in the chip.
 * Weird.
 */
#define nMSP_ISP_WBACCCNTH        (0xBE003064)
#define nMSP_ISP_WBACCCNTL        (0xBE003066)
#define nMSP_ISP_WBCBACCH         (0xBE003068)
#define nMSP_ISP_WBCBACCL         (0xBE00306A)
#define nMSP_ISP_WBCRACCH         (0xBE00306C)
#define nMSP_ISP_WBCRACCL         (0xBE00306E)
#define nMSP_ISP_AFACCH           (0xBE003070)
#define nMSP_ISP_AFACCL           (0xBE003072)

#define nMSP_ISP_WBCOORD0         (0xBE003076)
#define nMSP_ISP_WBCOORD1         (0xBE003078)
#define nMSP_ISP_WBCOORD2         (0xBE00307A)
#define nMSP_ISP_WBCOORD3         (0xBE00307C)
#define nMSP_ISP_WBCOORD4         (0xBE00307E)
#define nMSP_ISP_WBCOORD5         (0xBE003080)
#define nMSP_ISP_WBCOORD6         (0xBE003082)
#define nMSP_ISP_WBCOORD7         (0xBE003084)
#define nMSP_ISP_WBCOORD8         (0xBE003086)
#define nMSP_ISP_WBCOORD9         (0xBE003088)
#define nMSP_ISP_WBCOORD10        (0xBE00308A)
#define nMSP_ISP_WBCOORD11        (0xBE00308C)
#define nMSP_ISP_WBCOORD12        (0xBE00308E)
#define nMSP_ISP_WBCOORD13        (0xBE003090)
#define nMSP_ISP_WBCOORD14        (0xBE003092)
#define nMSP_ISP_WBCOORD15        (0xBE003094)

#define nMSP_ISP_MEMADDRL1        (0xBE003096)
#define nMSP_ISP_MEMADDRH1        (0xBE003098)

#define nMSP_ISP_GAMMA            (0xBE00309A)
#define nMSP_ISP_GAMMD            (0xBE00309C)

/* There's something strange going on here... the datasheet implies that the Auto Exposure Result
 * takes up 16 words which will explain the address difference between AEACCH and AEACCL, but
 * the explanation is sketchy.
 */
#define nMSP_ISP_AEACCH           (0xBE00309E)
#define nMSP_ISP_AEACCL           (0xBE0030DC)

#define nMSP_ISP_CNTST_LLCNTL     (0xBE0030DE)
#define nMSP_ISP_CNTST_LLCNTH     (0xBE0030E0)
#define nMSP_ISP_CNTST_HLCNTL     (0xBE0030E2)
#define nMSP_ISP_CNTST_HLCNTH     (0xBE0030E4)

#define nMSP_ISP_CNTST_LLACCL     (0xBE0030E6)
#define nMSP_ISP_CNTST_LLACCH     (0xBE0030E8)
#define nMSP_ISP_CNTST_HLACCL     (0xBE0030EA)
#define nMSP_ISP_CNTST_HLACCH     (0xBE0030EC)

/* MEMRGBADDR and MEMADDR2 share the same addresses, which might be alright, but it is suspicious... */
#define nMSP_ISP_MEMRGBADDRL      (0xBE0030EE)
#define nMSP_ISP_MEMRGBADDRH      (0xBE0030F0)
#define nMSP_ISP_MEMADDRL2        (0xBE0030EE)
#define nMSP_ISP_MEMADDRH2        (0xBE0030F0)

#define nMSP_ISP_DST_HWIDTH       (0xBE0030F2)

/* ***** Pulse Width Modulator ***** */

#define nMSP_PWM_PRES01           (0xBE000B00)
#define nMSP_PWM_DUTY0            (0xBE000B02)
#define nMSP_PWM_DUTY1            (0xBE000B04)
#define nMSP_PWM_PERIOD0          (0xBE000B06)
#define nMSP_PWM_PERIOD1          (0xBE000B08)
#define nMSP_PWM_PRES23           (0xBE000B10)
#define nMSP_PWM_DUTY2            (0xBE000B12)
#define nMSP_PWM_DUTY3            (0xBE000B14)
#define nMSP_PWM_PERIOD2          (0xBE000B16)
#define nMSP_PWM_PERIOD3          (0xBE000B18)

/* ***** Pulse Period Measurement ***** */

#define nMSP_PPMCTRL              (0xBE003E00)
#define nMSP_PPMSTATUS            (0xBE003E04)
#define nMSP_PPMLOWPERIOD         (0xBE003E06)
#define nMSP_PPMHIGHPERIOD        (0xBE003E08)

/* ***** USB Host ***** */

/* USB Host is standard OHCI, and that means its control registers are in RAM. Somewhere. */

/* ***** USB Device ****** */

#define nMSP_FUNC_ADDR_REG        (0xBE001400)
#define nMSP_PWR_REG              (0xBE001402)
#define nMSP_EP_INT_REG           (0xBE001404)
#define nMSP_USB_INT_REG          (0xBE00140C)
#define nMSP_EP_INT_EN_REG        (0xBE00140E)
#define nMSP_USB_INT_EN_REG       (0xBE001416)
#define nMSP_FRAME_NUM1_REG       (0xBE001418)
#define nMSP_FRAME_NUM2_REG       (0xBE00141A)
#define nMSP_INDEX_REG            (0xBE00141C)

#define nMSP_MAXP_REG             (0xBE001420)
#define nMSP_EP0_CSR              (0xBE001422)
#define nMSP_IN_CSR1_REG          (0xBE001422)
#define nMSP_IN_CSR2_REG          (0xBE001424)
#define nMSP_OUT_MAXP_REG         (0xBE001426)
#define nMSP_OUT_CSR1_REG         (0xBE001428)
#define nMSP_OUT_CSR2_REG         (0xBE00142A)
#define nMSP_OUT_FIFO_CNT1_REG    (0xBE00142C)
#define nMSP_OUT_FIFO_CNT2_REG    (0xBE00142E)

#define nMSP_EP0_FIFO_REG         (0xBE001440)
#define nMSP_EP1_FIFO_REG         (0xBE001442)
#define nMSP_EP2_FIFO_REG         (0xBE001444)
#define nMSP_EP3_FIFO_REG         (0xBE001446)
#define nMSP_EP4_FIFO_REG         (0xBE001448)

#define nMSP_EP1_DMA_CON          (0xBE001460)
#define nMSP_EP1_DMA_FIFO         (0xBE001464)
#define nMSP_EP1_DMA_TTC_L        (0xBE001466)
#define nMSP_EP1_DMA_TTC_M        (0xBE001468)
#define nMSP_EP1_DMA_TTC_H        (0xBE00146A)

#define nMSP_EP2_DMA_CON          (0xBE00146C)
#define nMSP_EP2_DMA_FIFO         (0xBE001470)
#define nMSP_EP2_DMA_TTC_L        (0xBE001472)
#define nMSP_EP2_DMA_TTC_M        (0xBE001474)
#define nMSP_EP2_DMA_TTC_H        (0xBE001476)

#define nMSP_EP3_DMA_CON          (0xBE001480)
#define nMSP_EP3_DMA_FIFO         (0xBE001484)
#define nMSP_EP3_DMA_TTC_L        (0xBE001486)
#define nMSP_EP3_DMA_TTC_M        (0xBE001488)
#define nMSP_EP3_DMA_TTC_H        (0xBE00148A)

#define nMSP_EP4_DMA_CON          (0xBE00148C)
#define nMSP_EP4_DMA_FIFO         (0xBE001490)
#define nMSP_EP4_DMA_TTC_L        (0xBE001492)
#define nMSP_EP4_DMA_TTC_M        (0xBE001494)
#define nMSP_EP4_DMA_TTC_H        (0xBE001496)

/* ***** I2C Bus ***** */

#define nMSP_I2CCON               (0xBE000D00)
#define nMSP_I2CSTAT              (0xBE000D02)
#define nMSP_I2CADD               (0xBE000D04)
#define nMSP_I2CDS                (0xBE000D06)

/* ***** S/PDIF Controller ***** */

#define nMSP_SPDIFCON             (0xBE004200)
#define nMSP_SPDIFINTENB          (0xBE004202)
#define nMSP_SPDIFSTATUS          (0xBE004204)
#define nMSP_SPDIFTXDIV           (0xBE004206)

/* ***** AC'97 ***** */

#define nMSP_AC_CTRL_REG          (0xBE000E00)
#define nMSP_AC_CONFIG_REG        (0xBE000E02)
#define nMSP_AC_STA_ENA_REG       (0xBE000E04)
#define nMSP_AC_STATUS_REG        (0xBE000E06)
#define nMSP_AC_STATE_MACHINE     (0xBE000E08)
#define nMSP_AC_CODEC_ADDR        (0xBE000E0C)
#define nMSP_AC_CODEC_DATA        (0xBE000E0E)
#define nMSP_AC_CODEC_ACCESS      (0xBE000E10)

/* AC'97 Codec Registers are mapped to 0xC0000F00 - 0xC0000F7E */
#define nMSP_AC_CODEC_REG_BASE    (0xBE000F00)

/* ***** I2S ***** */

#define nMSP_I2S_CTRL_REG         (0xBE004400)
#define nMSP_I2S_CONFIG_REG       (0xBE004402)
#define nMSP_I2S_STA_ENA_REG      (0xBE004404)
#define nMSP_I2S_STATUS_REG       (0xBE004406)

/* ***** SSP/SPI ***** */

#define nMSP_SSPSPICONT0          (0xBE003F00)
#define nMSP_SSPSPICONT1          (0xBE003F02)
#define nMSP_SSPSPIDATA           (0xBE003F04)
#define nMSP_SSPSPISTAT           (0xBE003F06)

/* ***** Transport Stream Interface ***** */

#define nMSP_MPEGIFCONT           (0xBE003C00)

/* ***** ADC & Touch Panel Controller ***** */

/* All touch panel and ADC registers are referenced to a value called "TP_BASE", which
 * is never actually given a value. Open2x project kernel source says it's C0004600.
 */

/* ***** SD/MMC Controller ***** */

#define nMSP_SDICON               (0xBE001500)
#define nMSP_SDIPRE               (0xBE001502)
#define nMSP_SDICMDARGL           (0xBE001504)
#define nMSP_SDICMDARGH           (0xBE001506)
#define nMSP_SDICMDCON            (0xBE001508)
#define nMSP_SDICMDSTA            (0xBE00150A)
#define nMSP_SDIRSP0              (0xBE00150C)
#define nMSP_SDIRSP1              (0xBE00150E)
#define nMSP_SDIRSP2              (0xBE001510)
#define nMSP_SDIRSP3              (0xBE001512)
#define nMSP_SDIRSP4              (0xBE001514)
#define nMSP_SDIRSP5              (0xBE001516)
#define nMSP_SDIRSP6              (0xBE001518)
#define nMSP_SDIRSP7              (0xBE00151A)
#define nMSP_SDIBSIZE             (0xBE00151E)
#define nMSP_SDIDATCONL           (0xBE001520)
#define nMSP_SDIDATCONH           (0xBE001522)
#define nMSP_SDIDATCNTL           (0xBE001524)
#define nMSP_SDIDATCNTH           (0xBE001526)
#define nMSP_SDIDATSTA            (0xBE001528)
#define nMSP_SDIFSTA              (0xBE00152A)
#define nMSP_SDIDAT               (0xBE00152C)
#define nMSP_SDIINTMSK0           (0xBE001530)
#define nMSP_SDIINTMSK1           (0xBE001532)
#define nMSP_SDIDTIMERL           (0xBE001536)
#define nMSP_SDIDTIMERH           (0xBE001538)
#define nMSP_SDISFTREGL           (0xBE00153A)
#define nMSP_SDISFTREGH           (0xBE00153C)

/* ***** Memory Stick Controller ***** */

#define nMSP_MSCTRL               (0xBE001600)
#define nMSP_MSFIFO               (0xBE001602)
#define nMSP_MSFIFOST             (0xBE001604)
#define nMSP_MSINTCTRL            (0xBE001606)
#define nMSP_MSCRC                (0xBE001608)
#define nMSP_MSTIMEOUT            (0xBE00160A)
#define nMSP_MSTPC                (0xBE00160C)
#define nMSP_MSTPCCNT             (0xBE00160E)

/* ***** CDROM Decoder ***** */

/* Addresses of registers are not specified in the datasheet. */

/* ***** 1-Wire Master ***** */

#define nMSP_OWM_CMD              (0xBE004700)
#define nMSP_OWM_BUF              (0xBE004702)
#define nMSP_OWM_STA              (0xBE004704)
#define nMSP_OWM_STA_ENB          (0xBE004706)
#define nMSP_OWM_CLK_DIVISOR      (0xBE004708)
#define nMSP_OWM_CTRL             (0xBE00470A)
#define nMSP_OWM_IRQ              (0xBE00470C)

/* ***** DeCSS ***** */

#define nMSP_DECSSCTRL            (0xBE001700)
#define nMSP_DECSSKEY0            (0xBE001702)
#define nMSP_DECSSKEY1            (0xBE001704)
#define nMSP_DECSSKEY2            (0xBE001706)
#define nMSP_DECSSDATA            (0xBE001708)
#define nMSP_DECSSNUM             (0xBE00170A)

/* Below here are the pointer definitions. The comments are not repeated. */

#define pMSP_MEMCFGW              ((volatile unsigned short *)nMSP_MEMCFGW)
#define pMSP_MEMTIMEW0            ((volatile unsigned short *)nMSP_MEMTIMEW0)
#define pMSP_MEMTIMEW1            ((volatile unsigned short *)nMSP_MEMTIMEW1)
#define pMSP_MRMTIMEW2            ((volatile unsigned short *)nMSP_MRMTIMEW2)
#define pMSP_MEMTIMEW3            ((volatile unsigned short *)nMSP_MEMTIMEW3)
#define pMSP_MEMTIMEW4            ((volatile unsigned short *)nMSP_MEMTIMEW4)
#define pMSP_MEMWAITCTRLW         ((volatile unsigned short *)nMSP_MEMWAITCTRLW)
#define pMSP_MEMPAGEW             ((volatile unsigned short *)nMSP_MEMPAGEW)
#define pMSP_MEMIDETIMEW          ((volatile unsigned short *)nMSP_MEMIDETIMEW)
#define pMSP_MEMPCMCIAMW          ((volatile unsigned short *)nMSP_MEMPCMCIAMW)
#define pMSP_MEMPCMCIAAW          ((volatile unsigned short *)nMSP_MEMPCMCIAAW)
#define pMSP_MEMPCMCIAIW          ((volatile unsigned short *)nMSP_MEMPCMCIAIW)
#define pMSP_MEMPCMCIAWAITW       ((volatile unsigned short *)nMSP_MEMPCMCIAWAITW)
#define pMSP_MEMIDEWAITW          ((volatile unsigned short *)nMSP_MEMIDEWAITW)
#define pMSP_MEMDTIMEOUTW         ((volatile unsigned short *)nMSP_MEMDTIMEOUTW)
#define pMSP_MEMDMACTRLW          ((volatile unsigned short *)nMSP_MEMDMACTRLW)
#define pMSP_MEMDMAPOLW           ((volatile unsigned short *)nMSP_MEMDMAPOLW)
#define pMSP_MEMDMATIMEW0         ((volatile unsigned short *)nMSP_MEMDMATIMEW0)
#define pMSP_MEMDMATIMEW1         ((volatile unsigned short *)nMSP_MEMDMATIMEW1)
#define pMSP_MEMDMATIMEW2         ((volatile unsigned short *)nMSP_MEMDMATIMEW2)
#define pMSP_MEMDMATIMEW3         ((volatile unsigned short *)nMSP_MEMDMATIMEW3)
#define pMSP_MEMDMATIMEW4         ((volatile unsigned short *)nMSP_MEMDMATIMEW4)
#define pMSP_MEMDMATIMEW5         ((volatile unsigned short *)nMSP_MEMDMATIMEW5)
#define pMSP_MEMDMATIMEW6         ((volatile unsigned short *)nMSP_MEMDMATIMEW6)
#define pMSP_MEMDMATIMEW7         ((volatile unsigned short *)nMSP_MEMDMATIMEW7)
#define pMSP_MEMDMATIMEW8         ((volatile unsigned short *)nMSP_MEMDMATIMEW8)
#define pMSP_MEMDMASTRBW          ((volatile unsigned short *)nMSP_MEMDMASTRBW)
#define pMSP_MEMNANDCTRLW         ((volatile unsigned short *)nMSP_MEMNANDCTRLW)
#define pMSP_MEMNANDTIMEW         ((volatile unsigned short *)nMSP_MEMNANDTIMEW)
#define pMSP_MEMNANDECC0W         ((volatile unsigned short *)nMSP_MEMNANDECC0W)
#define pMSP_MEMNANDECC1W         ((volatile unsigned short *)nMSP_MEMNANDECC1W)
#define pMSP_MEMNANDECC2W         ((volatile unsigned short *)nMSP_MEMNANDECC2W)
#define pMSP_MEMNANDCNTW          ((volatile unsigned short *)nMSP_MEMNANDCNTW)
#define pMSP_NFDATA               ((volatile unsigned short *)nMSP_NFDATA)
#define pMSP_NFCMD                ((volatile unsigned short *)nMSP_NFCMD)
#define pMSP_NFADDR               ((volatile unsigned short *)nMSP_NFADDR)
#define pMSP_MEMCFGX              ((volatile unsigned short *)nMSP_MEMCFGX)
#define pMSP_MEMTIMEX0            ((volatile unsigned short *)nMSP_MEMTIMEX0)
#define pMSP_MEMTIMEX1            ((volatile unsigned short *)nMSP_MEMTIMEX1)
#define pMSP_MEMACTPWDX           ((volatile unsigned short *)nMSP_MEMACTPWDX)
#define pMSP_MEMREFX              ((volatile unsigned short *)nMSP_MEMREFX)
#define pMSP_DUAL920DATA0         ((volatile unsigned short *)nMSP_DUAL920DATA0)
#define pMSP_DUAL920DATA1         ((volatile unsigned short *)nMSP_DUAL920DATA1)
#define pMSP_DUAL920DATA2         ((volatile unsigned short *)nMSP_DUAL920DATA2)
#define pMSP_DUAL920DATA3         ((volatile unsigned short *)nMSP_DUAL920DATA3)
#define pMSP_DUAL920DATA4         ((volatile unsigned short *)nMSP_DUAL920DATA4)
#define pMSP_DUAL920DATA5         ((volatile unsigned short *)nMSP_DUAL920DATA5)
#define pMSP_DUAL920DATA6         ((volatile unsigned short *)nMSP_DUAL920DATA6)
#define pMSP_DUAL920DATA7         ((volatile unsigned short *)nMSP_DUAL920DATA7)
#define pMSP_DUAL920DATA8         ((volatile unsigned short *)nMSP_DUAL920DATA8)
#define pMSP_DUAL920DATA9         ((volatile unsigned short *)nMSP_DUAL920DATA9)
#define pMSP_DUAL920DATA10        ((volatile unsigned short *)nMSP_DUAL920DATA10)
#define pMSP_DUAL920DATA11        ((volatile unsigned short *)nMSP_DUAL920DATA11)
#define pMSP_DUAL920DATA12        ((volatile unsigned short *)nMSP_DUAL920DATA12)
#define pMSP_DUAL920DATA13        ((volatile unsigned short *)nMSP_DUAL920DATA13)
#define pMSP_DUAL920DATA14        ((volatile unsigned short *)nMSP_DUAL920DATA14)
#define pMSP_DUAL920DATA15        ((volatile unsigned short *)nMSP_DUAL920DATA15)
#define pMSP_DUAL940DATA0         ((volatile unsigned short *)nMSP_DUAL940DATA0)
#define pMSP_DUAL940DATA1         ((volatile unsigned short *)nMSP_DUAL940DATA1)
#define pMSP_DUAL940DATA2         ((volatile unsigned short *)nMSP_DUAL940DATA2)
#define pMSP_DUAL940DATA3         ((volatile unsigned short *)nMSP_DUAL940DATA3)
#define pMSP_DUAL940DATA4         ((volatile unsigned short *)nMSP_DUAL940DATA4)
#define pMSP_DUAL940DATA5         ((volatile unsigned short *)nMSP_DUAL940DATA5)
#define pMSP_DUAL940DATA6         ((volatile unsigned short *)nMSP_DUAL940DATA6)
#define pMSP_DUAL940DATA7         ((volatile unsigned short *)nMSP_DUAL940DATA7)
#define pMSP_DUAL940DATA8         ((volatile unsigned short *)nMSP_DUAL940DATA8)
#define pMSP_DUAL940DATA9         ((volatile unsigned short *)nMSP_DUAL940DATA9)
#define pMSP_DUAL940DATA10        ((volatile unsigned short *)nMSP_DUAL940DATA10)
#define pMSP_DUAL940DATA11        ((volatile unsigned short *)nMSP_DUAL940DATA11)
#define pMSP_DUAL940DATA12        ((volatile unsigned short *)nMSP_DUAL940DATA12)
#define pMSP_DUAL940DATA13        ((volatile unsigned short *)nMSP_DUAL940DATA13)
#define pMSP_DUAL940DATA14        ((volatile unsigned short *)nMSP_DUAL940DATA14)
#define pMSP_DUAL940DATA15        ((volatile unsigned short *)nMSP_DUAL940DATA15)
#define pMSP_DUALINT920           ((volatile unsigned short *)nMSP_DUALINT920)
#define pMSP_DUALINT940           ((volatile unsigned short *)nMSP_DUALINT940)
#define pMSP_DUALPEND920          ((volatile unsigned short *)nMSP_DUALPEND920)
#define pMSP_DUALPEND940          ((volatile unsigned short *)nMSP_DUALPEND940)
#define pMSP_DUALCTRL940          ((volatile unsigned short *)nMSP_DUALCTRL940)
#define pMSP_PWMODEREG            ((volatile unsigned short *)nMSP_PWMODEREG)
#define pMSP_CLKCHGSTREG          ((volatile unsigned short *)nMSP_CLKCHGSTREG)
#define pMSP_SYSCLKENREG          ((volatile unsigned short *)nMSP_SYSCLKENREG)
#define pMSP_COMCLKENREG          ((volatile unsigned short *)nMSP_COMCLKENREG)
#define pMSP_VGCLKENREG           ((volatile unsigned short *)nMSP_VGCLKENREG)
#define pMSP_ASCLKENREG           ((volatile unsigned short *)nMSP_ASCLKENREG)
#define pMSP_FPLLSETVREG          ((volatile unsigned short *)nMSP_FPLLSETVREG)
#define pMSP_FPLLVSETREG          ((volatile unsigned short *)nMSP_FPLLVSETREG)
#define pMSP_UPLLSETVREG          ((volatile unsigned short *)nMSP_UPLLSETVREG)
#define pMSP_UPLLVSETREG          ((volatile unsigned short *)nMSP_UPLLVSETREG)
#define pMSP_APLLSETVREG          ((volatile unsigned short *)nMSP_APLLSETVREG)
#define pMSP_APLLVSETREG          ((volatile unsigned short *)nMSP_APLLVSETREG)
#define pMSP_SYSCSETREG           ((volatile unsigned short *)nMSP_SYSCSETREG)
#define pMSP_ESYSCSETREG          ((volatile unsigned short *)nMSP_ESYSCSETREG)
#define pMSP_UIRMCSETREG          ((volatile unsigned short *)nMSP_UIRMCSETREG)
#define pMSP_AUDICSETREG          ((volatile unsigned short *)nMSP_AUDICSETREG)
#define pMSP_DISPCSETREG          ((volatile unsigned short *)nMSP_DISPCSETREG)
#define pMSP_IMAGCSETREG          ((volatile unsigned short *)nMSP_IMAGCSETREG)
#define pMSP_URT0CSETREG          ((volatile unsigned short *)nMSP_URT0CSETREG)
#define pMSP_URT1CSETREG          ((volatile unsigned short *)nMSP_URT1CSETREG)
#define pMSP_A940TIDLE            ((volatile unsigned short *)nMSP_A940TIDLE)
#define pMSP_SPDICSETREG          ((volatile unsigned short *)nMSP_SPDICSETREG)
#define pMSP_RTCTSETREG           ((volatile unsigned int *)nMSP_RTCTSETREG)
#define pMSP_RTCTCNTREG           ((volatile unsigned int *)nMSP_RTCTCNTREG)
#define pMSP_RTCSTCNTREG          ((volatile unsigned short *)nMSP_RTCSTCNTREG)
#define pMSP_TICKSETREG           ((volatile unsigned short *)nMSP_TICKSETREG)
#define pMSP_ALARMTREG            ((volatile unsigned short *)nMSP_ALARMTREG)
#define pMSP_PWRMGRREG            ((volatile unsigned short *)nMSP_PWRMGRREG)
#define pMSP_CLKMGRREG            ((volatile unsigned short *)nMSP_CLKMGRREG)
#define pMSP_RSTCTRLREG           ((volatile unsigned short *)nMSP_RSTCTRLREG)
#define pMSP_RSTSTREG             ((volatile unsigned short *)nMSP_RSTSTREG)
#define pMSP_BOOTCTRLREG          ((volatile unsigned short *)nMSP_BOOTCTRLREG)
#define pMSP_LOCKTIMEREG          ((volatile unsigned short *)nMSP_LOCKTIMEREG)
#define pMSP_RSTTIMEREG           ((volatile unsigned short *)nMSP_RSTTIMEREG)
#define pMSP_EXTCTRLREG           ((volatile unsigned short *)nMSP_EXTCTRLREG)
#define pMSP_STOPTSETREG          ((volatile unsigned short *)nMSP_STOPTSETREG)
#define pMSP_RTCCTRLREG           ((volatile unsigned short *)nMSP_RTCCTRLREG)
#define pMSP_RTCSTREG             ((volatile unsigned short *)nMSP_RTCSTREG)
#define pMSP_DMAINT               ((volatile unsigned short *)nMSP_DMAINT)
#define pMSP_DCH0SRM              ((volatile unsigned short *)nMSP_DCH0SRM)
#define pMSP_DCH0TRM              ((volatile unsigned short *)nMSP_DCH0TRM)
#define pMSP_DCH1SRM              ((volatile unsigned short *)nMSP_DCH1SRM)
#define pMSP_DCH1TRM              ((volatile unsigned short *)nMSP_DCH1TRM)
#define pMSP_DCH2SRM              ((volatile unsigned short *)nMSP_DCH2SRM)
#define pMSP_DCH2TRM              ((volatile unsigned short *)nMSP_DCH2TRM)
#define pMSP_DCH3SRM              ((volatile unsigned short *)nMSP_DCH3SRM)
#define pMSP_DCH3TRM              ((volatile unsigned short *)nMSP_DCH3TRM)
#define pMSP_DCH4SRM              ((volatile unsigned short *)nMSP_DCH4SRM)
#define pMSP_DCH4TRM              ((volatile unsigned short *)nMSP_DCH4TRM)
#define pMSP_DCH5SRM              ((volatile unsigned short *)nMSP_DCH5SRM)
#define pMSP_DCH5TRM              ((volatile unsigned short *)nMSP_DCH5TRM)
#define pMSP_DCH6SRM              ((volatile unsigned short *)nMSP_DCH6SRM)
#define pMSP_DCH6TRM              ((volatile unsigned short *)nMSP_DCH6TRM)
#define pMSP_DCH7SRM              ((volatile unsigned short *)nMSP_DCH7SRM)
#define pMSP_DCH7TRM              ((volatile unsigned short *)nMSP_DCH7TRM)
#define pMSP_DCH8SRM              ((volatile unsigned short *)nMSP_DCH8SRM)
#define pMSP_DCH8TRM              ((volatile unsigned short *)nMSP_DCH8TRM)
#define pMSP_DCH9SRM              ((volatile unsigned short *)nMSP_DCH9SRM)
#define pMSP_DCH9TRM              ((volatile unsigned short *)nMSP_DCH9TRM)
#define pMSP_DCH10SRM             ((volatile unsigned short *)nMSP_DCH10SRM)
#define pMSP_DCH10TRM             ((volatile unsigned short *)nMSP_DCH10TRM)
#define pMSP_DCH11SRM             ((volatile unsigned short *)nMSP_DCH11SRM)
#define pMSP_DCH11TRM             ((volatile unsigned short *)nMSP_DCH11TRM)
#define pMSP_DCH12SRM             ((volatile unsigned short *)nMSP_DCH12SRM)
#define pMSP_DCH12TRM             ((volatile unsigned short *)nMSP_DCH12TRM)
#define pMSP_DCH13SRM             ((volatile unsigned short *)nMSP_DCH13SRM)
#define pMSP_DCH13TRM             ((volatile unsigned short *)nMSP_DCH13TRM)
#define pMSP_DCH14SRM             ((volatile unsigned short *)nMSP_DCH14SRM)
#define pMSP_DCH14TRM             ((volatile unsigned short *)nMSP_DCH14TRM)
#define pMSP_DCH15SRM             ((volatile unsigned short *)nMSP_DCH15SRM)
#define pMSP_DCH15TRM             ((volatile unsigned short *)nMSP_DCH15TRM)
#define pMSP_DMA0COM0             ((volatile unsigned short *)nMSP_DMA0COM0)
#define pMSP_DMA0COM1             ((volatile unsigned short *)nMSP_DMA0COM1)
#define pMSP_DMA0COM2             ((volatile unsigned short *)nMSP_DMA0COM2)
#define pMSP_DMA0CONS             ((volatile unsigned short *)nMSP_DMA0CONS)
#define pMSP_SRC0LADDR            ((volatile unsigned short *)nMSP_SRC0LADDR)
#define pMSP_SRC0HADDR            ((volatile unsigned short *)nMSP_SRC0HADDR)
#define pMSP_TRG0LADDR            ((volatile unsigned short *)nMSP_TRG0LADDR)
#define pMSP_TRG0HADDR            ((volatile unsigned short *)nMSP_TRG0HADDR)
#define pMSP_DMA1COM0             ((volatile unsigned short *)nMSP_DMA1COM0)
#define pMSP_DMA1COM1             ((volatile unsigned short *)nMSP_DMA1COM1)
#define pMSP_DMA1COM2             ((volatile unsigned short *)nMSP_DMA1COM2)
#define pMSP_DMA1CONS             ((volatile unsigned short *)nMSP_DMA1CONS)
#define pMSP_SRC1LADDR            ((volatile unsigned short *)nMSP_SRC1LADDR)
#define pMSP_SRC1HADDR            ((volatile unsigned short *)nMSP_SRC1HADDR)
#define pMSP_TRG1LADDR            ((volatile unsigned short *)nMSP_TRG1LADDR)
#define pMSP_TRG1HADDR            ((volatile unsigned short *)nMSP_TRG1HADDR)
#define pMSP_DMA2COM0             ((volatile unsigned short *)nMSP_DMA2COM0)
#define pMSP_DMA2COM1             ((volatile unsigned short *)nMSP_DMA2COM1)
#define pMSP_DMA2COM2             ((volatile unsigned short *)nMSP_DMA2COM2)
#define pMSP_DMA2CONS             ((volatile unsigned short *)nMSP_DMA2CONS)
#define pMSP_SRC2LADDR            ((volatile unsigned short *)nMSP_SRC2LADDR)
#define pMSP_SRC2HADDR            ((volatile unsigned short *)nMSP_SRC2HADDR)
#define pMSP_TRG2LADDR            ((volatile unsigned short *)nMSP_TRG2LADDR)
#define pMSP_TRG2HADDR            ((volatile unsigned short *)nMSP_TRG2HADDR)
#define pMSP_DMA3COM0             ((volatile unsigned short *)nMSP_DMA3COM0)
#define pMSP_DMA3COM1             ((volatile unsigned short *)nMSP_DMA3COM1)
#define pMSP_DMA3COM2             ((volatile unsigned short *)nMSP_DMA3COM2)
#define pMSP_DMA3CONS             ((volatile unsigned short *)nMSP_DMA3CONS)
#define pMSP_SRC3LADDR            ((volatile unsigned short *)nMSP_SRC3LADDR)
#define pMSP_SRC3HADDR            ((volatile unsigned short *)nMSP_SRC3HADDR)
#define pMSP_TRG3LADDR            ((volatile unsigned short *)nMSP_TRG3LADDR)
#define pMSP_TRG3HADDR            ((volatile unsigned short *)nMSP_TRG3HADDR)
#define pMSP_DMA4COM0             ((volatile unsigned short *)nMSP_DMA4COM0)
#define pMSP_DMA4COM1             ((volatile unsigned short *)nMSP_DMA4COM1)
#define pMSP_DMA4COM2             ((volatile unsigned short *)nMSP_DMA4COM2)
#define pMSP_DMA4CONS             ((volatile unsigned short *)nMSP_DMA4CONS)
#define pMSP_SRC4LADDR            ((volatile unsigned short *)nMSP_SRC4LADDR)
#define pMSP_SRC4HADDR            ((volatile unsigned short *)nMSP_SRC4HADDR)
#define pMSP_TRG4LADDR            ((volatile unsigned short *)nMSP_TRG4LADDR)
#define pMSP_TRG4HADDR            ((volatile unsigned short *)nMSP_TRG4HADDR)
#define pMSP_DMA5COM0             ((volatile unsigned short *)nMSP_DMA5COM0)
#define pMSP_DMA5COM1             ((volatile unsigned short *)nMSP_DMA5COM1)
#define pMSP_DMA5COM2             ((volatile unsigned short *)nMSP_DMA5COM2)
#define pMSP_DMA5CONS             ((volatile unsigned short *)nMSP_DMA5CONS)
#define pMSP_SRC5LADDR            ((volatile unsigned short *)nMSP_SRC5LADDR)
#define pMSP_SRC5HADDR            ((volatile unsigned short *)nMSP_SRC5HADDR)
#define pMSP_TRG5LADDR            ((volatile unsigned short *)nMSP_TRG5LADDR)
#define pMSP_TRG5HADDR            ((volatile unsigned short *)nMSP_TRG5HADDR)
#define pMSP_DMA6COM0             ((volatile unsigned short *)nMSP_DMA6COM0)
#define pMSP_DMA6COM1             ((volatile unsigned short *)nMSP_DMA6COM1)
#define pMSP_DMA6COM2             ((volatile unsigned short *)nMSP_DMA6COM2)
#define pMSP_DMA6CONS             ((volatile unsigned short *)nMSP_DMA6CONS)
#define pMSP_SRC6LADDR            ((volatile unsigned short *)nMSP_SRC6LADDR)
#define pMSP_SRC6HADDR            ((volatile unsigned short *)nMSP_SRC6HADDR)
#define pMSP_TRG6LADDR            ((volatile unsigned short *)nMSP_TRG6LADDR)
#define pMSP_TRG6HADDR            ((volatile unsigned short *)nMSP_TRG6HADDR)
#define pMSP_DMA7COM0             ((volatile unsigned short *)nMSP_DMA7COM0)
#define pMSP_DMA7COM1             ((volatile unsigned short *)nMSP_DMA7COM1)
#define pMSP_DMA7COM2             ((volatile unsigned short *)nMSP_DMA7COM2)
#define pMSP_DMA7CONS             ((volatile unsigned short *)nMSP_DMA7CONS)
#define pMSP_SRC7LADDR            ((volatile unsigned short *)nMSP_SRC7LADDR)
#define pMSP_SRC7HADDR            ((volatile unsigned short *)nMSP_SRC7HADDR)
#define pMSP_TRG7LADDR            ((volatile unsigned short *)nMSP_TRG7LADDR)
#define pMSP_TRG7HADDR            ((volatile unsigned short *)nMSP_TRG7HADDR)
#define pMSP_DMA8COM0             ((volatile unsigned short *)nMSP_DMA8COM0)
#define pMSP_DMA8COM1             ((volatile unsigned short *)nMSP_DMA8COM1)
#define pMSP_DMA8COM2             ((volatile unsigned short *)nMSP_DMA8COM2)
#define pMSP_DMA8CONS             ((volatile unsigned short *)nMSP_DMA8CONS)
#define pMSP_SRC8LADDR            ((volatile unsigned short *)nMSP_SRC8LADDR)
#define pMSP_SRC8HADDR            ((volatile unsigned short *)nMSP_SRC8HADDR)
#define pMSP_TRG8LADDR            ((volatile unsigned short *)nMSP_TRG8LADDR)
#define pMSP_TRG8HADDR            ((volatile unsigned short *)nMSP_TRG8HADDR)
#define pMSP_DMA9COM0             ((volatile unsigned short *)nMSP_DMA9COM0)
#define pMSP_DMA9COM1             ((volatile unsigned short *)nMSP_DMA9COM1)
#define pMSP_DMA9COM2             ((volatile unsigned short *)nMSP_DMA9COM2)
#define pMSP_DMA9CONS             ((volatile unsigned short *)nMSP_DMA9CONS)
#define pMSP_SRC9LADDR            ((volatile unsigned short *)nMSP_SRC9LADDR)
#define pMSP_SRC9HADDR            ((volatile unsigned short *)nMSP_SRC9HADDR)
#define pMSP_TRG9LADDR            ((volatile unsigned short *)nMSP_TRG9LADDR)
#define pMSP_TRG9HADDR            ((volatile unsigned short *)nMSP_TRG9HADDR)
#define pMSP_DMA10COM0            ((volatile unsigned short *)nMSP_DMA10COM0)
#define pMSP_DMA10COM1            ((volatile unsigned short *)nMSP_DMA10COM1)
#define pMSP_DMA10COM2            ((volatile unsigned short *)nMSP_DMA10COM2)
#define pMSP_DMA10CONS            ((volatile unsigned short *)nMSP_DMA10CONS)
#define pMSP_SRC10LADDR           ((volatile unsigned short *)nMSP_SRC10LADDR)
#define pMSP_SRC10HADDR           ((volatile unsigned short *)nMSP_SRC10HADDR)
#define pMSP_TRG10LADDR           ((volatile unsigned short *)nMSP_TRG10LADDR)
#define pMSP_TRG10HADDR           ((volatile unsigned short *)nMSP_TRG10HADDR)
#define pMSP_DMA11COM0            ((volatile unsigned short *)nMSP_DMA11COM0)
#define pMSP_DMA11COM1            ((volatile unsigned short *)nMSP_DMA11COM1)
#define pMSP_DMA11COM2            ((volatile unsigned short *)nMSP_DMA11COM2)
#define pMSP_DMA11CONS            ((volatile unsigned short *)nMSP_DMA11CONS)
#define pMSP_SRC11LADDR           ((volatile unsigned short *)nMSP_SRC11LADDR)
#define pMSP_SRC11HADDR           ((volatile unsigned short *)nMSP_SRC11HADDR)
#define pMSP_TRG11LADDR           ((volatile unsigned short *)nMSP_TRG11LADDR)
#define pMSP_TRG11HADDR           ((volatile unsigned short *)nMSP_TRG11HADDR)
#define pMSP_DMA12COM0            ((volatile unsigned short *)nMSP_DMA12COM0)
#define pMSP_DMA12COM1            ((volatile unsigned short *)nMSP_DMA12COM1)
#define pMSP_DMA12COM2            ((volatile unsigned short *)nMSP_DMA12COM2)
#define pMSP_DMA12CONS            ((volatile unsigned short *)nMSP_DMA12CONS)
#define pMSP_SRC12LADDR           ((volatile unsigned short *)nMSP_SRC12LADDR)
#define pMSP_SRC12HADDR           ((volatile unsigned short *)nMSP_SRC12HADDR)
#define pMSP_TRG12LADDR           ((volatile unsigned short *)nMSP_TRG12LADDR)
#define pMSP_TRG12HADDR           ((volatile unsigned short *)nMSP_TRG12HADDR)
#define pMSP_DMA13COM0            ((volatile unsigned short *)nMSP_DMA13COM0)
#define pMSP_DMA13COM1            ((volatile unsigned short *)nMSP_DMA13COM1)
#define pMSP_DMA13COM2            ((volatile unsigned short *)nMSP_DMA13COM2)
#define pMSP_DMA13CONS            ((volatile unsigned short *)nMSP_DMA13CONS)
#define pMSP_SRC13LADDR           ((volatile unsigned short *)nMSP_SRC13LADDR)
#define pMSP_SRC13HADDR           ((volatile unsigned short *)nMSP_SRC13HADDR)
#define pMSP_TRG13LADDR           ((volatile unsigned short *)nMSP_TRG13LADDR)
#define pMSP_TRG13HADDR           ((volatile unsigned short *)nMSP_TRG13HADDR)
#define pMSP_DMA14COM0            ((volatile unsigned short *)nMSP_DMA14COM0)
#define pMSP_DMA14COM1            ((volatile unsigned short *)nMSP_DMA14COM1)
#define pMSP_DMA14COM2            ((volatile unsigned short *)nMSP_DMA14COM2)
#define pMSP_DMA14CONS            ((volatile unsigned short *)nMSP_DMA14CONS)
#define pMSP_SRC14LADDR           ((volatile unsigned short *)nMSP_SRC14LADDR)
#define pMSP_SRC14HADDR           ((volatile unsigned short *)nMSP_SRC14HADDR)
#define pMSP_TRG14LADDR           ((volatile unsigned short *)nMSP_TRG14LADDR)
#define pMSP_TRG14HADDR           ((volatile unsigned short *)nMSP_TRG14HADDR)
#define pMSP_DMA15COM0            ((volatile unsigned short *)nMSP_DMA15COM0)
#define pMSP_DMA15COM1            ((volatile unsigned short *)nMSP_DMA15COM1)
#define pMSP_DMA15COM2            ((volatile unsigned short *)nMSP_DMA15COM2)
#define pMSP_DMA15CONS            ((volatile unsigned short *)nMSP_DMA15CONS)
#define pMSP_SRC15LADDR           ((volatile unsigned short *)nMSP_SRC15LADDR)
#define pMSP_SRC15HADDR           ((volatile unsigned short *)nMSP_SRC15HADDR)
#define pMSP_TRG15LADDR           ((volatile unsigned short *)nMSP_TRG15LADDR)
#define pMSP_TRG15HADDR           ((volatile unsigned short *)nMSP_TRG15HADDR)
#define pMSP_SRCPEND              ((volatile unsigned int *)nMSP_SRCPEND)
#define pMSP_INTMODE              ((volatile unsigned int *)nMSP_INTMODE)
#define pMSP_INTMASK              ((volatile unsigned int *)nMSP_INTMASK)
#define pMSP_PRIORITY             ((volatile unsigned int *)nMSP_PRIORITY)
#define pMSP_INTPEND              ((volatile unsigned int *)nMSP_INTPEND)
#define pMSP_INTOFFSET            ((volatile unsigned short *)nMSP_INTOFFSET)
#define pMSP_TCOUNT               ((volatile unsigned int *)nMSP_TCOUNT)
#define pMSP_TMATCH0              ((volatile unsigned int *)nMSP_TMATCH0)
#define pMSP_TMATCH1              ((volatile unsigned int *)nMSP_TMATCH1)
#define pMSP_TMATCH2              ((volatile unsigned int *)nMSP_TMATCH2)
#define pMSP_TMATCH3              ((volatile unsigned int *)nMSP_TMATCH3)
#define pMSP_TCONTROL             ((volatile unsigned short *)nMSP_TCONTROL)
#define pMSP_TSTATUS              ((volatile unsigned short *)nMSP_TSTATUS)
#define pMSP_TINTEN               ((volatile unsigned short *)nMSP_TINTEN)
#define pMSP_LCON0                ((volatile unsigned short *)nMSP_LCON0)
#define pMSP_UCON0                ((volatile unsigned short *)nMSP_UCON0)
#define pMSP_FCON0                ((volatile unsigned short *)nMSP_FCON0)
#define pMSP_MCON0                ((volatile unsigned short *)nMSP_MCON0)
#define pMSP_TRSTATUS0            ((volatile unsigned short *)nMSP_TRSTATUS0)
#define pMSP_ESTATUS0             ((volatile unsigned short *)nMSP_ESTATUS0)
#define pMSP_FSTATUS0             ((volatile unsigned short *)nMSP_FSTATUS0)
#define pMSP_MSTATUS0             ((volatile unsigned short *)nMSP_MSTATUS0)
#define pMSP_THB0                 ((volatile unsigned char *)nMSP_THB0)
#define pMSP_RHB0                 ((volatile unsigned char *)nMSP_RHB0)
#define pMSP_BRD0                 ((volatile unsigned short *)nMSP_BRD0)
#define pMSP_TIMEOUTREG0          ((volatile unsigned short *)nMSP_TIMEOUTREG0)
#define pMSP_LCON1                ((volatile unsigned short *)nMSP_LCON1)
#define pMSP_UCON1                ((volatile unsigned short *)nMSP_UCON1)
#define pMSP_FCON1                ((volatile unsigned short *)nMSP_FCON1)
#define pMSP_MCON1                ((volatile unsigned short *)nMSP_MCON1)
#define pMSP_TRSTATUS1            ((volatile unsigned short *)nMSP_TRSTATUS1)
#define pMSP_ESTATUS1             ((volatile unsigned short *)nMSP_ESTATUS1)
#define pMSP_FSTATUS1             ((volatile unsigned short *)nMSP_FSTATUS1)
#define pMSP_MSTATUS1             ((volatile unsigned short *)nMSP_MSTATUS1)
#define pMSP_THB1                 ((volatile unsigned char *)nMSP_THB1)
#define pMSP_RHB1                 ((volatile unsigned char *)nMSP_RHB1)
#define pMSP_BRD1                 ((volatile unsigned short *)nMSP_BRD1)
#define pMSP_TIMEOUTREG1          ((volatile unsigned short *)nMSP_TIMEOUTREG1)
#define pMSP_LCON2                ((volatile unsigned short *)nMSP_LCON2)
#define pMSP_UCON2                ((volatile unsigned short *)nMSP_UCON2)
#define pMSP_FCON2                ((volatile unsigned short *)nMSP_FCON2)
#define pMSP_MCON2                ((volatile unsigned short *)nMSP_MCON2)
#define pMSP_TRSTATUS2            ((volatile unsigned short *)nMSP_TRSTATUS2)
#define pMSP_ESTATUS2             ((volatile unsigned short *)nMSP_ESTATUS2)
#define pMSP_FSTATUS2             ((volatile unsigned short *)nMSP_FSTATUS2)
#define pMSP_MSTATUS2             ((volatile unsigned short *)nMSP_MSTATUS2)
#define pMSP_THB2                 ((volatile unsigned char *)nMSP_THB2)
#define pMSP_RHB2                 ((volatile unsigned char *)nMSP_RHB2)
#define pMSP_BRD2                 ((volatile unsigned short *)nMSP_BRD2)
#define pMSP_TIMEOUTREG2          ((volatile unsigned short *)nMSP_TIMEOUTREG2)
#define pMSP_LCON3                ((volatile unsigned short *)nMSP_LCON3)
#define pMSP_UCON3                ((volatile unsigned short *)nMSP_UCON3)
#define pMSP_FCON3                ((volatile unsigned short *)nMSP_FCON3)
#define pMSP_MCON3                ((volatile unsigned short *)nMSP_MCON3)
#define pMSP_TRSTATUS3            ((volatile unsigned short *)nMSP_TRSTATUS3)
#define pMSP_ESTATUS3             ((volatile unsigned short *)nMSP_ESTATUS3)
#define pMSP_FSTATUS3             ((volatile unsigned short *)nMSP_FSTATUS3)
#define pMSP_MSTATUS3             ((volatile unsigned short *)nMSP_MSTATUS3)
#define pMSP_THB3                 ((volatile unsigned char *)nMSP_THB3)
#define pMSP_RHB3                 ((volatile unsigned char *)nMSP_RHB3)
#define pMSP_BRD3                 ((volatile unsigned short *)nMSP_BRD3)
#define pMSP_TIMEOUTREG3          ((volatile unsigned short *)nMSP_TIMEOUTREG3)
#define pMSP_INTSTATREG           ((volatile unsigned short *)nMSP_INTSTATREG)
#define pMSP_PORTCON              ((volatile unsigned short *)nMSP_PORTCON)
#define pMSP_GPIOAALTFNLOW        ((volatile unsigned short *)nMSP_GPIOAALTFNLOW)
#define pMSP_GPIOBALTFNLOW        ((volatile unsigned short *)nMSP_GPIOBALTFNLOW)
#define pMSP_GPIOCALTFNLOW        ((volatile unsigned short *)nMSP_GPIOCALTFNLOW)
#define pMSP_GPIODALTFNLOW        ((volatile unsigned short *)nMSP_GPIODALTFNLOW)
#define pMSP_GPIOEALTFNLOW        ((volatile unsigned short *)nMSP_GPIOEALTFNLOW)
#define pMSP_GPIOFALTFNLOW        ((volatile unsigned short *)nMSP_GPIOFALTFNLOW)
#define pMSP_GPIOGALTFNLOW        ((volatile unsigned short *)nMSP_GPIOGALTFNLOW)
#define pMSP_GPIOHALTFNLOW        ((volatile unsigned short *)nMSP_GPIOHALTFNLOW)
#define pMSP_GPIOIALTFNLOW        ((volatile unsigned short *)nMSP_GPIOIALTFNLOW)
#define pMSP_GPIOJALTFNLOW        ((volatile unsigned short *)nMSP_GPIOJALTFNLOW)
#define pMSP_GPIOKALTFNLOW        ((volatile unsigned short *)nMSP_GPIOKALTFNLOW)
#define pMSP_GPIOLALTFNLOW        ((volatile unsigned short *)nMSP_GPIOLALTFNLOW)
#define pMSP_GPIOMALTFNLOW        ((volatile unsigned short *)nMSP_GPIOMALTFNLOW)
#define pMSP_GPIONALTFNLOW        ((volatile unsigned short *)nMSP_GPIONALTFNLOW)
#define pMSP_GPIOOALTFNLOW        ((volatile unsigned short *)nMSP_GPIOOALTFNLOW)
#define pMSP_GPIOAALTFNHI         ((volatile unsigned short *)nMSP_GPIOAALTFNHI)
#define pMSP_GPIOBALTFNHI         ((volatile unsigned short *)nMSP_GPIOBALTFNHI)
#define pMSP_GPIOCALTFNHI         ((volatile unsigned short *)nMSP_GPIOCALTFNHI)
#define pMSP_GPIODALTFNHI         ((volatile unsigned short *)nMSP_GPIODALTFNHI)
#define pMSP_GPIOEALTFNHI         ((volatile unsigned short *)nMSP_GPIOEALTFNHI)
#define pMSP_GPIOFALTFNHI         ((volatile unsigned short *)nMSP_GPIOFALTFNHI)
#define pMSP_GPIOGALTFNHI         ((volatile unsigned short *)nMSP_GPIOGALTFNHI)
#define pMSP_GPIOHALTFNHI         ((volatile unsigned short *)nMSP_GPIOHALTFNHI)
#define pMSP_GPIOIALTFNHI         ((volatile unsigned short *)nMSP_GPIOIALTFNHI)
#define pMSP_GPIOJALTFNHI         ((volatile unsigned short *)nMSP_GPIOJALTFNHI)
#define pMSP_GPIOKALTFNHI         ((volatile unsigned short *)nMSP_GPIOKALTFNHI)
#define pMSP_GPIOLALTFNHI         ((volatile unsigned short *)nMSP_GPIOLALTFNHI)
#define pMSP_GPIOMALTFNHI         ((volatile unsigned short *)nMSP_GPIOMALTFNHI)
#define pMSP_GPIONALTFNHI         ((volatile unsigned short *)nMSP_GPIONALTFNHI)
#define pMSP_GPIOOALTFNHI         ((volatile unsigned short *)nMSP_GPIOOALTFNHI)
#define pMSP_GPIOAOUT             ((volatile unsigned short *)nMSP_GPIOAOUT)
#define pMSP_GPIOBOUT             ((volatile unsigned short *)nMSP_GPIOBOUT)
#define pMSP_GPIOCOUT             ((volatile unsigned short *)nMSP_GPIOCOUT)
#define pMSP_GPIODOUT             ((volatile unsigned short *)nMSP_GPIODOUT)
#define pMSP_GPIOEOUT             ((volatile unsigned short *)nMSP_GPIOEOUT)
#define pMSP_GPIOFOUT             ((volatile unsigned short *)nMSP_GPIOFOUT)
#define pMSP_GPIOGOUT             ((volatile unsigned short *)nMSP_GPIOGOUT)
#define pMSP_GPIOHOUT             ((volatile unsigned short *)nMSP_GPIOHOUT)
#define pMSP_GPIOIOUT             ((volatile unsigned short *)nMSP_GPIOIOUT)
#define pMSP_GPIOJOUT             ((volatile unsigned short *)nMSP_GPIOJOUT)
#define pMSP_GPIOKOUT             ((volatile unsigned short *)nMSP_GPIOKOUT)
#define pMSP_GPIOLOUT             ((volatile unsigned short *)nMSP_GPIOLOUT)
#define pMSP_GPIOMOUT             ((volatile unsigned short *)nMSP_GPIOMOUT)
#define pMSP_GPIONOUT             ((volatile unsigned short *)nMSP_GPIONOUT)
#define pMSP_GPIOOOUT             ((volatile unsigned short *)nMSP_GPIOOOUT)
#define pMSP_GPIOAEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOAEVTTYPLOW)
#define pMSP_GPIOBEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOBEVTTYPLOW)
#define pMSP_GPIOCEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOCEVTTYPLOW)
#define pMSP_GPIODEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIODEVTTYPLOW)
#define pMSP_GPIOEEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOEEVTTYPLOW)
#define pMSP_GPIOFEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOFEVTTYPLOW)
#define pMSP_GPIOGEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOGEVTTYPLOW)
#define pMSP_GPIOHEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOHEVTTYPLOW)
#define pMSP_GPIOIEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOIEVTTYPLOW)
#define pMSP_GPIOJEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOJEVTTYPLOW)
#define pMSP_GPIOKEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOKEVTTYPLOW)
#define pMSP_GPIOLEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOLEVTTYPLOW)
#define pMSP_GPIOMEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOMEVTTYPLOW)
#define pMSP_GPIONEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIONEVTTYPLOW)
#define pMSP_GPIOOEVTTYPLOW       ((volatile unsigned short *)nMSP_GPIOOEVTTYPLOW)
#define pMSP_GPIOAEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOAEVTTYPHI)
#define pMSP_GPIOBEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOBEVTTYPHI)
#define pMSP_GPIOCEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOCEVTTYPHI)
#define pMSP_GPIODEVTTYPHI        ((volatile unsigned short *)nMSP_GPIODEVTTYPHI)
#define pMSP_GPIOEEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOEEVTTYPHI)
#define pMSP_GPIOFEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOFEVTTYPHI)
#define pMSP_GPIOGEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOGEVTTYPHI)
#define pMSP_GPIOHEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOHEVTTYPHI)
#define pMSP_GPIOIEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOIEVTTYPHI)
#define pMSP_GPIOJEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOJEVTTYPHI)
#define pMSP_GPIOKEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOKEVTTYPHI)
#define pMSP_GPIOLEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOLEVTTYPHI)
#define pMSP_GPIOMEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOMEVTTYPHI)
#define pMSP_GPIONEVTTYPHI        ((volatile unsigned short *)nMSP_GPIONEVTTYPHI)
#define pMSP_GPIOOEVTTYPHI        ((volatile unsigned short *)nMSP_GPIOOEVTTYPHI)
#define pMSP_GPIOAPUENB           ((volatile unsigned short *)nMSP_GPIOAPUENB)
#define pMSP_GPIOBPUENB           ((volatile unsigned short *)nMSP_GPIOBPUENB)
#define pMSP_GPIOCPUENB           ((volatile unsigned short *)nMSP_GPIOCPUENB)
#define pMSP_GPIODPUENB           ((volatile unsigned short *)nMSP_GPIODPUENB)
#define pMSP_GPIOEPUENB           ((volatile unsigned short *)nMSP_GPIOEPUENB)
#define pMSP_GPIOFPUENB           ((volatile unsigned short *)nMSP_GPIOFPUENB)
#define pMSP_GPIOGPUENB           ((volatile unsigned short *)nMSP_GPIOGPUENB)
#define pMSP_GPIOHPUENB           ((volatile unsigned short *)nMSP_GPIOHPUENB)
#define pMSP_GPIOIPUENB           ((volatile unsigned short *)nMSP_GPIOIPUENB)
#define pMSP_GPIOJPUENB           ((volatile unsigned short *)nMSP_GPIOJPUENB)
#define pMSP_GPIOKPUENB           ((volatile unsigned short *)nMSP_GPIOKPUENB)
#define pMSP_GPIOLPUENB           ((volatile unsigned short *)nMSP_GPIOLPUENB)
#define pMSP_GPIOMPUENB           ((volatile unsigned short *)nMSP_GPIOMPUENB)
#define pMSP_GPIONPUENB           ((volatile unsigned short *)nMSP_GPIONPUENB)
#define pMSP_GPIOOPUENB           ((volatile unsigned short *)nMSP_GPIOOPUENB)
#define pMSP_GPIOAINTENB          ((volatile unsigned short *)nMSP_GPIOAINTENB)
#define pMSP_GPIOBINTENB          ((volatile unsigned short *)nMSP_GPIOBINTENB)
#define pMSP_GPIOCINTENB          ((volatile unsigned short *)nMSP_GPIOCINTENB)
#define pMSP_GPIODINTENB          ((volatile unsigned short *)nMSP_GPIODINTENB)
#define pMSP_GPIOEINTENB          ((volatile unsigned short *)nMSP_GPIOEINTENB)
#define pMSP_GPIOFINTENB          ((volatile unsigned short *)nMSP_GPIOFINTENB)
#define pMSP_GPIOGINTENB          ((volatile unsigned short *)nMSP_GPIOGINTENB)
#define pMSP_GPIOHINTENB          ((volatile unsigned short *)nMSP_GPIOHINTENB)
#define pMSP_GPIOIINTENB          ((volatile unsigned short *)nMSP_GPIOIINTENB)
#define pMSP_GPIOJINTENB          ((volatile unsigned short *)nMSP_GPIOJINTENB)
#define pMSP_GPIOKINTENB          ((volatile unsigned short *)nMSP_GPIOKINTENB)
#define pMSP_GPIOLINTENB          ((volatile unsigned short *)nMSP_GPIOLINTENB)
#define pMSP_GPIOMINTENB          ((volatile unsigned short *)nMSP_GPIOMINTENB)
#define pMSP_GPIONINTENB          ((volatile unsigned short *)nMSP_GPIONINTENB)
#define pMSP_GPIOOINTENB          ((volatile unsigned short *)nMSP_GPIOOINTENB)
#define pMSP_GPIOAEVT             ((volatile unsigned short *)nMSP_GPIOAEVT)
#define pMSP_GPIOBEVT             ((volatile unsigned short *)nMSP_GPIOBEVT)
#define pMSP_GPIOCEVT             ((volatile unsigned short *)nMSP_GPIOCEVT)
#define pMSP_GPIODEVT             ((volatile unsigned short *)nMSP_GPIODEVT)
#define pMSP_GPIOEEVT             ((volatile unsigned short *)nMSP_GPIOEEVT)
#define pMSP_GPIOFEVT             ((volatile unsigned short *)nMSP_GPIOFEVT)
#define pMSP_GPIOGEVT             ((volatile unsigned short *)nMSP_GPIOGEVT)
#define pMSP_GPIOHEVT             ((volatile unsigned short *)nMSP_GPIOHEVT)
#define pMSP_GPIOIEVT             ((volatile unsigned short *)nMSP_GPIOIEVT)
#define pMSP_GPIOJEVT             ((volatile unsigned short *)nMSP_GPIOJEVT)
#define pMSP_GPIOKEVT             ((volatile unsigned short *)nMSP_GPIOKEVT)
#define pMSP_GPIOLEVT             ((volatile unsigned short *)nMSP_GPIOLEVT)
#define pMSP_GPIOMEVT             ((volatile unsigned short *)nMSP_GPIOMEVT)
#define pMSP_GPIONEVT             ((volatile unsigned short *)nMSP_GPIONEVT)
#define pMSP_GPIOOEVT             ((volatile unsigned short *)nMSP_GPIOOEVT)
#define pMSP_GPIOPADSEL           ((volatile unsigned short *)nMSP_GPIOPADSEL)
#define pMSP_GPIOGREVTSTAT        ((volatile unsigned short *)nMSP_GPIOGREVTSTAT)
#define pMSP_GPIOAPINLVL          ((volatile unsigned short *)nMSP_GPIOAPINLVL)
#define pMSP_GPIOBPINLVL          ((volatile unsigned short *)nMSP_GPIOBPINLVL)
#define pMSP_GPIOCPINLVL          ((volatile unsigned short *)nMSP_GPIOCPINLVL)
#define pMSP_GPIODPINLVL          ((volatile unsigned short *)nMSP_GPIODPINLVL)
#define pMSP_GPIOEPINLVL          ((volatile unsigned short *)nMSP_GPIOEPINLVL)
#define pMSP_GPIOFPINLVL          ((volatile unsigned short *)nMSP_GPIOFPINLVL)
#define pMSP_GPIOGPINLVL          ((volatile unsigned short *)nMSP_GPIOGPINLVL)
#define pMSP_GPIOHPINLVL          ((volatile unsigned short *)nMSP_GPIOHPINLVL)
#define pMSP_GPIOIPINLVL          ((volatile unsigned short *)nMSP_GPIOIPINLVL)
#define pMSP_GPIOJPINLVL          ((volatile unsigned short *)nMSP_GPIOJPINLVL)
#define pMSP_GPIOKPINLVL          ((volatile unsigned short *)nMSP_GPIOKPINLVL)
#define pMSP_GPIOLPINLVL          ((volatile unsigned short *)nMSP_GPIOLPINLVL)
#define pMSP_GPIOMPINLVL          ((volatile unsigned short *)nMSP_GPIOMPINLVL)
#define pMSP_GPIONPINLVL          ((volatile unsigned short *)nMSP_GPIONPINLVL)
#define pMSP_GPIOOPINLVL          ((volatile unsigned short *)nMSP_GPIOOPINLVL)
#define pMSP_GPIOEOUT_ALIVE       ((volatile unsigned short *)nMSP_GPIOEOUT_ALIVE)
#define pMSP_GPIOEOUTENB_ALIVE    ((volatile unsigned short *)nMSP_GPIOEOUTENB_ALIVE)
#define pMSP_GPIOEPUENB_ALIVE     ((volatile unsigned short *)nMSP_GPIOEPUENB_ALIVE)
#define pMSP_GPIOFOUT_ALIVE       ((volatile unsigned short *)nMSP_GPIOFOUT_ALIVE)
#define pMSP_GPIOFOUTENB_ALIVE    ((volatile unsigned short *)nMSP_GPIOFOUTENB_ALIVE)
#define pMSP_GPIOFPUENB_ALIVE     ((volatile unsigned short *)nMSP_GPIOFPUENB_ALIVE)
#define pMSP_GPIOGOUT_ALIVE       ((volatile unsigned short *)nMSP_GPIOGOUT_ALIVE)
#define pMSP_GPIOGOUTENB_ALIVE    ((volatile unsigned short *)nMSP_GPIOGOUTENB_ALIVE)
#define pMSP_GPIOGPUENB_ALIVE     ((volatile unsigned short *)nMSP_GPIOGPUENB_ALIVE)
#define pMSP_GPIOIOUT_ALIVE       ((volatile unsigned short *)nMSP_GPIOIOUT_ALIVE)
#define pMSP_GPIOIOUTENB_ALIVE    ((volatile unsigned short *)nMSP_GPIOIOUTENB_ALIVE)
#define pMSP_GPIOIPUENB_ALIVE     ((volatile unsigned short *)nMSP_GPIOIPUENB_ALIVE)
#define pMSP_GPIOJOUT_ALIVE       ((volatile unsigned short *)nMSP_GPIOJOUT_ALIVE)
#define pMSP_GPIOJOUTENB_ALIVE    ((volatile unsigned short *)nMSP_GPIOJOUTENB_ALIVE)
#define pMSP_GPIOJPUENB_ALIVE     ((volatile unsigned short *)nMSP_GPIOJPUENB_ALIVE)
#define pMSP_SC_CMD               ((volatile unsigned short *)nMSP_SC_CMD)
#define pMSP_SC_STATUS            ((volatile unsigned short *)nMSP_SC_STATUS)
#define pMSP_SC_MIRROR            ((volatile unsigned short *)nMSP_SC_MIRROR)
#define pMSP_SC_PRE_VRATIO        ((volatile unsigned short *)nMSP_SC_PRE_VRATIO)
#define pMSP_SC_PRE_HRATIO        ((volatile unsigned short *)nMSP_SC_PRE_HRATIO)
#define pMSP_SC_POST_VRATIOL      ((volatile unsigned short *)nMSP_SC_POST_VRATIOL)
#define pMSP_SC_POST_VRATIOH      ((volatile unsigned short *)nMSP_SC_POST_VRATIOH)
#define pMSP_SC_POST_HRATIOL      ((volatile unsigned short *)nMSP_SC_POST_HRATIOL)
#define pMSP_SC_POST_HRATIOH      ((volatile unsigned short *)nMSP_SC_POST_HRATIOH)
#define pMSP_SC_SRC_ODD_ADDRL     ((volatile unsigned short *)nMSP_SC_SRC_ODD_ADDRL)
#define pMSP_SC_SRC_ODD_ADDRH     ((volatile unsigned short *)nMSP_SC_SRC_ODD_ADDRH)
#define pMSP_SC_SRC_EVEN_ADDRL    ((volatile unsigned short *)nMSP_SC_SRC_EVEN_ADDRL)
#define pMSP_SC_SRC_EVEN_ADDRH    ((volatile unsigned short *)nMSP_SC_SRC_EVEN_ADDRH)
#define pMSP_SC_DST_ADDRL         ((volatile unsigned short *)nMSP_SC_DST_ADDRL)
#define pMSP_SC_DST_ADDRH         ((volatile unsigned short *)nMSP_SC_DST_ADDRH)
#define pMSP_SC_SRC_PXL_HEIGHT    ((volatile unsigned short *)nMSP_SC_SRC_PXL_HEIGHT)
#define pMSP_SC_SRC_PXL_WIDTH     ((volatile unsigned short *)nMSP_SC_SRC_PXL_WIDTH)
#define pMSP_SC_SRC_PXL_REQCNT    ((volatile unsigned short *)nMSP_SC_SRC_PXL_REQCNT)
#define pMSP_SC_DST_PXL_HEIGHT    ((volatile unsigned short *)nMSP_SC_DST_PXL_HEIGHT)
#define pMSP_SC_DST_PXL_WIDTH     ((volatile unsigned short *)nMSP_SC_DST_PXL_WIDTH)
#define pMSP_SC_DST_WPXL_WIDTH    ((volatile unsigned short *)nMSP_SC_DST_WPXL_WIDTH)
#define pMSP_SC_LUMA_OFFSET       ((volatile unsigned short *)nMSP_SC_LUMA_OFFSET)
#define pMSP_SC_CB_OFFSET         ((volatile unsigned short *)nMSP_SC_CB_OFFSET)
#define pMSP_SC_CR_OFFSET         ((volatile unsigned short *)nMSP_SC_CR_OFFSET)
#define pMSP_SC_DELAY             ((volatile unsigned short *)nMSP_SC_DELAY)
#define pMSP_SC_MEM_CNTR          ((volatile unsigned short *)nMSP_SC_MEM_CNTR)
#define pMSP_SC_IRQ               ((volatile unsigned short *)nMSP_SC_IRQ)
#define pMSP_SC_RESERVED          ((volatile unsigned short *)nMSP_SC_RESERVED)
#define pMSP_MLC_OVLAY_CNTR       ((volatile unsigned short *)nMSP_MLC_OVLAY_CNTR)
#define pMSP_MLC_YUV_EFFECT       ((volatile unsigned short *)nMSP_MLC_YUV_EFFECT)
#define pMSP_MLC_YUV_CNTL         ((volatile unsigned short *)nMSP_MLC_YUV_CNTL)
#define pMSP_MLC_YUVA_TP_HSC      ((volatile unsigned short *)nMSP_MLC_YUVA_TP_HSC)
#define pMSP_MLC_YUVA_BT_HSC      ((volatile unsigned short *)nMSP_MLC_YUVA_BT_HSC)
#define pMSP_MLC_YUVA_TP_HSCL     ((volatile unsigned short *)nMSP_MLC_YUVA_TP_HSCL)
#define pMSP_MLC_YUVA_TP_HSCH     ((volatile unsigned short *)nMSP_MLC_YUVA_TP_HSCH)
#define pMSP_MLC_YUVA_BT_HSCL     ((volatile unsigned short *)nMSP_MLC_YUVA_BT_HSCL)
#define pMSP_MLC_YUVA_BT_HSCH     ((volatile unsigned short *)nMSP_MLC_YUVA_BT_HSCH)
#define pMSP_MLC_YUVA_TP_PXW      ((volatile unsigned short *)nMSP_MLC_YUVA_TP_PXW)
#define pMSP_MLC_YUVA_BT_PXW      ((volatile unsigned short *)nMSP_MLC_YUVA_BT_PXW)
#define pMSP_MLC_YUVA_STX         ((volatile unsigned short *)nMSP_MLC_YUVA_STX)
#define pMSP_MLC_YUVA_ENDX        ((volatile unsigned short *)nMSP_MLC_YUVA_ENDX)
#define pMSP_MLC_YUVA_TP_STY      ((volatile unsigned short *)nMSP_MLC_YUVA_TP_STY)
#define pMSP_MLC_YUVA_TP_ENDY     ((volatile unsigned short *)nMSP_MLC_YUVA_TP_ENDY)
#define pMSP_MLC_YUVA_BT_ENDY     ((volatile unsigned short *)nMSP_MLC_YUVA_BT_ENDY)
#define pMSP_MLC_YUVA_TP_OADRL    ((volatile unsigned short *)nMSP_MLC_YUVA_TP_OADRL)
#define pMSP_MLC_YUVA_TP_OADRH    ((volatile unsigned short *)nMSP_MLC_YUVA_TP_OADRH)
#define pMSP_MLC_YUVA_TP_EADRL    ((volatile unsigned short *)nMSP_MLC_YUVA_TP_EADRL)
#define pMSP_MLC_YUVA_TP_EADRH    ((volatile unsigned short *)nMSP_MLC_YUVA_TP_EADRH)
#define pMSP_MLC_YUVA_BT_OADRL    ((volatile unsigned short *)nMSP_MLC_YUVA_BT_OADRL)
#define pMSP_MLC_YUVA_BT_OADRH    ((volatile unsigned short *)nMSP_MLC_YUVA_BT_OADRH)
#define pMSP_MLC_YUVA_BT_EADRL    ((volatile unsigned short *)nMSP_MLC_YUVA_BT_EADRL)
#define pMSP_MLC_YUVA_BT_EADRH    ((volatile unsigned short *)nMSP_MLC_YUVA_BT_EADRH)
#define pMSP_MLC_YUVB_TP_HSC      ((volatile unsigned short *)nMSP_MLC_YUVB_TP_HSC)
#define pMSP_MLC_YUVB_BT_HSC      ((volatile unsigned short *)nMSP_MLC_YUVB_BT_HSC)
#define pMSP_MLC_YUVB_TP_VSCL     ((volatile unsigned short *)nMSP_MLC_YUVB_TP_VSCL)
#define pMSP_MLC_YUVB_TP_VSCH     ((volatile unsigned short *)nMSP_MLC_YUVB_TP_VSCH)
#define pMSP_MLC_YUVB_BT_VSCL     ((volatile unsigned short *)nMSP_MLC_YUVB_BT_VSCL)
#define pMSP_MLC_YUVB_BT_VSCH     ((volatile unsigned short *)nMSP_MLC_YUVB_BT_VSCH)
#define pMSP_MLC_YUVB_TP_PXW      ((volatile unsigned short *)nMSP_MLC_YUVB_TP_PXW)
#define pMSP_MLC_YUVB_BT_PXW      ((volatile unsigned short *)nMSP_MLC_YUVB_BT_PXW)
#define pMSP_MLC_YUVB_STX         ((volatile unsigned short *)nMSP_MLC_YUVB_STX)
#define pMSP_MLC_YUVB_ENDX        ((volatile unsigned short *)nMSP_MLC_YUVB_ENDX)
#define pMSP_MLC_YUVB_TP_STY      ((volatile unsigned short *)nMSP_MLC_YUVB_TP_STY)
#define pMSP_MLC_YUVB_TP_ENDY     ((volatile unsigned short *)nMSP_MLC_YUVB_TP_ENDY)
#define pMSP_MLC_YUVB_BT_ENDY     ((volatile unsigned short *)nMSP_MLC_YUVB_BT_ENDY)
#define pMSP_MLC_YUVB_TP_OADRL    ((volatile unsigned short *)nMSP_MLC_YUVB_TP_OADRL)
#define pMSP_MLC_YUVB_TP_OADRH    ((volatile unsigned short *)nMSP_MLC_YUVB_TP_OADRH)
#define pMSP_MLC_YUVB_TP_EADRL    ((volatile unsigned short *)nMSP_MLC_YUVB_TP_EADRL)
#define pMSP_MLC_YUVB_TP_EADRH    ((volatile unsigned short *)nMSP_MLC_YUVB_TP_EADRH)
#define pMSP_MLC_YUVB_BT_OADRL    ((volatile unsigned short *)nMSP_MLC_YUVB_BT_OADRL)
#define pMSP_MLC_YUVB_BT_OADRH    ((volatile unsigned short *)nMSP_MLC_YUVB_BT_OADRH)
#define pMSP_MLC_YUVB_BT_EADRL    ((volatile unsigned short *)nMSP_MLC_YUVB_BT_EADRL)
#define pMSP_MLC_YUVB_BT_EADRH    ((volatile unsigned short *)nMSP_MLC_YUVB_BT_EADRH)
#define pMSP_MLC_STL_CNTL         ((volatile unsigned short *)nMSP_MLC_STL_CNTL)
#define pMSP_MLC_STL_MIXMUX       ((volatile unsigned short *)nMSP_MLC_STL_MIXMUX)
#define pMSP_MLC_RGB_ALPHAL       ((volatile unsigned short *)nMSP_MLC_RGB_ALPHAL)
#define pMSP_MLC_STL_ALPHAL       ((volatile unsigned short *)nMSP_MLC_STL_ALPHAL)
#define pMSP_MLC_STL_ALPHAH       ((volatile unsigned short *)nMSP_MLC_STL_ALPHAH)
#define pMSP_MLC_STL1_STX         ((volatile unsigned short *)nMSP_MLC_STL1_STX)
#define pMSP_MLC_STL1_ENDX        ((volatile unsigned short *)nMSP_MLC_STL1_ENDX)
#define pMSP_MLC_STL1_STY         ((volatile unsigned short *)nMSP_MLC_STL1_STY)
#define pMSP_MLC_STL1_ENDY        ((volatile unsigned short *)nMSP_MLC_STL1_ENDY)
#define pMSP_MLC_STL2_STX         ((volatile unsigned short *)nMSP_MLC_STL2_STX)
#define pMSP_MLC_STL2_ENDX        ((volatile unsigned short *)nMSP_MLC_STL2_ENDX)
#define pMSP_MLC_STL2_STY         ((volatile unsigned short *)nMSP_MLC_STL2_STY)
#define pMSP_MLC_STL2_ENDY        ((volatile unsigned short *)nMSP_MLC_STL2_ENDY)
#define pMSP_MLC_STL3_STX         ((volatile unsigned short *)nMSP_MLC_STL3_STX)
#define pMSP_MLC_STL3_ENDX        ((volatile unsigned short *)nMSP_MLC_STL3_ENDX)
#define pMSP_MLC_STL3_STY         ((volatile unsigned short *)nMSP_MLC_STL3_STY)
#define pMSP_MLC_STL3_ENDY        ((volatile unsigned short *)nMSP_MLC_STL3_ENDY)
#define pMSP_MLC_STL4_STX         ((volatile unsigned short *)nMSP_MLC_STL4_STX)
#define pMSP_MLC_STL4_ENDX        ((volatile unsigned short *)nMSP_MLC_STL4_ENDX)
#define pMSP_MLC_STL4_STY         ((volatile unsigned short *)nMSP_MLC_STL4_STY)
#define pMSP_MLC_STL4_ENDY        ((volatile unsigned short *)nMSP_MLC_STL4_ENDY)
#define pMSP_MLC_STL_CKEY_GB      ((volatile unsigned short *)nMSP_MLC_STL_CKEY_GB)
#define pMSP_MLC_STL_CKEY_R       ((volatile unsigned short *)nMSP_MLC_STL_CKEY_R)
#define pMSP_MLC_STL_HSC          ((volatile unsigned short *)nMSP_MLC_STL_HSC)
#define pMSP_MLC_STL_VSCL         ((volatile unsigned short *)nMSP_MLC_STL_VSCL)
#define pMSP_MLC_STL_VSCH         ((volatile unsigned short *)nMSP_MLC_STL_VSCH)
#define pMSP_MLC_STL_HW           ((volatile unsigned short *)nMSP_MLC_STL_HW)
#define pMSP_MLC_STL_OADRL        ((volatile unsigned short *)nMSP_MLC_STL_OADRL)
#define pMSP_MLC_STL_OADRH        ((volatile unsigned short *)nMSP_MLC_STL_OADRH)
#define pMSP_MLC_STL_EADRL        ((volatile unsigned short *)nMSP_MLC_STL_EADRL)
#define pMSP_MLC_STL_EADRH        ((volatile unsigned short *)nMSP_MLC_STL_EADRH)
#define pMSP_MLC_OSD_OADRL        ((volatile unsigned short *)nMSP_MLC_OSD_OADRL)
#define pMSP_MLC_OSD_OADRH        ((volatile unsigned short *)nMSP_MLC_OSD_OADRH)
#define pMSP_MLC_OSD_EADRL        ((volatile unsigned short *)nMSP_MLC_OSD_EADRL)
#define pMSP_MLC_OSD_EADRH        ((volatile unsigned short *)nMSP_MLC_OSD_EADRH)
#define pMSP_MLC_HWC_CNTR         ((volatile unsigned short *)nMSP_MLC_HWC_CNTR)
#define pMSP_MLC_HWC_STX          ((volatile unsigned short *)nMSP_MLC_HWC_STX)
#define pMSP_MLC_HWC_STY          ((volatile unsigned short *)nMSP_MLC_HWC_STY)
#define pMSP_MLC_HWC_FGR          ((volatile unsigned short *)nMSP_MLC_HWC_FGR)
#define pMSP_MLC_HWC_FB           ((volatile unsigned short *)nMSP_MLC_HWC_FB)
#define pMSP_MLC_HWC_BGR          ((volatile unsigned short *)nMSP_MLC_HWC_BGR)
#define pMSP_MLC_HWC_BB           ((volatile unsigned short *)nMSP_MLC_HWC_BB)
#define pMSP_MLC_HWC_OADRL        ((volatile unsigned short *)nMSP_MLC_HWC_OADRL)
#define pMSP_MLC_HWC_OADRH        ((volatile unsigned short *)nMSP_MLC_HWC_OADRH)
#define pMSP_MLC_HWC_EADRL        ((volatile unsigned short *)nMSP_MLC_HWC_EADRL)
#define pMSP_MLC_HWC_EADRH        ((volatile unsigned short *)nMSP_MLC_HWC_EADRH)
#define pMSP_MLC_LUMA_ENH         ((volatile unsigned short *)nMSP_MLC_LUMA_ENH)
#define pMSP_MLC_HUECB1AB         ((volatile unsigned short *)nMSP_MLC_HUECB1AB)
#define pMSP_MLC_HUECR1AB         ((volatile unsigned short *)nMSP_MLC_HUECR1AB)
#define pMSP_MLC_HUECB2AB         ((volatile unsigned short *)nMSP_MLC_HUECB2AB)
#define pMSP_MLC_HUECR2AB         ((volatile unsigned short *)nMSP_MLC_HUECR2AB)
#define pMSP_MLC_HUECB3AB         ((volatile unsigned short *)nMSP_MLC_HUECB3AB)
#define pMSP_MLC_HUECR3AB         ((volatile unsigned short *)nMSP_MLC_HUECR3AB)
#define pMSP_MLC_HUECB4AB         ((volatile unsigned short *)nMSP_MLC_HUECB4AB)
#define pMSP_MLC_HUECR4AB         ((volatile unsigned short *)nMSP_MLC_HUECR4AB)
#define pMSP_MLC_DITHER           ((volatile unsigned short *)nMSP_MLC_DITHER)
#define pMSP_MLC_OSD_PALLT_A      ((volatile unsigned short *)nMSP_MLC_OSD_PALLT_A)
#define pMSP_MLC_OSD_PALLT_D      ((volatile unsigned short *)nMSP_MLC_OSD_PALLT_D)
#define pMSP_MLC_STL_PALLT_A      ((volatile unsigned short *)nMSP_MLC_STL_PALLT_A)
#define pMSP_MLC_STL_PALLT_D      ((volatile unsigned short *)nMSP_MLC_STL_PALLT_D)
#define pMSP_MLC_GAMMA_A          ((volatile unsigned short *)nMSP_MLC_GAMMA_A)
#define pMSP_MLC_GAMMA_D          ((volatile unsigned short *)nMSP_MLC_GAMMA_D)
#define pMSP_MLC_SPU_CTRL         ((volatile unsigned short *)nMSP_MLC_SPU_CTRL)
#define pMSP_MLC_SPU_DELAY        ((volatile unsigned short *)nMSP_MLC_SPU_DELAY)
#define pMSP_MLC_SPU_BASEI_ADDRL  ((volatile unsigned short *)nMSP_MLC_SPU_BASEI_ADDRL)
#define pMSP_MLC_SPU_BASEI_ADDRH  ((volatile unsigned short *)nMSP_MLC_SPU_BASEI_ADDRH)
#define pMSP_MLC_SPU_HLI_ADDRL    ((volatile unsigned short *)nMSP_MLC_SPU_HLI_ADDRL)
#define pMSP_MLC_SPU_HLI_ADDRH    ((volatile unsigned short *)nMSP_MLC_SPU_HLI_ADDRH)
#define pMSP_MLC_SPU_PAL          ((volatile unsigned short *)nMSP_MLC_SPU_PAL)
#define pMSP_DPC_CNTL             ((volatile unsigned short *)nMSP_DPC_CNTL)
#define pMSP_DPC_FPICNTL          ((volatile unsigned short *)nMSP_DPC_FPICNTL)
#define pMSP_DPC_FPIPOL1          ((volatile unsigned short *)nMSP_DPC_FPIPOL1)
#define pMSP_DPC_FPIPOL2          ((volatile unsigned short *)nMSP_DPC_FPIPOL2)
#define pMSP_DPC_FPIPOL3          ((volatile unsigned short *)nMSP_DPC_FPIPOL3)
#define pMSP_DPC_FPIATV1          ((volatile unsigned short *)nMSP_DPC_FPIATV1)
#define pMSP_DPC_FPIATV2          ((volatile unsigned short *)nMSP_DPC_FPIATV2)
#define pMSP_DPC_FPIATV3          ((volatile unsigned short *)nMSP_DPC_FPIATV3)
#define pMSP_DPC_X_MAX            ((volatile unsigned short *)nMSP_DPC_X_MAX)
#define pMSP_DPC_Y_MAX            ((volatile unsigned short *)nMSP_DPC_Y_MAX)
#define pMSP_DPC_HS_WIDTH         ((volatile unsigned short *)nMSP_DPC_HS_WIDTH)
#define pMSP_DPC_HS_STR           ((volatile unsigned short *)nMSP_DPC_HS_STR)
#define pMSP_DPC_HS_END           ((volatile unsigned short *)nMSP_DPC_HS_END)
#define pMSP_DPC_V_SYNC           ((volatile unsigned short *)nMSP_DPC_V_SYNC)
#define pMSP_DPC_V_END            ((volatile unsigned short *)nMSP_DPC_V_END)
#define pMSP_DPC_TV_BNK           ((volatile unsigned short *)nMSP_DPC_TV_BNK)
#define pMSP_DPC_DE               ((volatile unsigned short *)nMSP_DPC_DE)
#define pMSP_DPC_PS               ((volatile unsigned short *)nMSP_DPC_PS)
#define pMSP_DPC_FG               ((volatile unsigned short *)nMSP_DPC_FG)
#define pMSP_DPC_LP               ((volatile unsigned short *)nMSP_DPC_LP)
#define pMSP_DPC_CLKVH            ((volatile unsigned short *)nMSP_DPC_CLKVH)
#define pMSP_DPC_CLKVL            ((volatile unsigned short *)nMSP_DPC_CLKVL)
#define pMSP_DPC_POL              ((volatile unsigned short *)nMSP_DPC_POL)
#define pMSP_DPC_CISSYNC          ((volatile unsigned short *)nMSP_DPC_CISSYNC)
#define pMSP_DPC_MID_SYNC         ((volatile unsigned short *)nMSP_DPC_MID_SYNC)
#define pMSP_DPC_C_SYNC           ((volatile unsigned short *)nMSP_DPC_C_SYNC)
#define pMSP_DPC_Y_BLANK          ((volatile unsigned short *)nMSP_DPC_Y_BLANK)
#define pMSP_DPC_C_BLANK          ((volatile unsigned short *)nMSP_DPC_C_BLANK)
#define pMSP_DPC_YP_CSYNC         ((volatile unsigned short *)nMSP_DPC_YP_CSYNC)
#define pMSP_DPC_YN_CSYNC         ((volatile unsigned short *)nMSP_DPC_YN_CSYNC)
#define pMSP_DPC_CP_CSYNC         ((volatile unsigned short *)nMSP_DPC_CP_CSYNC)
#define pMSP_DPC_CN_CSYNC         ((volatile unsigned short *)nMSP_DPC_CN_CSYNC)
#define pMSP_DPC_INTR             ((volatile unsigned short *)nMSP_DPC_INTR)
#define pMSP_DPC_CLKCNTL          ((volatile unsigned short *)nMSP_DPC_CLKCNTL)
#define pMSP_MESGDSTCTRL          ((volatile unsigned int *)nMSP_MESGDSTCTRL)
#define pMSP_MESGDSTADDR          ((volatile unsigned int *)nMSP_MESGDSTADDR)
#define pMSP_MESGDSTSTRIDE        ((volatile unsigned int *)nMSP_MESGDSTSTRIDE)
#define pMSP_MESGSRCCTRL          ((volatile unsigned int *)nMSP_MESGSRCCTRL)
#define pMSP_MESGSRCADDR          ((volatile unsigned int *)nMSP_MESGSRCADDR)
#define pMSP_MESGSRCSTRIDE        ((volatile unsigned int *)nMSP_MESGSRCSTRIDE)
#define pMSP_MESGSRCFORCOLOR      ((volatile unsigned int *)nMSP_MESGSRCFORCOLOR)
#define pMSP_MESGSRCBACKCOLOR     ((volatile unsigned int *)nMSP_MESGSRCBACKCOLOR)
#define pMSP_MESGPATCTRL          ((volatile unsigned int *)nMSP_MESGPATCTRL)
#define pMSP_MESGFORCOLOR         ((volatile unsigned int *)nMSP_MESGFORCOLOR)
#define pMSP_MESGBACKCOLOR        ((volatile unsigned int *)nMSP_MESGBACKCOLOR)
#define pMSP_MESGSIZE             ((volatile unsigned int *)nMSP_MESGSIZE)
#define pMSP_MESGCTRL             ((volatile unsigned int *)nMSP_MESGCTRL)
#define pMSP_MESGSTATUS           ((volatile unsigned int *)nMSP_MESGSTATUS)
#define pMSP_MESGFIFOSTATUS       ((volatile unsigned int *)nMSP_MESGFIFOSTATUS)
#define pMSP_MESGFIFO             ((volatile unsigned int *)nMSP_MESGFIFO)
#define pMSP_MESGPAT              ((volatile unsigned int *)nMSP_MESGPAT)
#define pMSP_ISP_CON_1            ((volatile unsigned short *)nMSP_ISP_CON_1)
#define pMSP_ISP_CON_2            ((volatile unsigned short *)nMSP_ISP_CON_2)
#define pMSP_ISP_INT_CON          ((volatile unsigned short *)nMSP_ISP_INT_CON)
#define pMSP_ISP_VPERIOD          ((volatile unsigned short *)nMSP_ISP_VPERIOD)
#define pMSP_ISP_VPW              ((volatile unsigned short *)nMSP_ISP_VPW)
#define pMSP_ISP_HPERIOD          ((volatile unsigned short *)nMSP_ISP_HPERIOD)
#define pMSP_ISP_HPW              ((volatile unsigned short *)nMSP_ISP_HPW)
#define pMSP_ISP_HACT_ST          ((volatile unsigned short *)nMSP_ISP_HACT_ST)
#define pMSP_ISP_H_WIDTH          ((volatile unsigned short *)nMSP_ISP_H_WIDTH)
#define pMSP_ISP_VACT_ST          ((volatile unsigned short *)nMSP_ISP_VACT_ST)
#define pMSP_ISP_V_HEIGHT         ((volatile unsigned short *)nMSP_ISP_V_HEIGHT)
#define pMSP_ISP_SRC_HWIDTH       ((volatile unsigned short *)nMSP_ISP_SRC_HWIDTH)
#define pMSP_ISP_RGBSV            ((volatile unsigned short *)nMSP_ISP_RGBSV)
#define pMSP_ISP_RGBEV            ((volatile unsigned short *)nMSP_ISP_RGBEV)
#define pMSP_ISP_RGBSH            ((volatile unsigned short *)nMSP_ISP_RGBSH)
#define pMSP_ISP_RGBEH            ((volatile unsigned short *)nMSP_ISP_RGBEH)
#define pMSP_ISP_AESV             ((volatile unsigned short *)nMSP_ISP_AESV)
#define pMSP_ISP_AEEV             ((volatile unsigned short *)nMSP_ISP_AEEV)
#define pMSP_ISP_AESH             ((volatile unsigned short *)nMSP_ISP_AESH)
#define pMSP_ISP_AEEH             ((volatile unsigned short *)nMSP_ISP_AEEH)
#define pMSP_ISP_AELIM            ((volatile unsigned short *)nMSP_ISP_AELIM)
#define pMSP_ISP_AEVSTEP          ((volatile unsigned short *)nMSP_ISP_AEVSTEP)
#define pMSP_ISP_AEHSTEP          ((volatile unsigned short *)nMSP_ISP_AEHSTEP)
#define pMSP_ISP_AFSV             ((volatile unsigned short *)nMSP_ISP_AFSV)
#define pMSP_ISP_AFEV             ((volatile unsigned short *)nMSP_ISP_AFEV)
#define pMSP_ISP_AFSH             ((volatile unsigned short *)nMSP_ISP_AFSH)
#define pMSP_ISP_AFEH             ((volatile unsigned short *)nMSP_ISP_AFEH)
#define pMSP_ISP_CNTRSTGAIN       ((volatile unsigned short *)nMSP_ISP_CNTRSTGAIN)
#define pMSP_ISP_CNTRSTLIM        ((volatile unsigned short *)nMSP_ISP_CNTRSTLIM)
#define pMSP_ISP_RGBREC           ((volatile unsigned short *)nMSP_ISP_RGBREC)
#define pMSP_ISP_HUECB0AB         ((volatile unsigned short *)nMSP_ISP_HUECB0AB)
#define pMSP_ISP_HUECR0AB         ((volatile unsigned short *)nMSP_ISP_HUECR0AB)
#define pMSP_ISP_HUECB1AB         ((volatile unsigned short *)nMSP_ISP_HUECB1AB)
#define pMSP_ISP_HUECR1AB         ((volatile unsigned short *)nMSP_ISP_HUECR1AB)
#define pMSP_ISP_HUECB2AB         ((volatile unsigned short *)nMSP_ISP_HUECB2AB)
#define pMSP_ISP_HUECR2AB         ((volatile unsigned short *)nMSP_ISP_HUECR2AB)
#define pMSP_ISP_HUECB3AB         ((volatile unsigned short *)nMSP_ISP_HUECB3AB)
#define pMSP_ISP_HUECR3AB         ((volatile unsigned short *)nMSP_ISP_HUECR3AB)
#define pMSP_ISP_YGAIN            ((volatile unsigned short *)nMSP_ISP_YGAIN)
#define pMSP_ISP_CDLY             ((volatile unsigned short *)nMSP_ISP_CDLY)
#define pMSP_ISP_BRIGHT           ((volatile unsigned short *)nMSP_ISP_BRIGHT)
#define pMSP_ISP_YCLIP            ((volatile unsigned short *)nMSP_ISP_YCLIP)
#define pMSP_ISP_YCLIPSLANT       ((volatile unsigned short *)nMSP_ISP_YCLIPSLANT)
#define pMSP_ISP_CSUPSLANT        ((volatile unsigned short *)nMSP_ISP_CSUPSLANT)
#define pMSP_ISP_SLICELMT         ((volatile unsigned short *)nMSP_ISP_SLICELMT)
#define pMSP_ISP_WBLMT            ((volatile unsigned short *)nMSP_ISP_WBLMT)
#define pMSP_ISP_WBACCCNTH        ((volatile unsigned short *)nMSP_ISP_WBACCCNTH)
#define pMSP_ISP_WBACCCNTL        ((volatile unsigned short *)nMSP_ISP_WBACCCNTL)
#define pMSP_ISP_WBCBACCH         ((volatile unsigned short *)nMSP_ISP_WBCBACCH)
#define pMSP_ISP_WBCBACCL         ((volatile unsigned short *)nMSP_ISP_WBCBACCL)
#define pMSP_ISP_WBCRACCH         ((volatile unsigned short *)nMSP_ISP_WBCRACCH)
#define pMSP_ISP_WBCRACCL         ((volatile unsigned short *)nMSP_ISP_WBCRACCL)
#define pMSP_ISP_AFACCH           ((volatile unsigned short *)nMSP_ISP_AFACCH)
#define pMSP_ISP_AFACCL           ((volatile unsigned short *)nMSP_ISP_AFACCL)
#define pMSP_ISP_WBCOORD0         ((volatile unsigned short *)nMSP_ISP_WBCOORD0)
#define pMSP_ISP_WBCOORD1         ((volatile unsigned short *)nMSP_ISP_WBCOORD1)
#define pMSP_ISP_WBCOORD2         ((volatile unsigned short *)nMSP_ISP_WBCOORD2)
#define pMSP_ISP_WBCOORD3         ((volatile unsigned short *)nMSP_ISP_WBCOORD3)
#define pMSP_ISP_WBCOORD4         ((volatile unsigned short *)nMSP_ISP_WBCOORD4)
#define pMSP_ISP_WBCOORD5         ((volatile unsigned short *)nMSP_ISP_WBCOORD5)
#define pMSP_ISP_WBCOORD6         ((volatile unsigned short *)nMSP_ISP_WBCOORD6)
#define pMSP_ISP_WBCOORD7         ((volatile unsigned short *)nMSP_ISP_WBCOORD7)
#define pMSP_ISP_WBCOORD8         ((volatile unsigned short *)nMSP_ISP_WBCOORD8)
#define pMSP_ISP_WBCOORD9         ((volatile unsigned short *)nMSP_ISP_WBCOORD9)
#define pMSP_ISP_WBCOORD10        ((volatile unsigned short *)nMSP_ISP_WBCOORD10)
#define pMSP_ISP_WBCOORD11        ((volatile unsigned short *)nMSP_ISP_WBCOORD11)
#define pMSP_ISP_WBCOORD12        ((volatile unsigned short *)nMSP_ISP_WBCOORD12)
#define pMSP_ISP_WBCOORD13        ((volatile unsigned short *)nMSP_ISP_WBCOORD13)
#define pMSP_ISP_WBCOORD14        ((volatile unsigned short *)nMSP_ISP_WBCOORD14)
#define pMSP_ISP_WBCOORD15        ((volatile unsigned short *)nMSP_ISP_WBCOORD15)
#define pMSP_ISP_MEMADDRL1        ((volatile unsigned short *)nMSP_ISP_MEMADDRL1)
#define pMSP_ISP_MEMADDRH1        ((volatile unsigned short *)nMSP_ISP_MEMADDRH1)
#define pMSP_ISP_GAMMA            ((volatile unsigned short *)nMSP_ISP_GAMMA)
#define pMSP_ISP_GAMMD            ((volatile unsigned short *)nMSP_ISP_GAMMD)
#define pMSP_ISP_AEACCH           ((volatile unsigned short *)nMSP_ISP_AEACCH)
#define pMSP_ISP_AEACCL           ((volatile unsigned short *)nMSP_ISP_AEACCL)
#define pMSP_ISP_CNTST_LLCNTL     ((volatile unsigned short *)nMSP_ISP_CNTST_LLCNTL)
#define pMSP_ISP_CNTST_LLCNTH     ((volatile unsigned short *)nMSP_ISP_CNTST_LLCNTH)
#define pMSP_ISP_CNTST_HLCNTL     ((volatile unsigned short *)nMSP_ISP_CNTST_HLCNTL)
#define pMSP_ISP_CNTST_HLCNTH     ((volatile unsigned short *)nMSP_ISP_CNTST_HLCNTH)
#define pMSP_ISP_CNTST_LLACCL     ((volatile unsigned short *)nMSP_ISP_CNTST_LLACCL)
#define pMSP_ISP_CNTST_LLACCH     ((volatile unsigned short *)nMSP_ISP_CNTST_LLACCH)
#define pMSP_ISP_CNTST_HLACCL     ((volatile unsigned short *)nMSP_ISP_CNTST_HLACCL)
#define pMSP_ISP_CNTST_HLACCH     ((volatile unsigned short *)nMSP_ISP_CNTST_HLACCH)
#define pMSP_ISP_MEMRGBADDRL      ((volatile unsigned short *)nMSP_ISP_MEMRGBADDRL)
#define pMSP_ISP_MEMRGBADDRH      ((volatile unsigned short *)nMSP_ISP_MEMRGBADDRH)
#define pMSP_ISP_MEMADDRL2        ((volatile unsigned short *)nMSP_ISP_MEMADDRL2)
#define pMSP_ISP_MEMADDRH2        ((volatile unsigned short *)nMSP_ISP_MEMADDRH2)
#define pMSP_ISP_DST_HWIDTH       ((volatile unsigned short *)nMSP_ISP_DST_HWIDTH)
#define pMSP_PWM_PRES01           ((volatile unsigned short *)nMSP_PWM_PRES01)
#define pMSP_PWM_DUTY0            ((volatile unsigned short *)nMSP_PWM_DUTY0)
#define pMSP_PWM_DUTY1            ((volatile unsigned short *)nMSP_PWM_DUTY1)
#define pMSP_PWM_PERIOD0          ((volatile unsigned short *)nMSP_PWM_PERIOD0)
#define pMSP_PWM_PERIOD1          ((volatile unsigned short *)nMSP_PWM_PERIOD1)
#define pMSP_PWM_PRES23           ((volatile unsigned short *)nMSP_PWM_PRES23)
#define pMSP_PWM_DUTY2            ((volatile unsigned short *)nMSP_PWM_DUTY2)
#define pMSP_PWM_DUTY3            ((volatile unsigned short *)nMSP_PWM_DUTY3)
#define pMSP_PWM_PERIOD2          ((volatile unsigned short *)nMSP_PWM_PERIOD2)
#define pMSP_PWM_PERIOD3          ((volatile unsigned short *)nMSP_PWM_PERIOD3)
#define pMSP_PPMCTRL              ((volatile unsigned short *)nMSP_PPMCTRL)
#define pMSP_PPMSTATUS            ((volatile unsigned short *)nMSP_PPMSTATUS)
#define pMSP_PPMLOWPERIOD         ((volatile unsigned short *)nMSP_PPMLOWPERIOD)
#define pMSP_PPMHIGHPERIOD        ((volatile unsigned short *)nMSP_PPMHIGHPERIOD)
#define pMSP_FUNC_ADDR_REG        ((volatile unsigned short *)nMSP_FUNC_ADDR_REG)
#define pMSP_PWR_REG              ((volatile unsigned short *)nMSP_PWR_REG)
#define pMSP_EP_INT_REG           ((volatile unsigned short *)nMSP_EP_INT_REG)
#define pMSP_USB_INT_REG          ((volatile unsigned short *)nMSP_USB_INT_REG)
#define pMSP_EP_INT_EN_REG        ((volatile unsigned short *)nMSP_EP_INT_EN_REG)
#define pMSP_USB_INT_EN_REG       ((volatile unsigned short *)nMSP_USB_INT_EN_REG)
#define pMSP_FRAME_NUM1_REG       ((volatile unsigned short *)nMSP_FRAME_NUM1_REG)
#define pMSP_FRAME_NUM2_REG       ((volatile unsigned short *)nMSP_FRAME_NUM2_REG)
#define pMSP_INDEX_REG            ((volatile unsigned short *)nMSP_INDEX_REG)
#define pMSP_MAXP_REG             ((volatile unsigned short *)nMSP_MAXP_REG)
#define pMSP_EP0_CSR              ((volatile unsigned short *)nMSP_EP0_CSR)
#define pMSP_IN_CSR1_REG          ((volatile unsigned short *)nMSP_IN_CSR1_REG)
#define pMSP_IN_CSR2_REG          ((volatile unsigned short *)nMSP_IN_CSR2_REG)
#define pMSP_OUT_MAXP_REG         ((volatile unsigned short *)nMSP_OUT_MAXP_REG)
#define pMSP_OUT_CSR1_REG         ((volatile unsigned short *)nMSP_OUT_CSR1_REG)
#define pMSP_OUT_CSR2_REG         ((volatile unsigned short *)nMSP_OUT_CSR2_REG)
#define pMSP_OUT_FIFO_CNT1_REG    ((volatile unsigned short *)nMSP_OUT_FIFO_CNT1_REG)
#define pMSP_OUT_FIFO_CNT2_REG    ((volatile unsigned short *)nMSP_OUT_FIFO_CNT2_REG)
#define pMSP_EP0_FIFO_REG         ((volatile unsigned short *)nMSP_EP0_FIFO_REG)
#define pMSP_EP1_FIFO_REG         ((volatile unsigned short *)nMSP_EP1_FIFO_REG)
#define pMSP_EP2_FIFO_REG         ((volatile unsigned short *)nMSP_EP2_FIFO_REG)
#define pMSP_EP3_FIFO_REG         ((volatile unsigned short *)nMSP_EP3_FIFO_REG)
#define pMSP_EP4_FIFO_REG         ((volatile unsigned short *)nMSP_EP4_FIFO_REG)
#define pMSP_EP1_DMA_CON          ((volatile unsigned short *)nMSP_EP1_DMA_CON)
#define pMSP_EP1_DMA_FIFO         ((volatile unsigned short *)nMSP_EP1_DMA_FIFO)
#define pMSP_EP1_DMA_TTC_L        ((volatile unsigned short *)nMSP_EP1_DMA_TTC_L)
#define pMSP_EP1_DMA_TTC_M        ((volatile unsigned short *)nMSP_EP1_DMA_TTC_M)
#define pMSP_EP1_DMA_TTC_H        ((volatile unsigned short *)nMSP_EP1_DMA_TTC_H)
#define pMSP_EP2_DMA_CON          ((volatile unsigned short *)nMSP_EP2_DMA_CON)
#define pMSP_EP2_DMA_FIFO         ((volatile unsigned short *)nMSP_EP2_DMA_FIFO)
#define pMSP_EP2_DMA_TTC_L        ((volatile unsigned short *)nMSP_EP2_DMA_TTC_L)
#define pMSP_EP2_DMA_TTC_M        ((volatile unsigned short *)nMSP_EP2_DMA_TTC_M)
#define pMSP_EP2_DMA_TTC_H        ((volatile unsigned short *)nMSP_EP2_DMA_TTC_H)
#define pMSP_EP3_DMA_CON          ((volatile unsigned short *)nMSP_EP3_DMA_CON)
#define pMSP_EP3_DMA_FIFO         ((volatile unsigned short *)nMSP_EP3_DMA_FIFO)
#define pMSP_EP3_DMA_TTC_L        ((volatile unsigned short *)nMSP_EP3_DMA_TTC_L)
#define pMSP_EP3_DMA_TTC_M        ((volatile unsigned short *)nMSP_EP3_DMA_TTC_M)
#define pMSP_EP3_DMA_TTC_H        ((volatile unsigned short *)nMSP_EP3_DMA_TTC_H)
#define pMSP_EP4_DMA_CON          ((volatile unsigned short *)nMSP_EP4_DMA_CON)
#define pMSP_EP4_DMA_FIFO         ((volatile unsigned short *)nMSP_EP4_DMA_FIFO)
#define pMSP_EP4_DMA_TTC_L        ((volatile unsigned short *)nMSP_EP4_DMA_TTC_L)
#define pMSP_EP4_DMA_TTC_M        ((volatile unsigned short *)nMSP_EP4_DMA_TTC_M)
#define pMSP_EP4_DMA_TTC_H        ((volatile unsigned short *)nMSP_EP4_DMA_TTC_H)
#define pMSP_I2CCON               ((volatile unsigned short *)nMSP_I2CCON)
#define pMSP_I2CSTAT              ((volatile unsigned short *)nMSP_I2CSTAT)
#define pMSP_I2CADD               ((volatile unsigned short *)nMSP_I2CADD)
#define pMSP_I2CDS                ((volatile unsigned char *)nMSP_I2CDS)
#define pMSP_SPDIFCON             ((volatile unsigned short *)nMSP_SPDIFCON)
#define pMSP_SPDIFINTENB          ((volatile unsigned short *)nMSP_SPDIFINTENB)
#define pMSP_SPDIFSTATUS          ((volatile unsigned short *)nMSP_SPDIFSTATUS)
#define pMSP_SPDIFTXDIV           ((volatile unsigned short *)nMSP_SPDIFTXDIV)
#define pMSP_AC_CTRL_REG          ((volatile unsigned short *)nMSP_AC_CTRL_REG)
#define pMSP_AC_CONFIG_REG        ((volatile unsigned short *)nMSP_AC_CONFIG_REG)
#define pMSP_AC_STA_ENA_REG       ((volatile unsigned short *)nMSP_AC_STA_ENA_REG)
#define pMSP_AC_STATUS_REG        ((volatile unsigned short *)nMSP_AC_STATUS_REG)
#define pMSP_AC_STATE_MACHINE     ((volatile unsigned short *)nMSP_AC_STATE_MACHINE)
#define pMSP_AC_CODEC_ADDR        ((volatile unsigned short *)nMSP_AC_CODEC_ADDR)
#define pMSP_AC_CODEC_DATA        ((volatile unsigned short *)nMSP_AC_CODEC_DATA)
#define pMSP_AC_CODEC_ACCESS      ((volatile unsigned short *)nMSP_AC_CODEC_ACCESS)
#define pMSP_AC_CODEC_REG_BASE    ((volatile unsigned short *)nMSP_AC_CODEC_REG_BASE)
#define pMSP_I2S_CTRL_REG         ((volatile unsigned short *)nMSP_I2S_CTRL_REG)
#define pMSP_I2S_CONFIG_REG       ((volatile unsigned short *)nMSP_I2S_CONFIG_REG)
#define pMSP_I2S_STA_ENA_REG      ((volatile unsigned short *)nMSP_I2S_STA_ENA_REG)
#define pMSP_I2S_STATUS_REG       ((volatile unsigned short *)nMSP_I2S_STATUS_REG)
#define pMSP_SSPSPICONT0          ((volatile unsigned short *)nMSP_SSPSPICONT0)
#define pMSP_SSPSPICONT1          ((volatile unsigned short *)nMSP_SSPSPICONT1)
#define pMSP_SSPSPIDATA           ((volatile unsigned char *)nMSP_SSPSPIDATA)
#define pMSP_SSPSPISTAT           ((volatile unsigned short *)nMSP_SSPSPISTAT)
#define pMSP_MPEGIFCONT           ((volatile unsigned short *)nMSP_MPEGIFCONT)
#define pMSP_SDICON               ((volatile unsigned short *)nMSP_SDICON)
#define pMSP_SDIPRE               ((volatile unsigned char *)nMSP_SDIPRE)
#define pMSP_SDICMDARGL           ((volatile unsigned short *)nMSP_SDICMDARGL)
#define pMSP_SDICMDARGH           ((volatile unsigned short *)nMSP_SDICMDARGH)
#define pMSP_SDICMDCON            ((volatile unsigned short *)nMSP_SDICMDCON)
#define pMSP_SDICMDSTA            ((volatile unsigned short *)nMSP_SDICMDSTA)
#define pMSP_SDIRSP0              ((volatile unsigned short *)nMSP_SDIRSP0)
#define pMSP_SDIRSP1              ((volatile unsigned short *)nMSP_SDIRSP1)
#define pMSP_SDIRSP2              ((volatile unsigned short *)nMSP_SDIRSP2)
#define pMSP_SDIRSP3              ((volatile unsigned short *)nMSP_SDIRSP3)
#define pMSP_SDIRSP4              ((volatile unsigned short *)nMSP_SDIRSP4)
#define pMSP_SDIRSP5              ((volatile unsigned short *)nMSP_SDIRSP5)
#define pMSP_SDIRSP6              ((volatile unsigned short *)nMSP_SDIRSP6)
#define pMSP_SDIRSP7              ((volatile unsigned short *)nMSP_SDIRSP7)
#define pMSP_SDIBSIZE             ((volatile unsigned short *)nMSP_SDIBSIZE)
#define pMSP_SDIDATCONL           ((volatile unsigned short *)nMSP_SDIDATCONL)
#define pMSP_SDIDATCONH           ((volatile unsigned short *)nMSP_SDIDATCONH)
#define pMSP_SDIDATCNTL           ((volatile unsigned short *)nMSP_SDIDATCNTL)
#define pMSP_SDIDATCNTH           ((volatile unsigned short *)nMSP_SDIDATCNTH)
#define pMSP_SDIDATSTA            ((volatile unsigned short *)nMSP_SDIDATSTA)
#define pMSP_SDIFSTA              ((volatile unsigned short *)nMSP_SDIFSTA)
#define pMSP_SDIDAT               ((volatile unsigned char *)nMSP_SDIDAT)
#define pMSP_SDIINTMSK0           ((volatile unsigned short *)nMSP_SDIINTMSK0)
#define pMSP_SDIINTMSK1           ((volatile unsigned short *)nMSP_SDIINTMSK1)
#define pMSP_SDIDTIMERL           ((volatile unsigned short *)nMSP_SDIDTIMERL)
#define pMSP_SDIDTIMERH           ((volatile unsigned short *)nMSP_SDIDTIMERH)
#define pMSP_SDISFTREGL           ((volatile unsigned short *)nMSP_SDISFTREGL)
#define pMSP_SDISFTREGH           ((volatile unsigned short *)nMSP_SDISFTREGH)
#define pMSP_MSCTRL               ((volatile unsigned short *)nMSP_MSCTRL)
#define pMSP_MSFIFO               ((volatile unsigned short *)nMSP_MSFIFO)
#define pMSP_MSFIFOST             ((volatile unsigned short *)nMSP_MSFIFOST)
#define pMSP_MSINTCTRL            ((volatile unsigned short *)nMSP_MSINTCTRL)
#define pMSP_MSCRC                ((volatile unsigned short *)nMSP_MSCRC)
#define pMSP_MSTIMEOUT            ((volatile unsigned short *)nMSP_MSTIMEOUT)
#define pMSP_MSTPC                ((volatile unsigned short *)nMSP_MSTPC)
#define pMSP_MSTPCCNT             ((volatile unsigned short *)nMSP_MSTPCCNT)
#define pMSP_OWM_CMD              ((volatile unsigned short *)nMSP_OWM_CMD)
#define pMSP_OWM_BUF              ((volatile unsigned short *)nMSP_OWM_BUF)
#define pMSP_OWM_STA              ((volatile unsigned short *)nMSP_OWM_STA)
#define pMSP_OWM_STA_ENB          ((volatile unsigned short *)nMSP_OWM_STA_ENB)
#define pMSP_OWM_CLK_DIVISOR      ((volatile unsigned short *)nMSP_OWM_CLK_DIVISOR)
#define pMSP_OWM_CTRL             ((volatile unsigned short *)nMSP_OWM_CTRL)
#define pMSP_OWM_IRQ              ((volatile unsigned short *)nMSP_OWM_IRQ)
#define pMSP_DECSSCTRL            ((volatile unsigned short *)nMSP_DECSSCTRL)
#define pMSP_DECSSKEY0            ((volatile unsigned short *)nMSP_DECSSKEY0)
#define pMSP_DECSSKEY1            ((volatile unsigned short *)nMSP_DECSSKEY1)
#define pMSP_DECSSKEY2            ((volatile unsigned short *)nMSP_DECSSKEY2)
#define pMSP_DECSSDATA            ((volatile unsigned short *)nMSP_DECSSDATA)
#define pMSP_DECSSNUM             ((volatile unsigned short *)nMSP_DECSSNUM)

/* Below here are the memory access definitions. */

#define MSP_MEMCFGW               (*pMSP_MEMCFGW)
#define MSP_MEMTIMEW0             (*pMSP_MEMTIMEW0)
#define MSP_MEMTIMEW1             (*pMSP_MEMTIMEW1)
#define MSP_MRMTIMEW2             (*pMSP_MRMTIMEW2)
#define MSP_MEMTIMEW3             (*pMSP_MEMTIMEW3)
#define MSP_MEMTIMEW4             (*pMSP_MEMTIMEW4)
#define MSP_MEMWAITCTRLW          (*pMSP_MEMWAITCTRLW)
#define MSP_MEMPAGEW              (*pMSP_MEMPAGEW)
#define MSP_MEMIDETIMEW           (*pMSP_MEMIDETIMEW)
#define MSP_MEMPCMCIAMW           (*pMSP_MEMPCMCIAMW)
#define MSP_MEMPCMCIAAW           (*pMSP_MEMPCMCIAAW)
#define MSP_MEMPCMCIAIW           (*pMSP_MEMPCMCIAIW)
#define MSP_MEMPCMCIAWAITW        (*pMSP_MEMPCMCIAWAITW)
#define MSP_MEMIDEWAITW           (*pMSP_MEMIDEWAITW)
#define MSP_MEMDTIMEOUTW          (*pMSP_MEMDTIMEOUTW)
#define MSP_MEMDMACTRLW           (*pMSP_MEMDMACTRLW)
#define MSP_MEMDMAPOLW            (*pMSP_MEMDMAPOLW)
#define MSP_MEMDMATIMEW0          (*pMSP_MEMDMATIMEW0)
#define MSP_MEMDMATIMEW1          (*pMSP_MEMDMATIMEW1)
#define MSP_MEMDMATIMEW2          (*pMSP_MEMDMATIMEW2)
#define MSP_MEMDMATIMEW3          (*pMSP_MEMDMATIMEW3)
#define MSP_MEMDMATIMEW4          (*pMSP_MEMDMATIMEW4)
#define MSP_MEMDMATIMEW5          (*pMSP_MEMDMATIMEW5)
#define MSP_MEMDMATIMEW6          (*pMSP_MEMDMATIMEW6)
#define MSP_MEMDMATIMEW7          (*pMSP_MEMDMATIMEW7)
#define MSP_MEMDMATIMEW8          (*pMSP_MEMDMATIMEW8)
#define MSP_MEMDMASTRBW           (*pMSP_MEMDMASTRBW)
#define MSP_MEMNANDCTRLW          (*pMSP_MEMNANDCTRLW)
#define MSP_MEMNANDTIMEW          (*pMSP_MEMNANDTIMEW)
#define MSP_MEMNANDECC0W          (*pMSP_MEMNANDECC0W)
#define MSP_MEMNANDECC1W          (*pMSP_MEMNANDECC1W)
#define MSP_MEMNANDECC2W          (*pMSP_MEMNANDECC2W)
#define MSP_MEMNANDCNTW           (*pMSP_MEMNANDCNTW)
#define MSP_NFDATA                (*pMSP_NFDATA)
#define MSP_NFCMD                 (*pMSP_NFCMD)
#define MSP_NFADDR                (*pMSP_NFADDR)
#define MSP_MEMCFGX               (*pMSP_MEMCFGX)
#define MSP_MEMTIMEX0             (*pMSP_MEMTIMEX0)
#define MSP_MEMTIMEX1             (*pMSP_MEMTIMEX1)
#define MSP_MEMACTPWDX            (*pMSP_MEMACTPWDX)
#define MSP_MEMREFX               (*pMSP_MEMREFX)
#define MSP_DUAL920DATA0          (*pMSP_DUAL920DATA0)
#define MSP_DUAL920DATA1          (*pMSP_DUAL920DATA1)
#define MSP_DUAL920DATA2          (*pMSP_DUAL920DATA2)
#define MSP_DUAL920DATA3          (*pMSP_DUAL920DATA3)
#define MSP_DUAL920DATA4          (*pMSP_DUAL920DATA4)
#define MSP_DUAL920DATA5          (*pMSP_DUAL920DATA5)
#define MSP_DUAL920DATA6          (*pMSP_DUAL920DATA6)
#define MSP_DUAL920DATA7          (*pMSP_DUAL920DATA7)
#define MSP_DUAL920DATA8          (*pMSP_DUAL920DATA8)
#define MSP_DUAL920DATA9          (*pMSP_DUAL920DATA9)
#define MSP_DUAL920DATA10         (*pMSP_DUAL920DATA10)
#define MSP_DUAL920DATA11         (*pMSP_DUAL920DATA11)
#define MSP_DUAL920DATA12         (*pMSP_DUAL920DATA12)
#define MSP_DUAL920DATA13         (*pMSP_DUAL920DATA13)
#define MSP_DUAL920DATA14         (*pMSP_DUAL920DATA14)
#define MSP_DUAL920DATA15         (*pMSP_DUAL920DATA15)
#define MSP_DUAL940DATA0          (*pMSP_DUAL940DATA0)
#define MSP_DUAL940DATA1          (*pMSP_DUAL940DATA1)
#define MSP_DUAL940DATA2          (*pMSP_DUAL940DATA2)
#define MSP_DUAL940DATA3          (*pMSP_DUAL940DATA3)
#define MSP_DUAL940DATA4          (*pMSP_DUAL940DATA4)
#define MSP_DUAL940DATA5          (*pMSP_DUAL940DATA5)
#define MSP_DUAL940DATA6          (*pMSP_DUAL940DATA6)
#define MSP_DUAL940DATA7          (*pMSP_DUAL940DATA7)
#define MSP_DUAL940DATA8          (*pMSP_DUAL940DATA8)
#define MSP_DUAL940DATA9          (*pMSP_DUAL940DATA9)
#define MSP_DUAL940DATA10         (*pMSP_DUAL940DATA10)
#define MSP_DUAL940DATA11         (*pMSP_DUAL940DATA11)
#define MSP_DUAL940DATA12         (*pMSP_DUAL940DATA12)
#define MSP_DUAL940DATA13         (*pMSP_DUAL940DATA13)
#define MSP_DUAL940DATA14         (*pMSP_DUAL940DATA14)
#define MSP_DUAL940DATA15         (*pMSP_DUAL940DATA15)
#define MSP_DUALINT920            (*pMSP_DUALINT920)
#define MSP_DUALINT940            (*pMSP_DUALINT940)
#define MSP_DUALPEND920           (*pMSP_DUALPEND920)
#define MSP_DUALPEND940           (*pMSP_DUALPEND940)
#define MSP_DUALCTRL940           (*pMSP_DUALCTRL940)
#define MSP_PWMODEREG             (*pMSP_PWMODEREG)
#define MSP_CLKCHGSTREG           (*pMSP_CLKCHGSTREG)
#define MSP_SYSCLKENREG           (*pMSP_SYSCLKENREG)
#define MSP_COMCLKENREG           (*pMSP_COMCLKENREG)
#define MSP_VGCLKENREG            (*pMSP_VGCLKENREG)
#define MSP_ASCLKENREG            (*pMSP_ASCLKENREG)
#define MSP_FPLLSETVREG           (*pMSP_FPLLSETVREG)
#define MSP_FPLLVSETREG           (*pMSP_FPLLVSETREG)
#define MSP_UPLLSETVREG           (*pMSP_UPLLSETVREG)
#define MSP_UPLLVSETREG           (*pMSP_UPLLVSETREG)
#define MSP_APLLSETVREG           (*pMSP_APLLSETVREG)
#define MSP_APLLVSETREG           (*pMSP_APLLVSETREG)
#define MSP_SYSCSETREG            (*pMSP_SYSCSETREG)
#define MSP_ESYSCSETREG           (*pMSP_ESYSCSETREG)
#define MSP_UIRMCSETREG           (*pMSP_UIRMCSETREG)
#define MSP_AUDICSETREG           (*pMSP_AUDICSETREG)
#define MSP_DISPCSETREG           (*pMSP_DISPCSETREG)
#define MSP_IMAGCSETREG           (*pMSP_IMAGCSETREG)
#define MSP_URT0CSETREG           (*pMSP_URT0CSETREG)
#define MSP_URT1CSETREG           (*pMSP_URT1CSETREG)
#define MSP_A940TIDLE             (*pMSP_A940TIDLE)
#define MSP_SPDICSETREG           (*pMSP_SPDICSETREG)
#define MSP_RTCTSETREG            (*pMSP_RTCTSETREG)
#define MSP_RTCTCNTREG            (*pMSP_RTCTCNTREG)
#define MSP_RTCSTCNTREG           (*pMSP_RTCSTCNTREG)
#define MSP_TICKSETREG            (*pMSP_TICKSETREG)
#define MSP_ALARMTREG             (*pMSP_ALARMTREG)
#define MSP_PWRMGRREG             (*pMSP_PWRMGRREG)
#define MSP_CLKMGRREG             (*pMSP_CLKMGRREG)
#define MSP_RSTCTRLREG            (*pMSP_RSTCTRLREG)
#define MSP_RSTSTREG              (*pMSP_RSTSTREG)
#define MSP_BOOTCTRLREG           (*pMSP_BOOTCTRLREG)
#define MSP_LOCKTIMEREG           (*pMSP_LOCKTIMEREG)
#define MSP_RSTTIMEREG            (*pMSP_RSTTIMEREG)
#define MSP_EXTCTRLREG            (*pMSP_EXTCTRLREG)
#define MSP_STOPTSETREG           (*pMSP_STOPTSETREG)
#define MSP_RTCCTRLREG            (*pMSP_RTCCTRLREG)
#define MSP_RTCSTREG              (*pMSP_RTCSTREG)
#define MSP_DMAINT                (*pMSP_DMAINT)
#define MSP_DCH0SRM               (*pMSP_DCH0SRM)
#define MSP_DCH0TRM               (*pMSP_DCH0TRM)
#define MSP_DCH1SRM               (*pMSP_DCH1SRM)
#define MSP_DCH1TRM               (*pMSP_DCH1TRM)
#define MSP_DCH2SRM               (*pMSP_DCH2SRM)
#define MSP_DCH2TRM               (*pMSP_DCH2TRM)
#define MSP_DCH3SRM               (*pMSP_DCH3SRM)
#define MSP_DCH3TRM               (*pMSP_DCH3TRM)
#define MSP_DCH4SRM               (*pMSP_DCH4SRM)
#define MSP_DCH4TRM               (*pMSP_DCH4TRM)
#define MSP_DCH5SRM               (*pMSP_DCH5SRM)
#define MSP_DCH5TRM               (*pMSP_DCH5TRM)
#define MSP_DCH6SRM               (*pMSP_DCH6SRM)
#define MSP_DCH6TRM               (*pMSP_DCH6TRM)
#define MSP_DCH7SRM               (*pMSP_DCH7SRM)
#define MSP_DCH7TRM               (*pMSP_DCH7TRM)
#define MSP_DCH8SRM               (*pMSP_DCH8SRM)
#define MSP_DCH8TRM               (*pMSP_DCH8TRM)
#define MSP_DCH9SRM               (*pMSP_DCH9SRM)
#define MSP_DCH9TRM               (*pMSP_DCH9TRM)
#define MSP_DCH10SRM              (*pMSP_DCH10SRM)
#define MSP_DCH10TRM              (*pMSP_DCH10TRM)
#define MSP_DCH11SRM              (*pMSP_DCH11SRM)
#define MSP_DCH11TRM              (*pMSP_DCH11TRM)
#define MSP_DCH12SRM              (*pMSP_DCH12SRM)
#define MSP_DCH12TRM              (*pMSP_DCH12TRM)
#define MSP_DCH13SRM              (*pMSP_DCH13SRM)
#define MSP_DCH13TRM              (*pMSP_DCH13TRM)
#define MSP_DCH14SRM              (*pMSP_DCH14SRM)
#define MSP_DCH14TRM              (*pMSP_DCH14TRM)
#define MSP_DCH15SRM              (*pMSP_DCH15SRM)
#define MSP_DCH15TRM              (*pMSP_DCH15TRM)
#define MSP_DMA0COM0              (*pMSP_DMA0COM0)
#define MSP_DMA0COM1              (*pMSP_DMA0COM1)
#define MSP_DMA0COM2              (*pMSP_DMA0COM2)
#define MSP_DMA0CONS              (*pMSP_DMA0CONS)
#define MSP_SRC0LADDR             (*pMSP_SRC0LADDR)
#define MSP_SRC0HADDR             (*pMSP_SRC0HADDR)
#define MSP_TRG0LADDR             (*pMSP_TRG0LADDR)
#define MSP_TRG0HADDR             (*pMSP_TRG0HADDR)
#define MSP_DMA1COM0              (*pMSP_DMA1COM0)
#define MSP_DMA1COM1              (*pMSP_DMA1COM1)
#define MSP_DMA1COM2              (*pMSP_DMA1COM2)
#define MSP_DMA1CONS              (*pMSP_DMA1CONS)
#define MSP_SRC1LADDR             (*pMSP_SRC1LADDR)
#define MSP_SRC1HADDR             (*pMSP_SRC1HADDR)
#define MSP_TRG1LADDR             (*pMSP_TRG1LADDR)
#define MSP_TRG1HADDR             (*pMSP_TRG1HADDR)
#define MSP_DMA2COM0              (*pMSP_DMA2COM0)
#define MSP_DMA2COM1              (*pMSP_DMA2COM1)
#define MSP_DMA2COM2              (*pMSP_DMA2COM2)
#define MSP_DMA2CONS              (*pMSP_DMA2CONS)
#define MSP_SRC2LADDR             (*pMSP_SRC2LADDR)
#define MSP_SRC2HADDR             (*pMSP_SRC2HADDR)
#define MSP_TRG2LADDR             (*pMSP_TRG2LADDR)
#define MSP_TRG2HADDR             (*pMSP_TRG2HADDR)
#define MSP_DMA3COM0              (*pMSP_DMA3COM0)
#define MSP_DMA3COM1              (*pMSP_DMA3COM1)
#define MSP_DMA3COM2              (*pMSP_DMA3COM2)
#define MSP_DMA3CONS              (*pMSP_DMA3CONS)
#define MSP_SRC3LADDR             (*pMSP_SRC3LADDR)
#define MSP_SRC3HADDR             (*pMSP_SRC3HADDR)
#define MSP_TRG3LADDR             (*pMSP_TRG3LADDR)
#define MSP_TRG3HADDR             (*pMSP_TRG3HADDR)
#define MSP_DMA4COM0              (*pMSP_DMA4COM0)
#define MSP_DMA4COM1              (*pMSP_DMA4COM1)
#define MSP_DMA4COM2              (*pMSP_DMA4COM2)
#define MSP_DMA4CONS              (*pMSP_DMA4CONS)
#define MSP_SRC4LADDR             (*pMSP_SRC4LADDR)
#define MSP_SRC4HADDR             (*pMSP_SRC4HADDR)
#define MSP_TRG4LADDR             (*pMSP_TRG4LADDR)
#define MSP_TRG4HADDR             (*pMSP_TRG4HADDR)
#define MSP_DMA5COM0              (*pMSP_DMA5COM0)
#define MSP_DMA5COM1              (*pMSP_DMA5COM1)
#define MSP_DMA5COM2              (*pMSP_DMA5COM2)
#define MSP_DMA5CONS              (*pMSP_DMA5CONS)
#define MSP_SRC5LADDR             (*pMSP_SRC5LADDR)
#define MSP_SRC5HADDR             (*pMSP_SRC5HADDR)
#define MSP_TRG5LADDR             (*pMSP_TRG5LADDR)
#define MSP_TRG5HADDR             (*pMSP_TRG5HADDR)
#define MSP_DMA6COM0              (*pMSP_DMA6COM0)
#define MSP_DMA6COM1              (*pMSP_DMA6COM1)
#define MSP_DMA6COM2              (*pMSP_DMA6COM2)
#define MSP_DMA6CONS              (*pMSP_DMA6CONS)
#define MSP_SRC6LADDR             (*pMSP_SRC6LADDR)
#define MSP_SRC6HADDR             (*pMSP_SRC6HADDR)
#define MSP_TRG6LADDR             (*pMSP_TRG6LADDR)
#define MSP_TRG6HADDR             (*pMSP_TRG6HADDR)
#define MSP_DMA7COM0              (*pMSP_DMA7COM0)
#define MSP_DMA7COM1              (*pMSP_DMA7COM1)
#define MSP_DMA7COM2              (*pMSP_DMA7COM2)
#define MSP_DMA7CONS              (*pMSP_DMA7CONS)
#define MSP_SRC7LADDR             (*pMSP_SRC7LADDR)
#define MSP_SRC7HADDR             (*pMSP_SRC7HADDR)
#define MSP_TRG7LADDR             (*pMSP_TRG7LADDR)
#define MSP_TRG7HADDR             (*pMSP_TRG7HADDR)
#define MSP_DMA8COM0              (*pMSP_DMA8COM0)
#define MSP_DMA8COM1              (*pMSP_DMA8COM1)
#define MSP_DMA8COM2              (*pMSP_DMA8COM2)
#define MSP_DMA8CONS              (*pMSP_DMA8CONS)
#define MSP_SRC8LADDR             (*pMSP_SRC8LADDR)
#define MSP_SRC8HADDR             (*pMSP_SRC8HADDR)
#define MSP_TRG8LADDR             (*pMSP_TRG8LADDR)
#define MSP_TRG8HADDR             (*pMSP_TRG8HADDR)
#define MSP_DMA9COM0              (*pMSP_DMA9COM0)
#define MSP_DMA9COM1              (*pMSP_DMA9COM1)
#define MSP_DMA9COM2              (*pMSP_DMA9COM2)
#define MSP_DMA9CONS              (*pMSP_DMA9CONS)
#define MSP_SRC9LADDR             (*pMSP_SRC9LADDR)
#define MSP_SRC9HADDR             (*pMSP_SRC9HADDR)
#define MSP_TRG9LADDR             (*pMSP_TRG9LADDR)
#define MSP_TRG9HADDR             (*pMSP_TRG9HADDR)
#define MSP_DMA10COM0             (*pMSP_DMA10COM0)
#define MSP_DMA10COM1             (*pMSP_DMA10COM1)
#define MSP_DMA10COM2             (*pMSP_DMA10COM2)
#define MSP_DMA10CONS             (*pMSP_DMA10CONS)
#define MSP_SRC10LADDR            (*pMSP_SRC10LADDR)
#define MSP_SRC10HADDR            (*pMSP_SRC10HADDR)
#define MSP_TRG10LADDR            (*pMSP_TRG10LADDR)
#define MSP_TRG10HADDR            (*pMSP_TRG10HADDR)
#define MSP_DMA11COM0             (*pMSP_DMA11COM0)
#define MSP_DMA11COM1             (*pMSP_DMA11COM1)
#define MSP_DMA11COM2             (*pMSP_DMA11COM2)
#define MSP_DMA11CONS             (*pMSP_DMA11CONS)
#define MSP_SRC11LADDR            (*pMSP_SRC11LADDR)
#define MSP_SRC11HADDR            (*pMSP_SRC11HADDR)
#define MSP_TRG11LADDR            (*pMSP_TRG11LADDR)
#define MSP_TRG11HADDR            (*pMSP_TRG11HADDR)
#define MSP_DMA12COM0             (*pMSP_DMA12COM0)
#define MSP_DMA12COM1             (*pMSP_DMA12COM1)
#define MSP_DMA12COM2             (*pMSP_DMA12COM2)
#define MSP_DMA12CONS             (*pMSP_DMA12CONS)
#define MSP_SRC12LADDR            (*pMSP_SRC12LADDR)
#define MSP_SRC12HADDR            (*pMSP_SRC12HADDR)
#define MSP_TRG12LADDR            (*pMSP_TRG12LADDR)
#define MSP_TRG12HADDR            (*pMSP_TRG12HADDR)
#define MSP_DMA13COM0             (*pMSP_DMA13COM0)
#define MSP_DMA13COM1             (*pMSP_DMA13COM1)
#define MSP_DMA13COM2             (*pMSP_DMA13COM2)
#define MSP_DMA13CONS             (*pMSP_DMA13CONS)
#define MSP_SRC13LADDR            (*pMSP_SRC13LADDR)
#define MSP_SRC13HADDR            (*pMSP_SRC13HADDR)
#define MSP_TRG13LADDR            (*pMSP_TRG13LADDR)
#define MSP_TRG13HADDR            (*pMSP_TRG13HADDR)
#define MSP_DMA14COM0             (*pMSP_DMA14COM0)
#define MSP_DMA14COM1             (*pMSP_DMA14COM1)
#define MSP_DMA14COM2             (*pMSP_DMA14COM2)
#define MSP_DMA14CONS             (*pMSP_DMA14CONS)
#define MSP_SRC14LADDR            (*pMSP_SRC14LADDR)
#define MSP_SRC14HADDR            (*pMSP_SRC14HADDR)
#define MSP_TRG14LADDR            (*pMSP_TRG14LADDR)
#define MSP_TRG14HADDR            (*pMSP_TRG14HADDR)
#define MSP_DMA15COM0             (*pMSP_DMA15COM0)
#define MSP_DMA15COM1             (*pMSP_DMA15COM1)
#define MSP_DMA15COM2             (*pMSP_DMA15COM2)
#define MSP_DMA15CONS             (*pMSP_DMA15CONS)
#define MSP_SRC15LADDR            (*pMSP_SRC15LADDR)
#define MSP_SRC15HADDR            (*pMSP_SRC15HADDR)
#define MSP_TRG15LADDR            (*pMSP_TRG15LADDR)
#define MSP_TRG15HADDR            (*pMSP_TRG15HADDR)
#define MSP_SRCPEND               (*pMSP_SRCPEND)
#define MSP_INTMODE               (*pMSP_INTMODE)
#define MSP_INTMASK               (*pMSP_INTMASK)
#define MSP_PRIORITY              (*pMSP_PRIORITY)
#define MSP_INTPEND               (*pMSP_INTPEND)
#define MSP_INTOFFSET             (*pMSP_INTOFFSET)
#define MSP_TCOUNT                (*pMSP_TCOUNT)
#define MSP_TMATCH0               (*pMSP_TMATCH0)
#define MSP_TMATCH1               (*pMSP_TMATCH1)
#define MSP_TMATCH2               (*pMSP_TMATCH2)
#define MSP_TMATCH3               (*pMSP_TMATCH3)
#define MSP_TCONTROL              (*pMSP_TCONTROL)
#define MSP_TSTATUS               (*pMSP_TSTATUS)
#define MSP_TINTEN                (*pMSP_TINTEN)
#define MSP_LCON0                 (*pMSP_LCON0)
#define MSP_UCON0                 (*pMSP_UCON0)
#define MSP_FCON0                 (*pMSP_FCON0)
#define MSP_MCON0                 (*pMSP_MCON0)
#define MSP_TRSTATUS0             (*pMSP_TRSTATUS0)
#define MSP_ESTATUS0              (*pMSP_ESTATUS0)
#define MSP_FSTATUS0              (*pMSP_FSTATUS0)
#define MSP_MSTATUS0              (*pMSP_MSTATUS0)
#define MSP_THB0                  (*pMSP_THB0)
#define MSP_RHB0                  (*pMSP_RHB0)
#define MSP_BRD0                  (*pMSP_BRD0)
#define MSP_TIMEOUTREG0           (*pMSP_TIMEOUTREG0)
#define MSP_LCON1                 (*pMSP_LCON1)
#define MSP_UCON1                 (*pMSP_UCON1)
#define MSP_FCON1                 (*pMSP_FCON1)
#define MSP_MCON1                 (*pMSP_MCON1)
#define MSP_TRSTATUS1             (*pMSP_TRSTATUS1)
#define MSP_ESTATUS1              (*pMSP_ESTATUS1)
#define MSP_FSTATUS1              (*pMSP_FSTATUS1)
#define MSP_MSTATUS1              (*pMSP_MSTATUS1)
#define MSP_THB1                  (*pMSP_THB1)
#define MSP_RHB1                  (*pMSP_RHB1)
#define MSP_BRD1                  (*pMSP_BRD1)
#define MSP_TIMEOUTREG1           (*pMSP_TIMEOUTREG1)
#define MSP_LCON2                 (*pMSP_LCON2)
#define MSP_UCON2                 (*pMSP_UCON2)
#define MSP_FCON2                 (*pMSP_FCON2)
#define MSP_MCON2                 (*pMSP_MCON2)
#define MSP_TRSTATUS2             (*pMSP_TRSTATUS2)
#define MSP_ESTATUS2              (*pMSP_ESTATUS2)
#define MSP_FSTATUS2              (*pMSP_FSTATUS2)
#define MSP_MSTATUS2              (*pMSP_MSTATUS2)
#define MSP_THB2                  (*pMSP_THB2)
#define MSP_RHB2                  (*pMSP_RHB2)
#define MSP_BRD2                  (*pMSP_BRD2)
#define MSP_TIMEOUTREG2           (*pMSP_TIMEOUTREG2)
#define MSP_LCON3                 (*pMSP_LCON3)
#define MSP_UCON3                 (*pMSP_UCON3)
#define MSP_FCON3                 (*pMSP_FCON3)
#define MSP_MCON3                 (*pMSP_MCON3)
#define MSP_TRSTATUS3             (*pMSP_TRSTATUS3)
#define MSP_ESTATUS3              (*pMSP_ESTATUS3)
#define MSP_FSTATUS3              (*pMSP_FSTATUS3)
#define MSP_MSTATUS3              (*pMSP_MSTATUS3)
#define MSP_THB3                  (*pMSP_THB3)
#define MSP_RHB3                  (*pMSP_RHB3)
#define MSP_BRD3                  (*pMSP_BRD3)
#define MSP_TIMEOUTREG3           (*pMSP_TIMEOUTREG3)
#define MSP_INTSTATREG            (*pMSP_INTSTATREG)
#define MSP_PORTCON               (*pMSP_PORTCON)
#define MSP_GPIOAALTFNLOW         (*pMSP_GPIOAALTFNLOW)
#define MSP_GPIOBALTFNLOW         (*pMSP_GPIOBALTFNLOW)
#define MSP_GPIOCALTFNLOW         (*pMSP_GPIOCALTFNLOW)
#define MSP_GPIODALTFNLOW         (*pMSP_GPIODALTFNLOW)
#define MSP_GPIOEALTFNLOW         (*pMSP_GPIOEALTFNLOW)
#define MSP_GPIOFALTFNLOW         (*pMSP_GPIOFALTFNLOW)
#define MSP_GPIOGALTFNLOW         (*pMSP_GPIOGALTFNLOW)
#define MSP_GPIOHALTFNLOW         (*pMSP_GPIOHALTFNLOW)
#define MSP_GPIOIALTFNLOW         (*pMSP_GPIOIALTFNLOW)
#define MSP_GPIOJALTFNLOW         (*pMSP_GPIOJALTFNLOW)
#define MSP_GPIOKALTFNLOW         (*pMSP_GPIOKALTFNLOW)
#define MSP_GPIOLALTFNLOW         (*pMSP_GPIOLALTFNLOW)
#define MSP_GPIOMALTFNLOW         (*pMSP_GPIOMALTFNLOW)
#define MSP_GPIONALTFNLOW         (*pMSP_GPIONALTFNLOW)
#define MSP_GPIOOALTFNLOW         (*pMSP_GPIOOALTFNLOW)
#define MSP_GPIOAALTFNHI          (*pMSP_GPIOAALTFNHI)
#define MSP_GPIOBALTFNHI          (*pMSP_GPIOBALTFNHI)
#define MSP_GPIOCALTFNHI          (*pMSP_GPIOCALTFNHI)
#define MSP_GPIODALTFNHI          (*pMSP_GPIODALTFNHI)
#define MSP_GPIOEALTFNHI          (*pMSP_GPIOEALTFNHI)
#define MSP_GPIOFALTFNHI          (*pMSP_GPIOFALTFNHI)
#define MSP_GPIOGALTFNHI          (*pMSP_GPIOGALTFNHI)
#define MSP_GPIOHALTFNHI          (*pMSP_GPIOHALTFNHI)
#define MSP_GPIOIALTFNHI          (*pMSP_GPIOIALTFNHI)
#define MSP_GPIOJALTFNHI          (*pMSP_GPIOJALTFNHI)
#define MSP_GPIOKALTFNHI          (*pMSP_GPIOKALTFNHI)
#define MSP_GPIOLALTFNHI          (*pMSP_GPIOLALTFNHI)
#define MSP_GPIOMALTFNHI          (*pMSP_GPIOMALTFNHI)
#define MSP_GPIONALTFNHI          (*pMSP_GPIONALTFNHI)
#define MSP_GPIOOALTFNHI          (*pMSP_GPIOOALTFNHI)
#define MSP_GPIOAOUT              (*pMSP_GPIOAOUT)
#define MSP_GPIOBOUT              (*pMSP_GPIOBOUT)
#define MSP_GPIOCOUT              (*pMSP_GPIOCOUT)
#define MSP_GPIODOUT              (*pMSP_GPIODOUT)
#define MSP_GPIOEOUT              (*pMSP_GPIOEOUT)
#define MSP_GPIOFOUT              (*pMSP_GPIOFOUT)
#define MSP_GPIOGOUT              (*pMSP_GPIOGOUT)
#define MSP_GPIOHOUT              (*pMSP_GPIOHOUT)
#define MSP_GPIOIOUT              (*pMSP_GPIOIOUT)
#define MSP_GPIOJOUT              (*pMSP_GPIOJOUT)
#define MSP_GPIOKOUT              (*pMSP_GPIOKOUT)
#define MSP_GPIOLOUT              (*pMSP_GPIOLOUT)
#define MSP_GPIOMOUT              (*pMSP_GPIOMOUT)
#define MSP_GPIONOUT              (*pMSP_GPIONOUT)
#define MSP_GPIOOOUT              (*pMSP_GPIOOOUT)
#define MSP_GPIOAEVTTYPLOW        (*pMSP_GPIOAEVTTYPLOW)
#define MSP_GPIOBEVTTYPLOW        (*pMSP_GPIOBEVTTYPLOW)
#define MSP_GPIOCEVTTYPLOW        (*pMSP_GPIOCEVTTYPLOW)
#define MSP_GPIODEVTTYPLOW        (*pMSP_GPIODEVTTYPLOW)
#define MSP_GPIOEEVTTYPLOW        (*pMSP_GPIOEEVTTYPLOW)
#define MSP_GPIOFEVTTYPLOW        (*pMSP_GPIOFEVTTYPLOW)
#define MSP_GPIOGEVTTYPLOW        (*pMSP_GPIOGEVTTYPLOW)
#define MSP_GPIOHEVTTYPLOW        (*pMSP_GPIOHEVTTYPLOW)
#define MSP_GPIOIEVTTYPLOW        (*pMSP_GPIOIEVTTYPLOW)
#define MSP_GPIOJEVTTYPLOW        (*pMSP_GPIOJEVTTYPLOW)
#define MSP_GPIOKEVTTYPLOW        (*pMSP_GPIOKEVTTYPLOW)
#define MSP_GPIOLEVTTYPLOW        (*pMSP_GPIOLEVTTYPLOW)
#define MSP_GPIOMEVTTYPLOW        (*pMSP_GPIOMEVTTYPLOW)
#define MSP_GPIONEVTTYPLOW        (*pMSP_GPIONEVTTYPLOW)
#define MSP_GPIOOEVTTYPLOW        (*pMSP_GPIOOEVTTYPLOW)
#define MSP_GPIOAEVTTYPHI         (*pMSP_GPIOAEVTTYPHI)
#define MSP_GPIOBEVTTYPHI         (*pMSP_GPIOBEVTTYPHI)
#define MSP_GPIOCEVTTYPHI         (*pMSP_GPIOCEVTTYPHI)
#define MSP_GPIODEVTTYPHI         (*pMSP_GPIODEVTTYPHI)
#define MSP_GPIOEEVTTYPHI         (*pMSP_GPIOEEVTTYPHI)
#define MSP_GPIOFEVTTYPHI         (*pMSP_GPIOFEVTTYPHI)
#define MSP_GPIOGEVTTYPHI         (*pMSP_GPIOGEVTTYPHI)
#define MSP_GPIOHEVTTYPHI         (*pMSP_GPIOHEVTTYPHI)
#define MSP_GPIOIEVTTYPHI         (*pMSP_GPIOIEVTTYPHI)
#define MSP_GPIOJEVTTYPHI         (*pMSP_GPIOJEVTTYPHI)
#define MSP_GPIOKEVTTYPHI         (*pMSP_GPIOKEVTTYPHI)
#define MSP_GPIOLEVTTYPHI         (*pMSP_GPIOLEVTTYPHI)
#define MSP_GPIOMEVTTYPHI         (*pMSP_GPIOMEVTTYPHI)
#define MSP_GPIONEVTTYPHI         (*pMSP_GPIONEVTTYPHI)
#define MSP_GPIOOEVTTYPHI         (*pMSP_GPIOOEVTTYPHI)
#define MSP_GPIOAPUENB            (*pMSP_GPIOAPUENB)
#define MSP_GPIOBPUENB            (*pMSP_GPIOBPUENB)
#define MSP_GPIOCPUENB            (*pMSP_GPIOCPUENB)
#define MSP_GPIODPUENB            (*pMSP_GPIODPUENB)
#define MSP_GPIOEPUENB            (*pMSP_GPIOEPUENB)
#define MSP_GPIOFPUENB            (*pMSP_GPIOFPUENB)
#define MSP_GPIOGPUENB            (*pMSP_GPIOGPUENB)
#define MSP_GPIOHPUENB            (*pMSP_GPIOHPUENB)
#define MSP_GPIOIPUENB            (*pMSP_GPIOIPUENB)
#define MSP_GPIOJPUENB            (*pMSP_GPIOJPUENB)
#define MSP_GPIOKPUENB            (*pMSP_GPIOKPUENB)
#define MSP_GPIOLPUENB            (*pMSP_GPIOLPUENB)
#define MSP_GPIOMPUENB            (*pMSP_GPIOMPUENB)
#define MSP_GPIONPUENB            (*pMSP_GPIONPUENB)
#define MSP_GPIOOPUENB            (*pMSP_GPIOOPUENB)
#define MSP_GPIOAINTENB           (*pMSP_GPIOAINTENB)
#define MSP_GPIOBINTENB           (*pMSP_GPIOBINTENB)
#define MSP_GPIOCINTENB           (*pMSP_GPIOCINTENB)
#define MSP_GPIODINTENB           (*pMSP_GPIODINTENB)
#define MSP_GPIOEINTENB           (*pMSP_GPIOEINTENB)
#define MSP_GPIOFINTENB           (*pMSP_GPIOFINTENB)
#define MSP_GPIOGINTENB           (*pMSP_GPIOGINTENB)
#define MSP_GPIOHINTENB           (*pMSP_GPIOHINTENB)
#define MSP_GPIOIINTENB           (*pMSP_GPIOIINTENB)
#define MSP_GPIOJINTENB           (*pMSP_GPIOJINTENB)
#define MSP_GPIOKINTENB           (*pMSP_GPIOKINTENB)
#define MSP_GPIOLINTENB           (*pMSP_GPIOLINTENB)
#define MSP_GPIOMINTENB           (*pMSP_GPIOMINTENB)
#define MSP_GPIONINTENB           (*pMSP_GPIONINTENB)
#define MSP_GPIOOINTENB           (*pMSP_GPIOOINTENB)
#define MSP_GPIOAEVT              (*pMSP_GPIOAEVT)
#define MSP_GPIOBEVT              (*pMSP_GPIOBEVT)
#define MSP_GPIOCEVT              (*pMSP_GPIOCEVT)
#define MSP_GPIODEVT              (*pMSP_GPIODEVT)
#define MSP_GPIOEEVT              (*pMSP_GPIOEEVT)
#define MSP_GPIOFEVT              (*pMSP_GPIOFEVT)
#define MSP_GPIOGEVT              (*pMSP_GPIOGEVT)
#define MSP_GPIOHEVT              (*pMSP_GPIOHEVT)
#define MSP_GPIOIEVT              (*pMSP_GPIOIEVT)
#define MSP_GPIOJEVT              (*pMSP_GPIOJEVT)
#define MSP_GPIOKEVT              (*pMSP_GPIOKEVT)
#define MSP_GPIOLEVT              (*pMSP_GPIOLEVT)
#define MSP_GPIOMEVT              (*pMSP_GPIOMEVT)
#define MSP_GPIONEVT              (*pMSP_GPIONEVT)
#define MSP_GPIOOEVT              (*pMSP_GPIOOEVT)
#define MSP_GPIOPADSEL            (*pMSP_GPIOPADSEL)
#define MSP_GPIOGREVTSTAT         (*pMSP_GPIOGREVTSTAT)
#define MSP_GPIOAPINLVL           (*pMSP_GPIOAPINLVL)
#define MSP_GPIOBPINLVL           (*pMSP_GPIOBPINLVL)
#define MSP_GPIOCPINLVL           (*pMSP_GPIOCPINLVL)
#define MSP_GPIODPINLVL           (*pMSP_GPIODPINLVL)
#define MSP_GPIOEPINLVL           (*pMSP_GPIOEPINLVL)
#define MSP_GPIOFPINLVL           (*pMSP_GPIOFPINLVL)
#define MSP_GPIOGPINLVL           (*pMSP_GPIOGPINLVL)
#define MSP_GPIOHPINLVL           (*pMSP_GPIOHPINLVL)
#define MSP_GPIOIPINLVL           (*pMSP_GPIOIPINLVL)
#define MSP_GPIOJPINLVL           (*pMSP_GPIOJPINLVL)
#define MSP_GPIOKPINLVL           (*pMSP_GPIOKPINLVL)
#define MSP_GPIOLPINLVL           (*pMSP_GPIOLPINLVL)
#define MSP_GPIOMPINLVL           (*pMSP_GPIOMPINLVL)
#define MSP_GPIONPINLVL           (*pMSP_GPIONPINLVL)
#define MSP_GPIOOPINLVL           (*pMSP_GPIOOPINLVL)
#define MSP_GPIOEOUT_ALIVE        (*pMSP_GPIOEOUT_ALIVE)
#define MSP_GPIOEOUTENB_ALIVE     (*pMSP_GPIOEOUTENB_ALIVE)
#define MSP_GPIOEPUENB_ALIVE      (*pMSP_GPIOEPUENB_ALIVE)
#define MSP_GPIOFOUT_ALIVE        (*pMSP_GPIOFOUT_ALIVE)
#define MSP_GPIOFOUTENB_ALIVE     (*pMSP_GPIOFOUTENB_ALIVE)
#define MSP_GPIOFPUENB_ALIVE      (*pMSP_GPIOFPUENB_ALIVE)
#define MSP_GPIOGOUT_ALIVE        (*pMSP_GPIOGOUT_ALIVE)
#define MSP_GPIOGOUTENB_ALIVE     (*pMSP_GPIOGOUTENB_ALIVE)
#define MSP_GPIOGPUENB_ALIVE      (*pMSP_GPIOGPUENB_ALIVE)
#define MSP_GPIOIOUT_ALIVE        (*pMSP_GPIOIOUT_ALIVE)
#define MSP_GPIOIOUTENB_ALIVE     (*pMSP_GPIOIOUTENB_ALIVE)
#define MSP_GPIOIPUENB_ALIVE      (*pMSP_GPIOIPUENB_ALIVE)
#define MSP_GPIOJOUT_ALIVE        (*pMSP_GPIOJOUT_ALIVE)
#define MSP_GPIOJOUTENB_ALIVE     (*pMSP_GPIOJOUTENB_ALIVE)
#define MSP_GPIOJPUENB_ALIVE      (*pMSP_GPIOJPUENB_ALIVE)
#define MSP_SC_CMD                (*pMSP_SC_CMD)
#define MSP_SC_STATUS             (*pMSP_SC_STATUS)
#define MSP_SC_MIRROR             (*pMSP_SC_MIRROR)
#define MSP_SC_PRE_VRATIO         (*pMSP_SC_PRE_VRATIO)
#define MSP_SC_PRE_HRATIO         (*pMSP_SC_PRE_HRATIO)
#define MSP_SC_POST_VRATIOL       (*pMSP_SC_POST_VRATIOL)
#define MSP_SC_POST_VRATIOH       (*pMSP_SC_POST_VRATIOH)
#define MSP_SC_POST_HRATIOL       (*pMSP_SC_POST_HRATIOL)
#define MSP_SC_POST_HRATIOH       (*pMSP_SC_POST_HRATIOH)
#define MSP_SC_SRC_ODD_ADDRL      (*pMSP_SC_SRC_ODD_ADDRL)
#define MSP_SC_SRC_ODD_ADDRH      (*pMSP_SC_SRC_ODD_ADDRH)
#define MSP_SC_SRC_EVEN_ADDRL     (*pMSP_SC_SRC_EVEN_ADDRL)
#define MSP_SC_SRC_EVEN_ADDRH     (*pMSP_SC_SRC_EVEN_ADDRH)
#define MSP_SC_DST_ADDRL          (*pMSP_SC_DST_ADDRL)
#define MSP_SC_DST_ADDRH          (*pMSP_SC_DST_ADDRH)
#define MSP_SC_SRC_PXL_HEIGHT     (*pMSP_SC_SRC_PXL_HEIGHT)
#define MSP_SC_SRC_PXL_WIDTH      (*pMSP_SC_SRC_PXL_WIDTH)
#define MSP_SC_SRC_PXL_REQCNT     (*pMSP_SC_SRC_PXL_REQCNT)
#define MSP_SC_DST_PXL_HEIGHT     (*pMSP_SC_DST_PXL_HEIGHT)
#define MSP_SC_DST_PXL_WIDTH      (*pMSP_SC_DST_PXL_WIDTH)
#define MSP_SC_DST_WPXL_WIDTH     (*pMSP_SC_DST_WPXL_WIDTH)
#define MSP_SC_LUMA_OFFSET        (*pMSP_SC_LUMA_OFFSET)
#define MSP_SC_CB_OFFSET          (*pMSP_SC_CB_OFFSET)
#define MSP_SC_CR_OFFSET          (*pMSP_SC_CR_OFFSET)
#define MSP_SC_DELAY              (*pMSP_SC_DELAY)
#define MSP_SC_MEM_CNTR           (*pMSP_SC_MEM_CNTR)
#define MSP_SC_IRQ                (*pMSP_SC_IRQ)
#define MSP_SC_RESERVED           (*pMSP_SC_RESERVED)
#define MSP_MLC_OVLAY_CNTR        (*pMSP_MLC_OVLAY_CNTR)
#define MSP_MLC_YUV_EFFECT        (*pMSP_MLC_YUV_EFFECT)
#define MSP_MLC_YUV_CNTL          (*pMSP_MLC_YUV_CNTL)
#define MSP_MLC_YUVA_TP_HSC       (*pMSP_MLC_YUVA_TP_HSC)
#define MSP_MLC_YUVA_BT_HSC       (*pMSP_MLC_YUVA_BT_HSC)
#define MSP_MLC_YUVA_TP_HSCL      (*pMSP_MLC_YUVA_TP_HSCL)
#define MSP_MLC_YUVA_TP_HSCH      (*pMSP_MLC_YUVA_TP_HSCH)
#define MSP_MLC_YUVA_BT_HSCL      (*pMSP_MLC_YUVA_BT_HSCL)
#define MSP_MLC_YUVA_BT_HSCH      (*pMSP_MLC_YUVA_BT_HSCH)
#define MSP_MLC_YUVA_TP_PXW       (*pMSP_MLC_YUVA_TP_PXW)
#define MSP_MLC_YUVA_BT_PXW       (*pMSP_MLC_YUVA_BT_PXW)
#define MSP_MLC_YUVA_STX          (*pMSP_MLC_YUVA_STX)
#define MSP_MLC_YUVA_ENDX         (*pMSP_MLC_YUVA_ENDX)
#define MSP_MLC_YUVA_TP_STY       (*pMSP_MLC_YUVA_TP_STY)
#define MSP_MLC_YUVA_TP_ENDY      (*pMSP_MLC_YUVA_TP_ENDY)
#define MSP_MLC_YUVA_BT_ENDY      (*pMSP_MLC_YUVA_BT_ENDY)
#define MSP_MLC_YUVA_TP_OADRL     (*pMSP_MLC_YUVA_TP_OADRL)
#define MSP_MLC_YUVA_TP_OADRH     (*pMSP_MLC_YUVA_TP_OADRH)
#define MSP_MLC_YUVA_TP_EADRL     (*pMSP_MLC_YUVA_TP_EADRL)
#define MSP_MLC_YUVA_TP_EADRH     (*pMSP_MLC_YUVA_TP_EADRH)
#define MSP_MLC_YUVA_BT_OADRL     (*pMSP_MLC_YUVA_BT_OADRL)
#define MSP_MLC_YUVA_BT_OADRH     (*pMSP_MLC_YUVA_BT_OADRH)
#define MSP_MLC_YUVA_BT_EADRL     (*pMSP_MLC_YUVA_BT_EADRL)
#define MSP_MLC_YUVA_BT_EADRH     (*pMSP_MLC_YUVA_BT_EADRH)
#define MSP_MLC_YUVB_TP_HSC       (*pMSP_MLC_YUVB_TP_HSC)
#define MSP_MLC_YUVB_BT_HSC       (*pMSP_MLC_YUVB_BT_HSC)
#define MSP_MLC_YUVB_TP_VSCL      (*pMSP_MLC_YUVB_TP_VSCL)
#define MSP_MLC_YUVB_TP_VSCH      (*pMSP_MLC_YUVB_TP_VSCH)
#define MSP_MLC_YUVB_BT_VSCL      (*pMSP_MLC_YUVB_BT_VSCL)
#define MSP_MLC_YUVB_BT_VSCH      (*pMSP_MLC_YUVB_BT_VSCH)
#define MSP_MLC_YUVB_TP_PXW       (*pMSP_MLC_YUVB_TP_PXW)
#define MSP_MLC_YUVB_BT_PXW       (*pMSP_MLC_YUVB_BT_PXW)
#define MSP_MLC_YUVB_STX          (*pMSP_MLC_YUVB_STX)
#define MSP_MLC_YUVB_ENDX         (*pMSP_MLC_YUVB_ENDX)
#define MSP_MLC_YUVB_TP_STY       (*pMSP_MLC_YUVB_TP_STY)
#define MSP_MLC_YUVB_TP_ENDY      (*pMSP_MLC_YUVB_TP_ENDY)
#define MSP_MLC_YUVB_BT_ENDY      (*pMSP_MLC_YUVB_BT_ENDY)
#define MSP_MLC_YUVB_TP_OADRL     (*pMSP_MLC_YUVB_TP_OADRL)
#define MSP_MLC_YUVB_TP_OADRH     (*pMSP_MLC_YUVB_TP_OADRH)
#define MSP_MLC_YUVB_TP_EADRL     (*pMSP_MLC_YUVB_TP_EADRL)
#define MSP_MLC_YUVB_TP_EADRH     (*pMSP_MLC_YUVB_TP_EADRH)
#define MSP_MLC_YUVB_BT_OADRL     (*pMSP_MLC_YUVB_BT_OADRL)
#define MSP_MLC_YUVB_BT_OADRH     (*pMSP_MLC_YUVB_BT_OADRH)
#define MSP_MLC_YUVB_BT_EADRL     (*pMSP_MLC_YUVB_BT_EADRL)
#define MSP_MLC_YUVB_BT_EADRH     (*pMSP_MLC_YUVB_BT_EADRH)
#define MSP_MLC_STL_CNTL          (*pMSP_MLC_STL_CNTL)
#define MSP_MLC_STL_MIXMUX        (*pMSP_MLC_STL_MIXMUX)
#define MSP_MLC_RGB_ALPHAL        (*pMSP_MLC_RGB_ALPHAL)
#define MSP_MLC_STL_ALPHAL        (*pMSP_MLC_STL_ALPHAL)
#define MSP_MLC_STL_ALPHAH        (*pMSP_MLC_STL_ALPHAH)
#define MSP_MLC_STL1_STX          (*pMSP_MLC_STL1_STX)
#define MSP_MLC_STL1_ENDX         (*pMSP_MLC_STL1_ENDX)
#define MSP_MLC_STL1_STY          (*pMSP_MLC_STL1_STY)
#define MSP_MLC_STL1_ENDY         (*pMSP_MLC_STL1_ENDY)
#define MSP_MLC_STL2_STX          (*pMSP_MLC_STL2_STX)
#define MSP_MLC_STL2_ENDX         (*pMSP_MLC_STL2_ENDX)
#define MSP_MLC_STL2_STY          (*pMSP_MLC_STL2_STY)
#define MSP_MLC_STL2_ENDY         (*pMSP_MLC_STL2_ENDY)
#define MSP_MLC_STL3_STX          (*pMSP_MLC_STL3_STX)
#define MSP_MLC_STL3_ENDX         (*pMSP_MLC_STL3_ENDX)
#define MSP_MLC_STL3_STY          (*pMSP_MLC_STL3_STY)
#define MSP_MLC_STL3_ENDY         (*pMSP_MLC_STL3_ENDY)
#define MSP_MLC_STL4_STX          (*pMSP_MLC_STL4_STX)
#define MSP_MLC_STL4_ENDX         (*pMSP_MLC_STL4_ENDX)
#define MSP_MLC_STL4_STY          (*pMSP_MLC_STL4_STY)
#define MSP_MLC_STL4_ENDY         (*pMSP_MLC_STL4_ENDY)
#define MSP_MLC_STL_CKEY_GB       (*pMSP_MLC_STL_CKEY_GB)
#define MSP_MLC_STL_CKEY_R        (*pMSP_MLC_STL_CKEY_R)
#define MSP_MLC_STL_HSC           (*pMSP_MLC_STL_HSC)
#define MSP_MLC_STL_VSCL          (*pMSP_MLC_STL_VSCL)
#define MSP_MLC_STL_VSCH          (*pMSP_MLC_STL_VSCH)
#define MSP_MLC_STL_HW            (*pMSP_MLC_STL_HW)
#define MSP_MLC_STL_OADRL         (*pMSP_MLC_STL_OADRL)
#define MSP_MLC_STL_OADRH         (*pMSP_MLC_STL_OADRH)
#define MSP_MLC_STL_EADRL         (*pMSP_MLC_STL_EADRL)
#define MSP_MLC_STL_EADRH         (*pMSP_MLC_STL_EADRH)
#define MSP_MLC_OSD_OADRL         (*pMSP_MLC_OSD_OADRL)
#define MSP_MLC_OSD_OADRH         (*pMSP_MLC_OSD_OADRH)
#define MSP_MLC_OSD_EADRL         (*pMSP_MLC_OSD_EADRL)
#define MSP_MLC_OSD_EADRH         (*pMSP_MLC_OSD_EADRH)
#define MSP_MLC_HWC_CNTR          (*pMSP_MLC_HWC_CNTR)
#define MSP_MLC_HWC_STX           (*pMSP_MLC_HWC_STX)
#define MSP_MLC_HWC_STY           (*pMSP_MLC_HWC_STY)
#define MSP_MLC_HWC_FGR           (*pMSP_MLC_HWC_FGR)
#define MSP_MLC_HWC_FB            (*pMSP_MLC_HWC_FB)
#define MSP_MLC_HWC_BGR           (*pMSP_MLC_HWC_BGR)
#define MSP_MLC_HWC_BB            (*pMSP_MLC_HWC_BB)
#define MSP_MLC_HWC_OADRL         (*pMSP_MLC_HWC_OADRL)
#define MSP_MLC_HWC_OADRH         (*pMSP_MLC_HWC_OADRH)
#define MSP_MLC_HWC_EADRL         (*pMSP_MLC_HWC_EADRL)
#define MSP_MLC_HWC_EADRH         (*pMSP_MLC_HWC_EADRH)
#define MSP_MLC_LUMA_ENH          (*pMSP_MLC_LUMA_ENH)
#define MSP_MLC_HUECB1AB          (*pMSP_MLC_HUECB1AB)
#define MSP_MLC_HUECR1AB          (*pMSP_MLC_HUECR1AB)
#define MSP_MLC_HUECB2AB          (*pMSP_MLC_HUECB2AB)
#define MSP_MLC_HUECR2AB          (*pMSP_MLC_HUECR2AB)
#define MSP_MLC_HUECB3AB          (*pMSP_MLC_HUECB3AB)
#define MSP_MLC_HUECR3AB          (*pMSP_MLC_HUECR3AB)
#define MSP_MLC_HUECB4AB          (*pMSP_MLC_HUECB4AB)
#define MSP_MLC_HUECR4AB          (*pMSP_MLC_HUECR4AB)
#define MSP_MLC_DITHER            (*pMSP_MLC_DITHER)
#define MSP_MLC_OSD_PALLT_A       (*pMSP_MLC_OSD_PALLT_A)
#define MSP_MLC_OSD_PALLT_D       (*pMSP_MLC_OSD_PALLT_D)
#define MSP_MLC_STL_PALLT_A       (*pMSP_MLC_STL_PALLT_A)
#define MSP_MLC_STL_PALLT_D       (*pMSP_MLC_STL_PALLT_D)
#define MSP_MLC_GAMMA_A           (*pMSP_MLC_GAMMA_A)
#define MSP_MLC_GAMMA_D           (*pMSP_MLC_GAMMA_D)
#define MSP_MLC_SPU_CTRL          (*pMSP_MLC_SPU_CTRL)
#define MSP_MLC_SPU_DELAY         (*pMSP_MLC_SPU_DELAY)
#define MSP_MLC_SPU_BASEI_ADDRL   (*pMSP_MLC_SPU_BASEI_ADDRL)
#define MSP_MLC_SPU_BASEI_ADDRH   (*pMSP_MLC_SPU_BASEI_ADDRH)
#define MSP_MLC_SPU_HLI_ADDRL     (*pMSP_MLC_SPU_HLI_ADDRL)
#define MSP_MLC_SPU_HLI_ADDRH     (*pMSP_MLC_SPU_HLI_ADDRH)
#define MSP_MLC_SPU_PAL           (*pMSP_MLC_SPU_PAL)
#define MSP_DPC_CNTL              (*pMSP_DPC_CNTL)
#define MSP_DPC_FPICNTL           (*pMSP_DPC_FPICNTL)
#define MSP_DPC_FPIPOL1           (*pMSP_DPC_FPIPOL1)
#define MSP_DPC_FPIPOL2           (*pMSP_DPC_FPIPOL2)
#define MSP_DPC_FPIPOL3           (*pMSP_DPC_FPIPOL3)
#define MSP_DPC_FPIATV1           (*pMSP_DPC_FPIATV1)
#define MSP_DPC_FPIATV2           (*pMSP_DPC_FPIATV2)
#define MSP_DPC_FPIATV3           (*pMSP_DPC_FPIATV3)
#define MSP_DPC_X_MAX             (*pMSP_DPC_X_MAX)
#define MSP_DPC_Y_MAX             (*pMSP_DPC_Y_MAX)
#define MSP_DPC_HS_WIDTH          (*pMSP_DPC_HS_WIDTH)
#define MSP_DPC_HS_STR            (*pMSP_DPC_HS_STR)
#define MSP_DPC_HS_END            (*pMSP_DPC_HS_END)
#define MSP_DPC_V_SYNC            (*pMSP_DPC_V_SYNC)
#define MSP_DPC_V_END             (*pMSP_DPC_V_END)
#define MSP_DPC_TV_BNK            (*pMSP_DPC_TV_BNK)
#define MSP_DPC_DE                (*pMSP_DPC_DE)
#define MSP_DPC_PS                (*pMSP_DPC_PS)
#define MSP_DPC_FG                (*pMSP_DPC_FG)
#define MSP_DPC_LP                (*pMSP_DPC_LP)
#define MSP_DPC_CLKVH             (*pMSP_DPC_CLKVH)
#define MSP_DPC_CLKVL             (*pMSP_DPC_CLKVL)
#define MSP_DPC_POL               (*pMSP_DPC_POL)
#define MSP_DPC_CISSYNC           (*pMSP_DPC_CISSYNC)
#define MSP_DPC_MID_SYNC          (*pMSP_DPC_MID_SYNC)
#define MSP_DPC_C_SYNC            (*pMSP_DPC_C_SYNC)
#define MSP_DPC_Y_BLANK           (*pMSP_DPC_Y_BLANK)
#define MSP_DPC_C_BLANK           (*pMSP_DPC_C_BLANK)
#define MSP_DPC_YP_CSYNC          (*pMSP_DPC_YP_CSYNC)
#define MSP_DPC_YN_CSYNC          (*pMSP_DPC_YN_CSYNC)
#define MSP_DPC_CP_CSYNC          (*pMSP_DPC_CP_CSYNC)
#define MSP_DPC_CN_CSYNC          (*pMSP_DPC_CN_CSYNC)
#define MSP_DPC_INTR              (*pMSP_DPC_INTR)
#define MSP_DPC_CLKCNTL           (*pMSP_DPC_CLKCNTL)
#define MSP_MESGDSTCTRL           (*pMSP_MESGDSTCTRL)
#define MSP_MESGDSTADDR           (*pMSP_MESGDSTADDR)
#define MSP_MESGDSTSTRIDE         (*pMSP_MESGDSTSTRIDE)
#define MSP_MESGSRCCTRL           (*pMSP_MESGSRCCTRL)
#define MSP_MESGSRCADDR           (*pMSP_MESGSRCADDR)
#define MSP_MESGSRCSTRIDE         (*pMSP_MESGSRCSTRIDE)
#define MSP_MESGSRCFORCOLOR       (*pMSP_MESGSRCFORCOLOR)
#define MSP_MESGSRCBACKCOLOR      (*pMSP_MESGSRCBACKCOLOR)
#define MSP_MESGPATCTRL           (*pMSP_MESGPATCTRL)
#define MSP_MESGFORCOLOR          (*pMSP_MESGFORCOLOR)
#define MSP_MESGBACKCOLOR         (*pMSP_MESGBACKCOLOR)
#define MSP_MESGSIZE              (*pMSP_MESGSIZE)
#define MSP_MESGCTRL              (*pMSP_MESGCTRL)
#define MSP_MESGSTATUS            (*pMSP_MESGSTATUS)
#define MSP_MESGFIFOSTATUS        (*pMSP_MESGFIFOSTATUS)
#define MSP_MESGFIFO              (*pMSP_MESGFIFO)
#define MSP_MESGPAT               (*pMSP_MESGPAT)
#define MSP_ISP_CON_1             (*pMSP_ISP_CON_1)
#define MSP_ISP_CON_2             (*pMSP_ISP_CON_2)
#define MSP_ISP_INT_CON           (*pMSP_ISP_INT_CON)
#define MSP_ISP_VPERIOD           (*pMSP_ISP_VPERIOD)
#define MSP_ISP_VPW               (*pMSP_ISP_VPW)
#define MSP_ISP_HPERIOD           (*pMSP_ISP_HPERIOD)
#define MSP_ISP_HPW               (*pMSP_ISP_HPW)
#define MSP_ISP_HACT_ST           (*pMSP_ISP_HACT_ST)
#define MSP_ISP_H_WIDTH           (*pMSP_ISP_H_WIDTH)
#define MSP_ISP_VACT_ST           (*pMSP_ISP_VACT_ST)
#define MSP_ISP_V_HEIGHT          (*pMSP_ISP_V_HEIGHT)
#define MSP_ISP_SRC_HWIDTH        (*pMSP_ISP_SRC_HWIDTH)
#define MSP_ISP_RGBSV             (*pMSP_ISP_RGBSV)
#define MSP_ISP_RGBEV             (*pMSP_ISP_RGBEV)
#define MSP_ISP_RGBSH             (*pMSP_ISP_RGBSH)
#define MSP_ISP_RGBEH             (*pMSP_ISP_RGBEH)
#define MSP_ISP_AESV              (*pMSP_ISP_AESV)
#define MSP_ISP_AEEV              (*pMSP_ISP_AEEV)
#define MSP_ISP_AESH              (*pMSP_ISP_AESH)
#define MSP_ISP_AEEH              (*pMSP_ISP_AEEH)
#define MSP_ISP_AELIM             (*pMSP_ISP_AELIM)
#define MSP_ISP_AEVSTEP           (*pMSP_ISP_AEVSTEP)
#define MSP_ISP_AEHSTEP           (*pMSP_ISP_AEHSTEP)
#define MSP_ISP_AFSV              (*pMSP_ISP_AFSV)
#define MSP_ISP_AFEV              (*pMSP_ISP_AFEV)
#define MSP_ISP_AFSH              (*pMSP_ISP_AFSH)
#define MSP_ISP_AFEH              (*pMSP_ISP_AFEH)
#define MSP_ISP_CNTRSTGAIN        (*pMSP_ISP_CNTRSTGAIN)
#define MSP_ISP_CNTRSTLIM         (*pMSP_ISP_CNTRSTLIM)
#define MSP_ISP_RGBREC            (*pMSP_ISP_RGBREC)
#define MSP_ISP_HUECB0AB          (*pMSP_ISP_HUECB0AB)
#define MSP_ISP_HUECR0AB          (*pMSP_ISP_HUECR0AB)
#define MSP_ISP_HUECB1AB          (*pMSP_ISP_HUECB1AB)
#define MSP_ISP_HUECR1AB          (*pMSP_ISP_HUECR1AB)
#define MSP_ISP_HUECB2AB          (*pMSP_ISP_HUECB2AB)
#define MSP_ISP_HUECR2AB          (*pMSP_ISP_HUECR2AB)
#define MSP_ISP_HUECB3AB          (*pMSP_ISP_HUECB3AB)
#define MSP_ISP_HUECR3AB          (*pMSP_ISP_HUECR3AB)
#define MSP_ISP_YGAIN             (*pMSP_ISP_YGAIN)
#define MSP_ISP_CDLY              (*pMSP_ISP_CDLY)
#define MSP_ISP_BRIGHT            (*pMSP_ISP_BRIGHT)
#define MSP_ISP_YCLIP             (*pMSP_ISP_YCLIP)
#define MSP_ISP_YCLIPSLANT        (*pMSP_ISP_YCLIPSLANT)
#define MSP_ISP_CSUPSLANT         (*pMSP_ISP_CSUPSLANT)
#define MSP_ISP_SLICELMT          (*pMSP_ISP_SLICELMT)
#define MSP_ISP_WBLMT             (*pMSP_ISP_WBLMT)
#define MSP_ISP_WBACCCNTH         (*pMSP_ISP_WBACCCNTH)
#define MSP_ISP_WBACCCNTL         (*pMSP_ISP_WBACCCNTL)
#define MSP_ISP_WBCBACCH          (*pMSP_ISP_WBCBACCH)
#define MSP_ISP_WBCBACCL          (*pMSP_ISP_WBCBACCL)
#define MSP_ISP_WBCRACCH          (*pMSP_ISP_WBCRACCH)
#define MSP_ISP_WBCRACCL          (*pMSP_ISP_WBCRACCL)
#define MSP_ISP_AFACCH            (*pMSP_ISP_AFACCH)
#define MSP_ISP_AFACCL            (*pMSP_ISP_AFACCL)
#define MSP_ISP_WBCOORD0          (*pMSP_ISP_WBCOORD0)
#define MSP_ISP_WBCOORD1          (*pMSP_ISP_WBCOORD1)
#define MSP_ISP_WBCOORD2          (*pMSP_ISP_WBCOORD2)
#define MSP_ISP_WBCOORD3          (*pMSP_ISP_WBCOORD3)
#define MSP_ISP_WBCOORD4          (*pMSP_ISP_WBCOORD4)
#define MSP_ISP_WBCOORD5          (*pMSP_ISP_WBCOORD5)
#define MSP_ISP_WBCOORD6          (*pMSP_ISP_WBCOORD6)
#define MSP_ISP_WBCOORD7          (*pMSP_ISP_WBCOORD7)
#define MSP_ISP_WBCOORD8          (*pMSP_ISP_WBCOORD8)
#define MSP_ISP_WBCOORD9          (*pMSP_ISP_WBCOORD9)
#define MSP_ISP_WBCOORD10         (*pMSP_ISP_WBCOORD10)
#define MSP_ISP_WBCOORD11         (*pMSP_ISP_WBCOORD11)
#define MSP_ISP_WBCOORD12         (*pMSP_ISP_WBCOORD12)
#define MSP_ISP_WBCOORD13         (*pMSP_ISP_WBCOORD13)
#define MSP_ISP_WBCOORD14         (*pMSP_ISP_WBCOORD14)
#define MSP_ISP_WBCOORD15         (*pMSP_ISP_WBCOORD15)
#define MSP_ISP_MEMADDRL1         (*pMSP_ISP_MEMADDRL1)
#define MSP_ISP_MEMADDRH1         (*pMSP_ISP_MEMADDRH1)
#define MSP_ISP_GAMMA             (*pMSP_ISP_GAMMA)
#define MSP_ISP_GAMMD             (*pMSP_ISP_GAMMD)
#define MSP_ISP_AEACCH            (*pMSP_ISP_AEACCH)
#define MSP_ISP_AEACCL            (*pMSP_ISP_AEACCL)
#define MSP_ISP_CNTST_LLCNTL      (*pMSP_ISP_CNTST_LLCNTL)
#define MSP_ISP_CNTST_LLCNTH      (*pMSP_ISP_CNTST_LLCNTH)
#define MSP_ISP_CNTST_HLCNTL      (*pMSP_ISP_CNTST_HLCNTL)
#define MSP_ISP_CNTST_HLCNTH      (*pMSP_ISP_CNTST_HLCNTH)
#define MSP_ISP_CNTST_LLACCL      (*pMSP_ISP_CNTST_LLACCL)
#define MSP_ISP_CNTST_LLACCH      (*pMSP_ISP_CNTST_LLACCH)
#define MSP_ISP_CNTST_HLACCL      (*pMSP_ISP_CNTST_HLACCL)
#define MSP_ISP_CNTST_HLACCH      (*pMSP_ISP_CNTST_HLACCH)
#define MSP_ISP_MEMRGBADDRL       (*pMSP_ISP_MEMRGBADDRL)
#define MSP_ISP_MEMRGBADDRH       (*pMSP_ISP_MEMRGBADDRH)
#define MSP_ISP_MEMADDRL2         (*pMSP_ISP_MEMADDRL2)
#define MSP_ISP_MEMADDRH2         (*pMSP_ISP_MEMADDRH2)
#define MSP_ISP_DST_HWIDTH        (*pMSP_ISP_DST_HWIDTH)
#define MSP_PWM_PRES01            (*pMSP_PWM_PRES01)
#define MSP_PWM_DUTY0             (*pMSP_PWM_DUTY0)
#define MSP_PWM_DUTY1             (*pMSP_PWM_DUTY1)
#define MSP_PWM_PERIOD0           (*pMSP_PWM_PERIOD0)
#define MSP_PWM_PERIOD1           (*pMSP_PWM_PERIOD1)
#define MSP_PWM_PRES23            (*pMSP_PWM_PRES23)
#define MSP_PWM_DUTY2             (*pMSP_PWM_DUTY2)
#define MSP_PWM_DUTY3             (*pMSP_PWM_DUTY3)
#define MSP_PWM_PERIOD2           (*pMSP_PWM_PERIOD2)
#define MSP_PWM_PERIOD3           (*pMSP_PWM_PERIOD3)
#define MSP_PPMCTRL               (*pMSP_PPMCTRL)
#define MSP_PPMSTATUS             (*pMSP_PPMSTATUS)
#define MSP_PPMLOWPERIOD          (*pMSP_PPMLOWPERIOD)
#define MSP_PPMHIGHPERIOD         (*pMSP_PPMHIGHPERIOD)
#define MSP_FUNC_ADDR_REG         (*pMSP_FUNC_ADDR_REG)
#define MSP_PWR_REG               (*pMSP_PWR_REG)
#define MSP_EP_INT_REG            (*pMSP_EP_INT_REG)
#define MSP_USB_INT_REG           (*pMSP_USB_INT_REG)
#define MSP_EP_INT_EN_REG         (*pMSP_EP_INT_EN_REG)
#define MSP_USB_INT_EN_REG        (*pMSP_USB_INT_EN_REG)
#define MSP_FRAME_NUM1_REG        (*pMSP_FRAME_NUM1_REG)
#define MSP_FRAME_NUM2_REG        (*pMSP_FRAME_NUM2_REG)
#define MSP_INDEX_REG             (*pMSP_INDEX_REG)
#define MSP_MAXP_REG              (*pMSP_MAXP_REG)
#define MSP_EP0_CSR               (*pMSP_EP0_CSR)
#define MSP_IN_CSR1_REG           (*pMSP_IN_CSR1_REG)
#define MSP_IN_CSR2_REG           (*pMSP_IN_CSR2_REG)
#define MSP_OUT_MAXP_REG          (*pMSP_OUT_MAXP_REG)
#define MSP_OUT_CSR1_REG          (*pMSP_OUT_CSR1_REG)
#define MSP_OUT_CSR2_REG          (*pMSP_OUT_CSR2_REG)
#define MSP_OUT_FIFO_CNT1_REG     (*pMSP_OUT_FIFO_CNT1_REG)
#define MSP_OUT_FIFO_CNT2_REG     (*pMSP_OUT_FIFO_CNT2_REG)
#define MSP_EP0_FIFO_REG          (*pMSP_EP0_FIFO_REG)
#define MSP_EP1_FIFO_REG          (*pMSP_EP1_FIFO_REG)
#define MSP_EP2_FIFO_REG          (*pMSP_EP2_FIFO_REG)
#define MSP_EP3_FIFO_REG          (*pMSP_EP3_FIFO_REG)
#define MSP_EP4_FIFO_REG          (*pMSP_EP4_FIFO_REG)
#define MSP_EP1_DMA_CON           (*pMSP_EP1_DMA_CON)
#define MSP_EP1_DMA_FIFO          (*pMSP_EP1_DMA_FIFO)
#define MSP_EP1_DMA_TTC_L         (*pMSP_EP1_DMA_TTC_L)
#define MSP_EP1_DMA_TTC_M         (*pMSP_EP1_DMA_TTC_M)
#define MSP_EP1_DMA_TTC_H         (*pMSP_EP1_DMA_TTC_H)
#define MSP_EP2_DMA_CON           (*pMSP_EP2_DMA_CON)
#define MSP_EP2_DMA_FIFO          (*pMSP_EP2_DMA_FIFO)
#define MSP_EP2_DMA_TTC_L         (*pMSP_EP2_DMA_TTC_L)
#define MSP_EP2_DMA_TTC_M         (*pMSP_EP2_DMA_TTC_M)
#define MSP_EP2_DMA_TTC_H         (*pMSP_EP2_DMA_TTC_H)
#define MSP_EP3_DMA_CON           (*pMSP_EP3_DMA_CON)
#define MSP_EP3_DMA_FIFO          (*pMSP_EP3_DMA_FIFO)
#define MSP_EP3_DMA_TTC_L         (*pMSP_EP3_DMA_TTC_L)
#define MSP_EP3_DMA_TTC_M         (*pMSP_EP3_DMA_TTC_M)
#define MSP_EP3_DMA_TTC_H         (*pMSP_EP3_DMA_TTC_H)
#define MSP_EP4_DMA_CON           (*pMSP_EP4_DMA_CON)
#define MSP_EP4_DMA_FIFO          (*pMSP_EP4_DMA_FIFO)
#define MSP_EP4_DMA_TTC_L         (*pMSP_EP4_DMA_TTC_L)
#define MSP_EP4_DMA_TTC_M         (*pMSP_EP4_DMA_TTC_M)
#define MSP_EP4_DMA_TTC_H         (*pMSP_EP4_DMA_TTC_H)
#define MSP_I2CCON                (*pMSP_I2CCON)
#define MSP_I2CSTAT               (*pMSP_I2CSTAT)
#define MSP_I2CADD                (*pMSP_I2CADD)
#define MSP_I2CDS                 (*pMSP_I2CDS)
#define MSP_SPDIFCON              (*pMSP_SPDIFCON)
#define MSP_SPDIFINTENB           (*pMSP_SPDIFINTENB)
#define MSP_SPDIFSTATUS           (*pMSP_SPDIFSTATUS)
#define MSP_SPDIFTXDIV            (*pMSP_SPDIFTXDIV)
#define MSP_AC_CTRL_REG           (*pMSP_AC_CTRL_REG)
#define MSP_AC_CONFIG_REG         (*pMSP_AC_CONFIG_REG)
#define MSP_AC_STA_ENA_REG        (*pMSP_AC_STA_ENA_REG)
#define MSP_AC_STATUS_REG         (*pMSP_AC_STATUS_REG)
#define MSP_AC_STATE_MACHINE      (*pMSP_AC_STATE_MACHINE)
#define MSP_AC_CODEC_ADDR         (*pMSP_AC_CODEC_ADDR)
#define MSP_AC_CODEC_DATA         (*pMSP_AC_CODEC_DATA)
#define MSP_AC_CODEC_ACCESS       (*pMSP_AC_CODEC_ACCESS)
#define MSP_AC_CODEC_REG_BASE     (*pMSP_AC_CODEC_REG_BASE)
#define MSP_I2S_CTRL_REG          (*pMSP_I2S_CTRL_REG)
#define MSP_I2S_CONFIG_REG        (*pMSP_I2S_CONFIG_REG)
#define MSP_I2S_STA_ENA_REG       (*pMSP_I2S_STA_ENA_REG)
#define MSP_I2S_STATUS_REG        (*pMSP_I2S_STATUS_REG)
#define MSP_SSPSPICONT0           (*pMSP_SSPSPICONT0)
#define MSP_SSPSPICONT1           (*pMSP_SSPSPICONT1)
#define MSP_SSPSPIDATA            (*pMSP_SSPSPIDATA)
#define MSP_SSPSPISTAT            (*pMSP_SSPSPISTAT)
#define MSP_MPEGIFCONT            (*pMSP_MPEGIFCONT)
#define MSP_SDICON                (*pMSP_SDICON)
#define MSP_SDIPRE                (*pMSP_SDIPRE)
#define MSP_SDICMDARGL            (*pMSP_SDICMDARGL)
#define MSP_SDICMDARGH            (*pMSP_SDICMDARGH)
#define MSP_SDICMDCON             (*pMSP_SDICMDCON)
#define MSP_SDICMDSTA             (*pMSP_SDICMDSTA)
#define MSP_SDIRSP0               (*pMSP_SDIRSP0)
#define MSP_SDIRSP1               (*pMSP_SDIRSP1)
#define MSP_SDIRSP2               (*pMSP_SDIRSP2)
#define MSP_SDIRSP3               (*pMSP_SDIRSP3)
#define MSP_SDIRSP4               (*pMSP_SDIRSP4)
#define MSP_SDIRSP5               (*pMSP_SDIRSP5)
#define MSP_SDIRSP6               (*pMSP_SDIRSP6)
#define MSP_SDIRSP7               (*pMSP_SDIRSP7)
#define MSP_SDIBSIZE              (*pMSP_SDIBSIZE)
#define MSP_SDIDATCONL            (*pMSP_SDIDATCONL)
#define MSP_SDIDATCONH            (*pMSP_SDIDATCONH)
#define MSP_SDIDATCNTL            (*pMSP_SDIDATCNTL)
#define MSP_SDIDATCNTH            (*pMSP_SDIDATCNTH)
#define MSP_SDIDATSTA             (*pMSP_SDIDATSTA)
#define MSP_SDIFSTA               (*pMSP_SDIFSTA)
#define MSP_SDIDAT                (*pMSP_SDIDAT)
#define MSP_SDIINTMSK0            (*pMSP_SDIINTMSK0)
#define MSP_SDIINTMSK1            (*pMSP_SDIINTMSK1)
#define MSP_SDIDTIMERL            (*pMSP_SDIDTIMERL)
#define MSP_SDIDTIMERH            (*pMSP_SDIDTIMERH)
#define MSP_SDISFTREGL            (*pMSP_SDISFTREGL)
#define MSP_SDISFTREGH            (*pMSP_SDISFTREGH)
#define MSP_MSCTRL                (*pMSP_MSCTRL)
#define MSP_MSFIFO                (*pMSP_MSFIFO)
#define MSP_MSFIFOST              (*pMSP_MSFIFOST)
#define MSP_MSINTCTRL             (*pMSP_MSINTCTRL)
#define MSP_MSCRC                 (*pMSP_MSCRC)
#define MSP_MSTIMEOUT             (*pMSP_MSTIMEOUT)
#define MSP_MSTPC                 (*pMSP_MSTPC)
#define MSP_MSTPCCNT              (*pMSP_MSTPCCNT)
#define MSP_OWM_CMD               (*pMSP_OWM_CMD)
#define MSP_OWM_BUF               (*pMSP_OWM_BUF)
#define MSP_OWM_STA               (*pMSP_OWM_STA)
#define MSP_OWM_STA_ENB           (*pMSP_OWM_STA_ENB)
#define MSP_OWM_CLK_DIVISOR       (*pMSP_OWM_CLK_DIVISOR)
#define MSP_OWM_CTRL              (*pMSP_OWM_CTRL)
#define MSP_OWM_IRQ               (*pMSP_OWM_IRQ)
#define MSP_DECSSCTRL             (*pMSP_DECSSCTRL)
#define MSP_DECSSKEY0             (*pMSP_DECSSKEY0)
#define MSP_DECSSKEY1             (*pMSP_DECSSKEY1)
#define MSP_DECSSKEY2             (*pMSP_DECSSKEY2)
#define MSP_DECSSDATA             (*pMSP_DECSSDATA)
#define MSP_DECSSNUM              (*pMSP_DECSSNUM)

#endif /* MMSP2_940_2_H_INCLUDED */
